loadf `UU_UNITY` ;;
load_library `more_arithmetic` ;;
load_theory `compor` ;;
new_theory `MD1` ;;
new_parent `aci` ;;
loadf `OLD_RES`;;
loadf `faust88`;;

let Autoload thy =    
    let defs = map fst (definitions thy) in
        map (\name. autoload_theory(`definition`,thy,name)) defs;
    let thms = map fst (theorems thy) in
        map (\name. autoload_theory(`theorem`,thy,name)) thms; 
    delete_cache thy; ();;

Autoload `reach`;;
Autoload `con` ;;
Autoload `compor` ;;
Autoload `aci` ;;
loadf `reach_SUP`;;
map (delete_cache o fst) (cached_theories()) ;;

loadf `DerLem` ;;

ADD_TRANS ("==>", 
    prove("!p q r. (p==>q) /\ (q==>r) ==> (p==>r)",TAUT_TAC)) ;;

ADD_TRANS ("<==", 
    prove("!p q r. (p<==q) /\ (q<==r) ==> (p<==r)",
    REWRITE_TAC[definition `predicate` `IMPBY`] THEN TAUT_TAC)) ;;


%---------------------------------------------------------------------------------
  This section defines various basic operators and proves their properties.
  
       - Id: the identity of a binary operator, if one exists
       - Idempot: defines the meaning of idempotentcy 
       - Fold: set reduction. Defined as a special case of ACI_OP operator 
       - UNI: quantified union
       
---------------------------------------------------------------------------------%

let ID_LEMMA = prove
  ("?Id. !plus.  
          (?u:*. LEFT_ID plus u /\ RIGHT_ID plus u)
          ==>
          LEFT_ID plus (Id plus) /\ RIGHT_ID plus (Id plus)",
    EXISTS_TAC "\plus.@u:*. LEFT_ID plus u /\ RIGHT_ID plus u"
    THEN GEN_TAC THEN
    CONV_TAC (ONCE_DEPTH_CONV BETA_CONV) THEN
    DISCH_TAC THEN 
    CONV_TAC SELECT_CONV THEN
    FIRST_ASSUM MATCH_ACCEPT_TAC);;

let Id = new_specification `Id` [`constant`, `Id`] ID_LEMMA;;

let Idempot_DEF = new_definition
  (`Idempot_DEF`,
   "Idempot (op:*->*->*) = (!a. op a a = a)") ;;

let Fold_DEF = new_definition
  (`Fold_DEF`,
   "Fold (op:*->*->*) V = ACI_OP op (Id op) I V") ;;

%---------------------------------------------------------------------------------
    The identity of an operator op, if exists, is unique and it can be denoted
    by Id op.
---------------------------------------------------------------------------------%

let MK_ABS_CONV x f = SYM(BETA_CONV "(\^x. ^f) ^x") ;;

let Id_UNIQUE = prove_thm
  (`Id_UNIQUE`,
   "!plus (e:*). LEFT_ID plus e /\ RIGHT_ID plus e ==> (Id plus=e)",
    REWRITE_TAC [LEFT_ID_DEF; RIGHT_ID_DEF]
    THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC (REWRITE_RULE [LEFT_ID_DEF; RIGHT_ID_DEF] Id)
    THEN SUBGOAL_THEN "(e:*) = plus (Id plus) e" SUBST1_TAC
    THENL [ ASM_REWRITE_TAC[] ; ALL_TAC ]
    THEN REWRITE_TAC [ ASSUME "!x:*. plus x (e:*) = x"]) ;;

%---------------------------------------------------------------------------------
    A special case of above theorem
---------------------------------------------------------------------------------%

let ACI_Id_Id = prove_thm
  (`ACI_Id_Id`,
   "!(f:*->*->*) e. ASSOC_COMM_ID f e ==> (Id f = e)",
    REWRITE_TAC [ASSOC_COMM_ID_DEF] THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC Id_UNIQUE
    THEN REWRITE_TAC [LEFT_ID_DEF; RIGHT_ID_DEF]
    THEN REWRITE_TAC [ASSUME "!a:*. f a (e:*) = a"]
    THEN ONCE_REWRITE_TAC [ASSUME "!a b. (f:*->*->*) a b = f b a"]
    THEN REWRITE_TAC [ASSUME "!a:*. f a (e:*) = a"]) ;;

%---------------------------------------------------------------------------------
    Folding f on an empty set results the identity of f, if exists
---------------------------------------------------------------------------------%

let Fold_EMPTY = prove_thm
  (`Fold_EMPTY`,
   "!(f:*->*->*) e. ASSOC_COMM_ID f e ==> (Fold f {} = e)",
    REWRITE_TAC [Fold_DEF; ACI_OP_DEF] 
    THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC ACI_OP_DEF 
    THEN IMP_RES_TAC ACI_Id_Id THEN ASM_REWRITE_TAC[]) ;;

%---------------------------------------------------------------------------------
    Folding f on a singleton {a} results a
---------------------------------------------------------------------------------%

let Fold_SING = prove_thm
  (`Fold_SING`,
   "!(f:*->*->*) e a. ASSOC_COMM_ID f e ==> (Fold f {a} = a)",
    REWRITE_TAC [Fold_DEF; ACI_OP_DEF]
    THEN REPEAT STRIP_TAC
    THEN (CONV_TAC o RAND_CONV o ONCE_REWRITE_CONV) 
         [(GEN_ALL o SYM o SPEC_ALL) I_THM]
    THEN MATCH_MP_TAC ACI_OP_SING
    THEN IMP_RES_TAC ACI_Id_Id THEN ASM_REWRITE_TAC[]) ;;

%--------------------------------------------------------------------------------- 
   op/(a+V)  =  a op (op/V)      , if a is not already in V
---------------------------------------------------------------------------------%

let Fold_INSERT = prove_thm
  (`Fold_INSERT`,
   "!(f:*->*->*) e V a. 
        ASSOC_COMM_ID f e /\ FINITE V /\ ~(a IN V) ==> 
           (Fold f (a INSERT V) = f a (Fold f V))",
    REWRITE_TAC [Fold_DEF]
    THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC ACI_OP_DEF
    THEN IMP_RES_TAC ACI_Id_Id
    THEN ASM_REWRITE_TAC [I_THM]) ;;

%--------------------------------------------------------------------------------- 
    op/(V U W)   =   (op/V) op (op/V)    , if V and W are disjunct
---------------------------------------------------------------------------------%

let Fold_DISJOINT = prove_thm
  (`Fold_DISJOINT`,
   "!(f:*->*->*) e V W. 
        ASSOC_COMM_ID f e /\ FINITE V /\ FINITE W /\ DISJOINT V W
        ==>
        (Fold f (V UNION W) = f (Fold f V) (Fold f W))",
    REWRITE_TAC [Fold_DEF]
    THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC (QUAFF_RULE 4 ACI_OP_DISJOINT)
    THEN IMP_RES_TAC ACI_Id_Id
    THEN ASM_REWRITE_TAC []) ;;

%--------------------------------------------------------------------------------- 
    op/V   =   a op (op/(V-a))    , if a is a member of V
---------------------------------------------------------------------------------%

let Fold_DELETE = prove_thm
  (`Fold_DELETE`,
   "!(f:*->*->*) e V a. 
        ASSOC_COMM_ID f e /\ FINITE V /\ a IN V ==> 
           (Fold f V = f a (Fold f (V DELETE a)))",
    REPEAT STRIP_TAC
    THEN SUBGOAL_THEN "V = ((a:*) INSERT (V DELETE a))"
         (\thm. (CONV_TAC o RATOR_CONV o ONCE_DEPTH_CONV o ONCE_REWRITE_CONV) [thm])
    THENL [ IMP_RES_TAC INSERT_DELETE THEN ASM_REWRITE_TAC[]; 
            ALL_TAC ]
    THEN MATCH_MP_TAC Fold_INSERT
    THEN EXISTS_TAC "e:*" 
    THEN ASM_REWRITE_TAC [IN_DELETE; FINITE_DELETE]) ;;

%--------------------------------------------------------------------------------- 
   op/g*V  =  g.a op (op/((g*V)-g.a))      , if a is e member of V
---------------------------------------------------------------------------------%

let Fold_IMG_DELETE = prove_thm
  (`Fold_IMG_DELETE`,
   "!(f:*->*->*) e (P:**->*) V (a:**). 
        ASSOC_COMM_ID f e /\ FINITE V /\ a IN V ==> 
           (Fold f (IMAGE P V) = f (P a) (Fold f ((IMAGE P V) DELETE (P a))))",
    REPEAT STRIP_TAC
    THEN SUBGOAL_THEN "FINITE (IMAGE (P:**->*) V)" ASSUME_TAC
    THENL [ MATCH_MP_TAC (QUAFF_RULE 3 IMAGE_FINITE) THEN ASM_REWRITE_TAC[] ;
            ALL_TAC ]
    THEN IMP_RES_TAC IMAGE_IN
    THEN SUBGOAL_THEN "((P:**->*) a) IN (IMAGE P V)" ASSUME_TAC
    THENL [ MATCH_MP_TAC (QUAFF_RULE 3 IMAGE_IN) THEN ASM_REWRITE_TAC[] ;
            ALL_TAC ]
    THEN IMP_RES_TAC Fold_DELETE) ;;

%--------------------------------------------------------------------------------- 
   If op is idempotent then we can drop the requirement "a not already in V" in
   Fold_INSERT. So:

       op/(a+V)  =  a op (op/V)       holds for any a

---------------------------------------------------------------------------------%

let Fold_Idem_INSERT = prove_thm
  (`Fold_Idem_INSERT`,
   "!(f:*->*->*) e V a. 
        ASSOC_COMM_ID f e /\ Idempot f /\ FINITE V ==> 
           (Fold f (a INSERT V) = f a (Fold f V))",
    REWRITE_TAC [Idempot_DEF] THEN REPEAT STRIP_TAC
    THEN ASM_CASES_TAC "(a:*) IN V"
    THENL 
    [ %-- 1 --%
      IMP_RES_TAC Fold_DELETE THEN ASM_REWRITE_TAC[]
      THEN SUBGOAL_THEN "!a b c. (f:*->*->*) a(f b c) = f (f a b) c"
           (\thm. REWRITE_TAC [thm])
      THENL [ IMP_RES_TAC ASSOC_COMM_ID_DEF; ALL_TAC ]
      THEN XRULE_ASSUM_TAC (REWRITE_RULE [ABSORPTION])
      THEN ASM_REWRITE_TAC [] ;
      %-- 2 --%
      IMP_RES_TAC Fold_INSERT ]) ;;
   
%--------------------------------------------------------------------------------- 
   For an idempotent op:

       op/g*(a+V) = g.a op (op/g*V)

---------------------------------------------------------------------------------%

let Fold_Idem_IMG_INSERT = prove_thm
  (`Fold_Idem_IMG_INSERT`,
   "!(f:*->*->*) e (P:**->*) V (a:**). 
        ASSOC_COMM_ID f e /\ Idempot f /\ FINITE V ==> 
           (Fold f (IMAGE P ( a INSERT V)) = f (P a) (Fold f (IMAGE P V)))",
    REPEAT STRIP_TAC
    THEN REWRITE_TAC [IMAGE_INSERT]
    THEN SUBGOAL_THEN "FINITE (IMAGE (P:**->*) V)" ASSUME_TAC
    THENL [ MATCH_MP_TAC (QUAFF_RULE 3 IMAGE_FINITE) THEN ASM_REWRITE_TAC[] ;
            ALL_TAC ]
    THEN IMP_RES_TAC Fold_Idem_INSERT
    THEN ASM_REWRITE_TAC[]) ;;

%---------------------------------------------------------------------------------
                          Define quatified union
---------------------------------------------------------------------------------%

let RES_UNI = new_definition
  (`RES_UNI`,
   "RES_UNI P Q = { x:* | ?i:**. P i /\ x IN (Q i) }") ;;

new_binder (`UNI`, ":(* -> **) -> ***") ;;

associate_restriction (`UNI`,`RES_UNI`) ;;

%---------------------------------------------------------------------------------
                          Some properties of UNI
---------------------------------------------------------------------------------%

let IN_UNI = prove_thm
  (`IN_UNI`,
   "!(x:*) P Q. 
        x IN (UNI i::P. Q i) = (?i:**. P i /\ x IN (Q i))",
    REWRITE_TAC [RES_UNI] THEN BETA_TAC
    THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
    THEN REWRITE_TAC []) ;;

let FF_UNI = prove_thm
  (`FF_UNI`,
   "!Q. (UNI i:**::FF. Q i) = ({}:(*)set)",
    REWRITE_TAC [EXTENSION; IN_UNI; FF_DEF; NOT_IN_EMPTY]) ;;

%---------------------------------------------------------------------------------
   Now we proves more UNITY properties, mostly dealing with  Fold PAR. The main 
   result we are aiming is a generalized Transparency principle for Fold PAR.
---------------------------------------------------------------------------------%

%---------------------------------------------------------------------------------
   Prove that (FF,TT,FF,FF) is equal to Id PAR
---------------------------------------------------------------------------------%
    
let PAR_UNIT_Id = prove_thm
  (`PAR_UNIT_Id`,
   "Id $PAR = ((FF,TT,FF,FF):^Uprog)",
    MATCH_MP_TAC Id_UNIQUE
    THEN REWRITE_TAC [LEFT_ID_DEF; RIGHT_ID_DEF; PAR_UNIT]
    THEN ONCE_REWRITE_TAC [PAR_SYM]
    THEN REWRITE_TAC [PAR_UNIT]) ;;

%---------------------------------------------------------------------------------
   PAR is an ACI (assoc, comm, with id) operator
---------------------------------------------------------------------------------%

let ACI_PAR = prove_thm
  (`ACI_PAR`,
   "ASSOC_COMM_ID $PAR ((FF,TT,FF,FF):^Uprog)",
    REWRITE_TAC [ASSOC_COMM_ID_DEF; PAR_ASSOC; PAR_UNIT]
    THEN ACCEPT_TAC PAR_SYM) ;;

   
%---------------------------------------------------------------------------------
    PAR is an idempotent operator
---------------------------------------------------------------------------------%

let CHF_EMPTY = prove_thm
  (`CHF_EMPTY`,
   "CHF ({}:(*)set) = FF",
    CONV_TAC FUN_EQ_CONV
    THEN REWRITE_TAC [FF_DEF; SYM (SPEC_ALL IN_DEF)]
    THEN BETA_TAC THEN REWRITE_TAC [NOT_IN_EMPTY]) ;;

let PAR_Idem = prove_thm
  (`PAR_Idem`,
   "Idempot ($PAR:^Uprog->^Uprog->^Uprog)",
    REWRITE_TAC [Idempot_DEF; PAR; pAND_REFL; pOR_REFL; 
                 SYM (SPEC_ALL ID_UPROG)]) ;;

%---------------------------------------------------------------------------------
    Uprog extensionality theorem: defines when two UNITY programs are equal
---------------------------------------------------------------------------------%

let UPROG_EXT = prove_thm
  (`UPROG_EXT`,
   "!Pr (Qr:^Uprog).
       (Pr = Qr) = (PROG Pr = PROG Qr) /\
                    (INIT Pr = INIT Qr) /\
                    (READ Pr = READ Qr) /\
                    (WRITE Pr = WRITE Qr)",
    REPEAT GEN_TAC 
    THEN (CONV_TAC o RATOR_CONV o ONCE_REWRITE_CONV) [ID_UPROG]
    THEN REWRITE_TAC [PAIR_EQ]) ;;

%---------------------------------------------------------------------------------
    PAR/{} = the unit of PAR
---------------------------------------------------------------------------------%

let PAR_Fold_EMPTY = prove_thm
  (`PAR_Fold_EMPTY`,
   "Fold $PAR ({}:(^Uprog)set) = (FF,TT,FF,FF)",
    MATCH_MP_TAC Fold_EMPTY
    THEN REWRITE_TAC [ACI_PAR]) ;;
%---------------------------------------------------------------------------------
    PAR/{a} = a
---------------------------------------------------------------------------------%
  
let PAR_Fold_SING = prove_thm
  (`PAR_Fold_SING`,
   "!(a:^Uprog). Fold $PAR {a} = a",
    MATCH_MP_TAC Fold_SING 
    THEN EXISTS_TAC "(FF,TT,FF,FF):^Uprog"
    THEN REWRITE_TAC [ACI_PAR]) ;;

%---------------------------------------------------------------------------------
    PAR/(a+V) = a PAR (PAR/V)
---------------------------------------------------------------------------------%

let PAR_Fold_INSERT = prove_thm
  (`PAR_Fold_INSERT`,
  "!(V:(^Uprog)set) a. 
        FINITE V ==> 
           (Fold $PAR (a INSERT V) = a PAR (Fold $PAR V))",
    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC Fold_Idem_INSERT
    THEN EXISTS_TAC "(FF,TT,FF,FF):^Uprog"
    THEN ASM_REWRITE_TAC [ACI_PAR; PAR_Idem]) ;;

%---------------------------------------------------------------------------------
    PAR/g*(a+V) = g.a PAR (PAR/g*V)
---------------------------------------------------------------------------------%

let PAR_Fold_IMG_INSERT = prove_thm
  (`PAR_Fold_IMG_INSERT`,
  "!(P:*->^Uprog) V a. 
        FINITE V ==> 
           (Fold $PAR (IMAGE P (a INSERT V)) = 
            (P a) PAR (Fold $PAR (IMAGE P V)))",
    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC Fold_Idem_IMG_INSERT
    THEN EXISTS_TAC "(FF,TT,FF,FF):^Uprog"
    THEN ASM_REWRITE_TAC [ACI_PAR; PAR_Idem]) ;;

%---------------------------------------------------------------------------------
    PAR/V = a PAR (PAR/(V-a))             , if a is a member of V
---------------------------------------------------------------------------------%

let PAR_Fold_DELETE = prove_thm
  (`PAR_Fold_DELETE`,
   "!(V:(^Uprog)set) a. 
        FINITE V /\ a IN V ==> 
           (Fold $PAR V = a PAR (Fold $PAR (V DELETE a)))",
    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC Fold_DELETE
    THEN EXISTS_TAC "(FF,TT,FF,FF):^Uprog"
    THEN ASM_REWRITE_TAC [ACI_PAR; PAR_Idem]) ;;

%---------------------------------------------------------------------------------
    PAR/g*V = g.a PAR (PAR/(g*V-g.a))             , if a is a member of V
---------------------------------------------------------------------------------%

let PAR_Fold_IMG_DELETE = prove_thm
  (`PAR_Fold_IMG_DELETE`,
   "!(P:*->^Uprog) V a. 
        FINITE V /\ a IN V ==> 
           (Fold $PAR (IMAGE P V) =  
            (P a) PAR (Fold $PAR ((IMAGE P V) DELETE (P a))))",
    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC Fold_IMG_DELETE
    THEN EXISTS_TAC "(FF,TT,FF,FF):^Uprog"
    THEN ASM_REWRITE_TAC [ACI_PAR; PAR_Idem]) ;;

let lemma1 = prove
  ("CHF ((a:*) INSERT V) = (\x. x IN (a INSERT V))",
    CONV_TAC FUN_EQ_CONV THEN BETA_TAC
    THEN REWRITE_TAC [IN_DEF]) ;;

%---------------------------------------------------------------------------------
    defines PAR/V in terms of PROG, INIT, READ, and WRITE of the members of V
---------------------------------------------------------------------------------%

let PAR_Fold = prove_thm
  (`PAR_Fold`,
   "!V:(^Uprog)set.
      FINITE V ==>
        (Fold $PAR V =
             ((?? P::CHF V. PROG P),
              (!! P::CHF V. INIT P),
              (?? P::CHF V. READ P),
              (?? P::CHF V. WRITE P)))",
    (CONV_TAC o RAND_CONV o ABS_CONV o RAND_CONV o MK_ABS_CONV) 
    "V:(^Uprog)set" 
    THEN MATCH_MP_TAC FINITE_INDUCT THEN BETA_TAC THEN REPEAT STRIP_TAC
    THENL
    [ %-- 1 --%
      REWRITE_TAC [CHF_EMPTY; PAR_Fold_EMPTY; UPROG_EXT]
      THEN REWRITE_TAC [PROG; INIT; READ; WRITE]
      THEN REPEAT CONJ_TAC THEN pred_CYAN_TAC ;
      %-- 2 --%
      IMP_RES_TAC PAR_Fold_INSERT THEN ASM_REWRITE_TAC[]
      THEN DEL_ALL_TAC 
      THEN REWRITE_TAC [PAR; UPROG_EXT; PROG; INIT; WRITE; READ; lemma1]
      THEN REWRITE_TAC [IN_INSERT]
      THEN REPEAT CONJ_TAC THEN pred_JADE_TAC 0
      THEN REWRITE_TAC [SYM (SPEC_ALL IN_DEF)]
      THEN CONV_TAC FUN_EQ_CONV THEN BETA_TAC
      THEN GEN_TAC THEN EQ_TAC THEN REPEAT STRIP_TAC
      THEN
        ((EXISTS_TAC "e:^Uprog" THEN ASM_REWRITE_TAC[] THEN NO_TAC)
         ORELSE
         (EXISTS_TAC "i:^Uprog" THEN ASM_REWRITE_TAC[] THEN NO_TAC)
         ORELSE
         (FIRST_ASSUM (SUBST1_TAC o SYM) THEN ASM_REWRITE_TAC[] THEN NO_TAC)
         ORELSE
         (DISJ2_TAC THEN EXISTS_TAC "i:^Uprog" THEN ASM_REWRITE_TAC[] THEN NO_TAC)
         ORELSE ALL_TAC)
      THENL 
      [ RES_TAC ;
        POP_ASSUM MATCH_MP_TAC THEN REWRITE_TAC[] ;
        RES_TAC ]
      ]) ;;

%---------------------------------------------------------------------------------
    If all programs in V are UNITY programs then so is PAR/V
---------------------------------------------------------------------------------%

let UNITY_PAR_Fold = prove_thm
  (`UNITY_PAR_Fold`,
   "!(V:(^Uprog)set).
        FINITE V /\ ~(V={}) /\ (!a. a IN V ==> UNITY a) 
         ==>
         UNITY (Fold $PAR V)",
    REPEAT STRIP_TAC
    THEN UNDISCH_ALL_TAC
    THEN (CONV_TAC o RAND_CONV o MK_ABS_CONV) "V:(^Uprog)set" 
    THEN MATCH_MP_TAC FINITE_INDUCT THEN BETA_TAC THEN REPEAT STRIP_TAC
    THENL
    [ %-- 1 --%
      XRULE_ASSUM_TAC (REWRITE_RULE []) THEN UNDISCH_TAC "F"
      THEN REWRITE_TAC[] ;
      %-- 2 --%
      ASM_CASES_TAC "(s:(^Uprog) set)={}"
      THENL
      [ %-- 2.1 --%
        ASM_REWRITE_TAC [PAR_Fold_SING]
        THEN FIRST_ASSUM MATCH_MP_TAC THEN REWRITE_TAC [IN_INSERT] ;
        %-- 2.2 --%
        XRULE_ASSUM_TAC (QUAFF_RULE 4)
        THEN FIRST_ASSUM (\thm. REWRITE_TAC [MATCH_MP PAR_Fold_INSERT thm])
        THEN MATCH_MP_TAC UNITY_PAR
        THEN CONJ_TAC
        THENL
        [ FIRST_ASSUM MATCH_MP_TAC THEN REWRITE_TAC [IN_INSERT] ;
          FIRST_ASSUM (\thm. if is_forall (concl thm) then NO_TAC
                             else MATCH_MP_TAC thm)
          THEN ASM_REWRITE_TAC[] THEN REPEAT STRIP_TAC
          THEN FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC [IN_INSERT]
        ]
       ]
    ]) ;;

%---------------------------------------------------------------------------------
    If p is confined by WRITE P.a, for some a IN V, then p is also confined by
    WRITE (PAR/P*V)
---------------------------------------------------------------------------------%
        
let CONF_Fold_PAR = prove_thm
  (`CONF_Fold_PAR`,
   "!(P:*->^Uprog) V (a:*) (p:^Pred).
       FINITE V /\ a IN V /\ (WRITE (P a)) CONF p 
       ==> 
      (WRITE (Fold $PAR (IMAGE P V))) CONF p",
   REPEAT STRIP_TAC
   THEN SUBGOAL_THEN "FINITE (IMAGE (P:*->^Uprog) V)" ASSUME_TAC
   THENL [ MATCH_MP_TAC (QUAFF_RULE 3 IMAGE_FINITE) THEN ASM_REWRITE_TAC[] ;
           ALL_TAC ]
   THEN FIRST_ASSUM (\thm. REWRITE_TAC [MATCH_MP PAR_Fold thm])
   THEN REWRITE_TAC [WRITE]
   THEN MATCH_MP_TAC CONF_MONO
   THEN EXISTS_TAC "WRITE ((P:*->^Uprog) a)" THEN ASM_REWRITE_TAC []
   THEN pred_JADE_TAC 0 THEN REWRITE_TAC [SYM (SPEC_ALL IN_DEF)]
   THEN EXISTS_TAC "(P:*->^Uprog) a" THEN ASM_REWRITE_TAC []
   THEN MATCH_MP_TAC (QUAFF_RULE 1 IMAGE_IN)
   THEN ASM_REWRITE_TAC [] ) ;;

let lemma = prove
  ("~((f:*->^Uprog) a = f b) ==> ~(a=b)",
    CONV_TAC CONTRAPOS_CONV
    THEN REWRITE_TAC [prove("~~p=p", TAUT_TAC)]
    THEN STRIP_TAC THEN ASM_REWRITE_TAC[]) ;;

%---------------------------------------------------------------------------------
    If for any distinct members a and b of V we have P.a and P.b are write
    disjunct, then for any member a of V we have:

                P.a is write disjunct with PAR/(P*V-P.a)

---------------------------------------------------------------------------------%

let ioP_PAR_Fold = prove_thm
  (`ioP_PAR_Fold`,
   "!(P:*->^Uprog) V.
      FINITE V /\ (!a b. a IN V /\ b IN V /\ ~(a=b) ==> (P a) ioP (P b))
      ==>
      (!a. a IN V ==> (P a) ioP (Fold $PAR ((IMAGE P V) DELETE (P a))))",
    REPEAT STRIP_TAC
    THEN REPEAT STRIP_TAC
    THEN SUBGOAL_THEN "FINITE ((IMAGE (P:*->^Uprog) V) DELETE (P a))" ASSUME_TAC
    THENL [ REWRITE_TAC [FINITE_DELETE] 
            THEN MATCH_MP_TAC (QUAFF_RULE 3 IMAGE_FINITE) THEN ASM_REWRITE_TAC[] ;
            ALL_TAC ]
    THEN FIRST_ASSUM (\thm. REWRITE_TAC [MATCH_MP PAR_Fold thm])
    THEN UNDISCH_ALL_TAC THEN REWRITE_TAC [ioP; WRITE]
    THEN REWRITE_TAC [RES_qOR] THEN BETA_TAC
    THEN (CONV_TAC o DEPTH_CONV) NOT_EXISTS_CONV
    THEN REWRITE_TAC [SYM (SPEC_ALL IN_DEF)]
    THEN REWRITE_TAC [prove("~(p/\q)=p==>~q",TAUT_TAC)]
    THEN REWRITE_TAC [IN_DELETE; IN_IMAGE]
    THEN REPEAT_FIN 8 STRIP_TAC
    THEN SUBGOAL_THEN "~((a:*)=x)" ASSUME_TAC
    THENL [ ONCE_REWRITE_TAC [ EQ_SYM_EQ]
            THEN MATCH_MP_TAC (GEN_ALL lemma)
            THEN EXISTS_TAC "P:*->^Uprog" 
            THEN FIRST_ASSUM (SUBST1_TAC o SYM)
            THEN ASM_REWRITE_TAC [] ;
            ALL_TAC ]
    THEN XRULE_ASSUM_TAC (QUAFF_RULE 2)
    THEN ASM_REWRITE_TAC[] THEN RES_TAC ) ;;

%---------------------------------------------------------------------------------
    If J is stable in Pr PAR Qr, then is is also stable in Qr
---------------------------------------------------------------------------------%

let PAR_STABLE = prove_thm
  (`PAR_STABLE`,
   "!(Pr:^Uprog) Qr p. STABLE (Pr PAR Qr) p ==> STABLE Qr p",
    REWRITE_TAC [STABLE; UNLESS; PAR; PROG; pOR_DEF] 
    THEN RESTRICT_DEF_TAC THEN BETA_TAC
    THEN REPEAT STRIP_TAC THEN RES_TAC) ;;

%---------------------------------------------------------------------------------
   the stability over PAR/V can be implemented by enforcing the stabolity at
   each member of V.    
---------------------------------------------------------------------------------%

let STABLE_PAR_Fold = prove_thm
  (`STABLE_PAR_Fold`,
   "!(V:(^Uprog)set) p. 
       FINITE V /\ (!a. a IN V ==> STABLE a p) ==> STABLE (Fold $PAR V) p",
    REWRITE_TAC [STABLE; UNLESS] THEN RESTRICT_DEF_TAC
    THEN REPEAT STRIP_TAC
    THEN UNDISCH_TAC "PROG (Fold $PAR (V:(^Uprog)set)) x"
    THEN FIRST_ASSUM (\thm. REWRITE_TAC [MATCH_MP PAR_Fold thm])
    THEN REWRITE_TAC [PROG; RES_qOR; SYM (SPEC_ALL IN_DEF)]
    THEN BETA_TAC THEN STRIP_TAC
    THEN RES_TAC) ;;

%---------------------------------------------------------------------------------
    This is the generalized REACH Transparency Law for Fold PAR
---------------------------------------------------------------------------------%

let REACH_PAR_Fold_ioP = prove_thm
  (`REACH_PAR_Fold_ioP`,
   "!(P:*->^Uprog) V a J p q.
         FINITE V /\ a IN V /\ 
         STABLE (Fold $PAR (IMAGE P V)) J /\
         (!a. a IN V ==> UNITY (P a)) /\
         (!a b. a IN V /\ b IN V /\ ~(a=b) ==> (P a) ioP (P b)) /\
         REACH (P a) J p q
         ==>
         REACH (Fold $PAR (IMAGE P V)) J p q",
     REPEAT STRIP_TAC
     THEN IMP_RES_TAC PAR_Fold_IMG_DELETE
     THEN ASM_REWRITE_TAC[]
     THEN ASM_CASES_TAC "((IMAGE P V) DELETE ((P:*->^Uprog) a)) = {}"
     THENL
     [ %-- 1 --%
       ASM_REWRITE_TAC [ PAR_Fold_EMPTY; PAR_UNIT] ;
       %-- 2 --%
       MATCH_MP_TAC REACH_ioP
       THEN ASM_REWRITE_TAC[] THEN REPEAT CONJ_TAC
       THENL
       [ %-- 2.1: finiteness --%
         MATCH_MP_TAC UNITY_PAR_Fold 
         THEN ASM_REWRITE_TAC [IN_DELETE; IN_IMAGE; FINITE_DELETE]
         THEN REPEAT STRIP_TAC
         THENL [ MATCH_MP_TAC IMAGE_FINITE THEN ASM_REWRITE_TAC[] ;
                 ASM_REWRITE_TAC[] THEN RES_TAC ] ;
         %-- 2.2: ioP --%
         MATCH_MP_TAC (QUAFF_RULE 1 ioP_PAR_Fold) 
         THEN ASM_REWRITE_TAC[] ;
         %-- 2.3: stability --%
         MATCH_MP_TAC PAR_STABLE
         THEN EXISTS_TAC "(P:*->^Uprog) a"
         THEN FIRST_ASSUM 
              (\thm. SUBST1_TAC (SYM (SPEC "(P:*->^Uprog)" thm))
                     THEN DEL_TAC (concl thm))
         THEN ASM_REWRITE_TAC[]
        ]
     ]) ;;

%---------------------------------------------------------------------------------
    This is the generalized CON Transparency Law for Fold PAR
---------------------------------------------------------------------------------%


let CON_PAR_Fold_ioP = prove_thm
  (`CON_PAR_Fold_ioP`,
   "!(P:*->^Uprog) V a J p q.
         FINITE V /\ a IN V /\ 
         STABLE (Fold $PAR (IMAGE P V)) J /\
         (!a. a IN V ==> UNITY (P a)) /\
         (!a b. a IN V /\ b IN V /\ ~(a=b) ==> (P a) ioP (P b)) /\
         CON (P a) J p q
         ==>
         CON (Fold $PAR (IMAGE P V)) J p q",
     REPEAT STRIP_TAC
     THEN IMP_RES_TAC PAR_Fold_IMG_DELETE
     THEN ASM_REWRITE_TAC[]
     THEN ASM_CASES_TAC "((IMAGE P V) DELETE ((P:*->^Uprog) a)) = {}"
     THENL
     [ %-- 1 --%
       ASM_REWRITE_TAC [ PAR_Fold_EMPTY; PAR_UNIT] ;
       %-- 2 --%
       MATCH_MP_TAC CON_ioP
       THEN ASM_REWRITE_TAC[] THEN REPEAT CONJ_TAC
       THENL
       [ %-- 2.1: finiteness --%
         MATCH_MP_TAC UNITY_PAR_Fold 
         THEN ASM_REWRITE_TAC [IN_DELETE; IN_IMAGE; FINITE_DELETE]
         THEN REPEAT STRIP_TAC
         THENL [ MATCH_MP_TAC IMAGE_FINITE THEN ASM_REWRITE_TAC[] ;
                 ASM_REWRITE_TAC[] THEN RES_TAC ] ;
         %-- 2.2: ioP --%
         MATCH_MP_TAC (QUAFF_RULE 1 ioP_PAR_Fold) 
         THEN ASM_REWRITE_TAC[] ;
         %-- 2.3: stability --%
         MATCH_MP_TAC PAR_STABLE
         THEN EXISTS_TAC "(P:*->^Uprog) a"
         THEN FIRST_ASSUM 
              (\thm. SUBST1_TAC (SYM (SPEC "(P:*->^Uprog)" thm))
                     THEN DEL_TAC (concl thm))
         THEN ASM_REWRITE_TAC[]
        ]
      ]) ;;

%-----------------------------------------------------------------------------

  Let V be a set of nodes and E describes the connectivity among the nodes.
  Let A be a set over (type) *A describing the rounds. Let R be a relation 
  over rounds and (A,R) is intended to be well-founded.

  To each node a we associate a program P a. In each node a, we have a variable 
  x a in which we want to store the result of the calculation of  P a. For each
  neighbor b of a, that is each b such that b IN V and E a b, we have a link
  register r a b to which P a can copy the result in x a. This can subsequently 
  read by P b.

  Let OK be a predicate defining our criterion for the correct result of node
  a for round n. 

  Now we can define:

     dataOK (V,E) ok n x : the values of all x a with a IN V are ok for 
                           round n

     comOK (V,E) ok n r : the values of all copies r a b are ok for round n

     preOK (V,E) (A,R) ok n x r : the values of the x's and their copies are
                                  OK for all rounds that preceed (wrt R) 
                                  round n.

------------------------------------------------------------------------------%

let VarVec = ":*Vert->*var" ;;
let OK_Type = ":*A->*Vert->^VarVec->^Pred" ;;
let CRel = ":*Vert->*Vert->bool" ;;
let WRel = ":*A->*A->bool" ;;

let dataOK = new_definition
  (`dataOK`,
   "dataOK (V,E:^CRel) (OK:^OK_Type) n x =  (\s. !a. a IN V ==> OK n a x s)") ;;

let comOK = new_definition
  (`comOK`,
   "comOK (V,E:^CRel) (OK:^OK_Type) n r =
      (\s. !a b. a IN V /\ b IN V /\ E a b ==> OK n a (r b) s)") ;;

let preOK = new_definition
  (`preOK`,
   "preOK (V,E) (A,R:^WRel) (OK:^OK_Type) n x r = 
          (!!m::(\m. m IN A /\ R m n). 
                   (dataOK (V,E) OK m x) AND (comOK (V,E) OK m r))") ;;

%------------------------------------------------------------------------------
   We can express preOK n in terms of preOK, dataOK, and comOK at rounds 
   preceeding n.
------------------------------------------------------------------------------%

let preOK_pre = prove_thm
  (`preOK_pre`,
   "TRANS R ==>
      (preOK (V,E) (A,R) (OK:^OK_Type) n x r =
       (!!m::(\m. m IN A /\ R m n).
               (preOK (V,E) (A,R) OK m x r) AND 
               (dataOK (V,E) OK m x) AND (comOK (V,E) OK m r)))",
    REWRITE_TAC [preOK; dataOK; comOK; RES_qAND; pAND_DEF; TRANS_DEF]
    THEN (CONV_TAC o DEPTH_CONV) FUN_EQ_CONV THEN BETA_TAC
    THEN REPEAT STRIP_TAC THEN EQ_TAC
    THEN REPEAT STRIP_TAC
    THEN REPEAT RES_TAC) ;;

%------------------------------------------------------------------------------
   We can converge to DataOK by converging to the OK of each x a separately
------------------------------------------------------------------------------%

let lemma = prove
   ("(!!i:*A::W. TT) = (TT:^Pred)", pred_CYAN_TAC) ;;

let CON_CONJ_RIGHT = 
     GEN_ALL (REWRITE_RULE [lemma] (BETA_RULE
    (ISPECL ["W:*A->bool"; "Pr:^Uprog"; "J:^Pred"; "\i:*A. (TT:^Pred)"]
     CON_CONJ))) ;;

let CHF_FIN_FIN = prove
   ("CHF_FINITE (CHF V) = FINITE (V:(*)set)",
     REWRITE_TAC [CHF_FINITE; set_ISO_DEF]) ;;

let qAND_shift_dataOK  = prove
  ("(\s. !a. a IN V ==>  (OK:^OK_Type) n a x s) = (!!a::\a. a IN V. OK n a x)",
    pred_CYAN_TAC) ;;

let dataOK_split = prove_thm
  (`dataOK_split`,
   "UNITY (Pr:^Uprog) /\ STABLE Pr J /\
      FINITE V /\
      (!a. a IN V ==> CON Pr J TT ((OK:^OK_Type) n a x))
      ==>
      CON Pr J TT (dataOK (V,E) OK n x)",
    REWRITE_TAC [dataOK; qAND_shift_dataOK] THEN REPEAT STRIP_TAC
    THEN (CONV_TAC o RAND_CONV o RAND_CONV o ABS_CONV o MK_ABS_CONV) "a:*Vert" 
    THEN MATCH_MP_TAC CON_CONJ_RIGHT
    THEN RESTRICT_DEF_TAC
    THEN ASM_REWRITE_TAC[IN_DEF]
    THEN (CONV_TAC o DEPTH_CONV) ETA_CONV
    THEN ASM_REWRITE_TAC [CHF_FIN_FIN]) ;;

%------------------------------------------------------------------------------
   We can converge to comOK by converging to the OK of each link register 
   separately
------------------------------------------------------------------------------%

let qAND_shift_comOK = prove
  ("(\s. !a b. a IN V /\ b IN V /\ E a b ==> (OK:^OK_Type) n a (r b) s)
     =
    (!!a::\a. a IN V. (!!b::(\b. b IN V /\ E a b). OK n a (r b)))",
    REWRITE_TAC [RES_qAND]
    THEN CONV_TAC FUN_EQ_CONV  
    THEN BETA_TAC THEN GEN_TAC THEN EQ_TAC 
    THEN REPEAT STRIP_TAC
    THEN RES_TAC) ;;
 
let lemma2 = prove
  ("b IN V /\ (E:^CRel) x b = b IN (V INTER {c | E x c})",
    REWRITE_TAC [IN_INTER]
    THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
    THEN REFL_TAC) ;;

let comOK_split = prove_thm
  (`comOK_split`,
   "UNITY (Pr:^Uprog) /\ STABLE Pr J /\
      FINITE V /\
      (!a b. a IN V /\ b IN V /\ E a b ==> CON Pr J TT ((OK:^OK_Type) n a (r b)))
      ==>
      CON Pr J TT (comOK (V,E) OK n r)",
    REWRITE_TAC [comOK; qAND_shift_comOK] THEN REPEAT STRIP_TAC
    THEN (CONV_TAC o RAND_CONV o RAND_CONV o ABS_CONV o MK_ABS_CONV)  "a:*Vert" 
    THEN MATCH_MP_TAC CON_CONJ_RIGHT
    THEN RESTRICT_DEF_TAC
    THEN ASM_REWRITE_TAC [IN_DEF]
    THEN (CONV_TAC o DEPTH_CONV) ETA_CONV
    THEN ASM_REWRITE_TAC [CHF_FIN_FIN]
    THEN REWRITE_TAC [SYM (SPEC_ALL IN_DEF)]
    THEN REPEAT STRIP_TAC
    THEN (CONV_TAC o RAND_CONV o RAND_CONV o ABS_CONV o MK_ABS_CONV) 
         "b:*Vert" 
    THEN MATCH_MP_TAC CON_CONJ_RIGHT
    THEN RESTRICT_DEF_TAC 
    THEN ASM_REWRITE_TAC [] THEN REPEAT STRIP_TAC
    THENL
    [ REWRITE_TAC [lemma2; IN_DEF]
      THEN (CONV_TAC o RAND_CONV) ETA_CONV
      THEN IMP_RES_TAC INTER_FINITE
      THEN ASM_REWRITE_TAC [CHF_FIN_FIN] ;
      RES_TAC ] ) ;;
         
%---------------------------------------------------------------------------------
   Let P a be the program associated with process a. Define:

       Net P V : the parallel composition of all P a

       Proper P V : all processes in the net are UNITY programs

       Disjunct P V : every process in the net is write-disjunct with another

       ConfOk : for all round n and node a, the satisfaction predicate
                OK n a x is confined by (the write variables of) P.a.

       ConfOkReg : for all round n, node a and its neighbor b,
                   OK n a (r b) is confined by P.a.

       NetConfOk : for all round n and node a, OK n a x is confined by the net

       NetConfOkReg : for all round n and node a and its neighbor b, 
                      OK n a (r b) is confined by the net
      
---------------------------------------------------------------------------------%

let Net_DEF = new_definition
  (`Net_DEF`,
   "Net (P:*Vert->^Uprog) V = Fold $PAR (IMAGE P V)") ;;

let Proper_DEF = new_definition
  (`Proper_DEF`,
   "Proper (P:*Vert->^Uprog) V = !a. a IN V ==> UNITY (P a)") ;;

let Disjunct_DEF = new_definition
  (`Disjunct_DEF`,
   "Disjunct (P:*Vert->^Uprog) V = 
       !a b. a IN V /\ b IN V /\ ~(a=b) ==> (P a) ioP (P b)") ;;

let ConfOk = new_definition
  (`ConfOk`,
   "ConfOk (V,E:^CRel) (P:*Vert->^Uprog) (OK:^OK_Type) x = 
           (!n a. a IN V ==> (WRITE (P a)) CONF (OK n a x))") ;;

let NetConfOk = new_definition
  (`NetConfOk`,
   "NetConfOk (V,E:^CRel) (Pr:^Uprog) (OK:^OK_Type) x = 
            (!n a. a IN V ==> (WRITE Pr ) CONF (OK n a x))") ;;

let RegConfOk = new_definition
  (`RegConfOk`,
   "RegConfOk (V,E:^CRel) (P:*Vert->^Uprog) (OK:^OK_Type) r = 
        (!n a b. a IN V /\ b IN V /\ E a b ==> 
                 (WRITE (P a)) CONF (OK n a (r b)))") ;;

let NetRegConfOk = new_definition
  (`NetRegConfOk`,
   "NetRegConfOk (V,E:^CRel) (Pr:^Uprog) (OK:^OK_Type) r = 
        (!n a b. a IN V /\ b IN V /\ E a b ==> 
                 (WRITE Pr) CONF (OK n a (r b)))") ;;

%---------------------------------------------------------------------------------
   ConfOk (V,E) P implies NetConfOk (V,E) (Net P V)
---------------------------------------------------------------------------------%

let ConfOk_Net = prove_thm
  (`ConfOk_Net`,
   "FINITE V /\ 
    ConfOk (V,E) (P:*Vert->^Uprog) (OK:^OK_Type) x
    ==>
    NetConfOk (V,E) (Net P V) OK x",
    REWRITE_TAC [ConfOk; NetConfOk; Net_DEF]
    THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC CONF_Fold_PAR 
    THEN EXISTS_TAC "a:*Vert" 
    THEN RES_TAC THEN ASM_REWRITE_TAC[]) ;;

%---------------------------------------------------------------------------------
   RegConfOk (V,E) P implies NetRegConfOk (V,E) (Net P V)
---------------------------------------------------------------------------------%

let RegConfOk_Net = prove_thm
  (`RegConfOk_Net`,
   "FINITE V /\ 
    RegConfOk (V,E) (P:*Vert->^Uprog) (OK:^OK_Type) r
    ==>
    NetRegConfOk (V,E) (Net P V) OK r",
    REWRITE_TAC [RegConfOk; NetRegConfOk; Net_DEF]
    THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC CONF_Fold_PAR 
    THEN EXISTS_TAC "a:*Vert" 
    THEN RES_TAC THEN ASM_REWRITE_TAC[]) ;;

%---------------------------------------------------------------------------------
   NetConfOk Pr implies dataOK is confined by Pr
---------------------------------------------------------------------------------%

let dataOK_CONF = prove_thm
  (`dataOK_CONF`,
   "NetConfOk (V,E) (Pr:^Uprog) (OK:^OK_Type) x
    ==>
    (WRITE Pr) CONF (dataOK (V,E) OK n x)",
    REWRITE_TAC [NetConfOk; dataOK; qAND_shift_dataOK] THEN REPEAT STRIP_TAC
    THEN (CONV_TAC o RAND_CONV o RAND_CONV o ABS_CONV o MK_ABS_CONV)  
         "a:*Vert" 
    THEN CONF_TAC 1 THEN RESTRICT_DEF_TAC
    THEN REPEAT STRIP_TAC THEN RES_TAC
    THEN ASM_REWRITE_TAC []) ;;

%---------------------------------------------------------------------------------
   NetRegConfOk Pr imples NetOK is confined by Pr
---------------------------------------------------------------------------------%

let comOK_CONF = prove_thm
  (`comOK_CONF`,
   "NetRegConfOk (V,E) (Pr:^Uprog) (OK:^OK_Type) r
    ==>
    (WRITE Pr) CONF (comOK (V,E) OK n r)",
    REWRITE_TAC [NetRegConfOk; comOK; qAND_shift_comOK] THEN REPEAT STRIP_TAC
    THEN (CONV_TAC o RAND_CONV o RAND_CONV o ABS_CONV o MK_ABS_CONV)  
         "a:*Vert" 
    THEN CONF_TAC 1 THEN RESTRICT_DEF_TAC
    THEN REPEAT STRIP_TAC 
    THEN (CONV_TAC o RAND_CONV o RAND_CONV o ABS_CONV o MK_ABS_CONV)  
         "b:*Vert"
    THEN CONF_TAC 1 THEN RESTRICT_DEF_TAC
    THEN REPEAT STRIP_TAC THEN RES_TAC THEN RES_TAC
    THEN ASM_REWRITE_TAC []) ;;

%---------------------------------------------------------------------------------
   NetConfOK Pr and NetRegConfOk Pr imples PreOK is confined by Pr
---------------------------------------------------------------------------------%

let preOK_CONF = prove_thm
  (`preOK_CONF`,
   "NetConfOk (V,E) (Pr:^Uprog) (OK:^OK_Type) x  /\
    NetRegConfOk (V,E) Pr OK r
    ==>
    (WRITE Pr) CONF (preOK (V,E) (A,R) OK n x r)",
    REWRITE_TAC [preOK] THEN REPEAT STRIP_TAC
    THEN (CONV_TAC o RAND_CONV o RAND_CONV o ABS_CONV o MK_ABS_CONV)  
         "m:*A" 
    THEN CONF_TAC 1 THEN RESTRICT_DEF_TAC
    THEN REPEAT STRIP_TAC 
    THEN IMP_RES_TAC (GEN_ALL dataOK_CONF)
    THEN IMP_RES_TAC (GEN_ALL comOK_CONF)
    THEN CONF_TAC 1) ;;

%---------------------------------------------------------------------------------
   The round decomposition theorem for flat networks with link registers:

      A distributed program Pr can converge to dataOK by at any round n
      
        (1) converging to OK x.a for each node a in the network, given preOK n
            holds and stable

        (2) converging to OK r.a.b for each neighboring a and b, given preOK n
            and dataOK n holds and stable 

   Let MDC, MDC1, MDC2, MDS1 be defined as below. MDC is the specification
   of the problem. We will prove that MDC can be implemented with MDC1, MDC2
   and MDS1.
---------------------------------------------------------------------------------%

let MDC_DEF = new_definition
  (`MDC_DEF`,
   "MDC (Pr:^Uprog) J (V,E) (A:(*A)set) (OK:^OK_Type) x =    
         CON Pr J TT (!!n:: CHF A. dataOK (V,E) OK n x)") ;;

let MDC1_DEF = new_definition
  (`MDC1_DEF`,
   "MDC1 (Pr:^Uprog) J (V,E) (A,(R:^WRel)) (OK:^OK_Type) n a x r =             
          CON Pr (J AND (preOK (V,E) (A,R) OK n x r)) TT (OK n a x)") ;;

let MDC2_DEF = new_definition
  (`MDC2_DEF`,
   "MDC2 (Pr:^Uprog) J (V,E) (A,(R:^WRel)) (OK:^OK_Type) n a b x r =    
      CON Pr (J AND (preOK (V,E) (A,R) OK n x r) AND (dataOK (V,E) OK n x))
              TT
              (OK n a (r b))") ;;

let MDS1_DEF = new_definition
  (`MDS1_DEF`,
   "MDS1 (Pr:^Uprog) J (V,E) (A,(R:^WRel)) (OK:^OK_Type) n x r =    
    STABLE Pr (J AND (preOK (V,E) (A,R) OK n x r) AND (dataOK (V,E) OK n x))") ;;

let OK_decom = prove
  ("sWF A R /\ ~(A={}) /\ FINITE A /\
    NetConfOk (V,E) (Pr:^Uprog) (OK:^OK_Type) x  /\
    (!n:: CHF A. CON Pr J 
                     (preOK (V,E) (A,R) OK n x r)
                     ((dataOK (V,E) OK n x) AND (comOK (V,E) OK n r)))
    ==>
    MDC Pr J (V,E) A OK x",
    REWRITE_TAC [MDC_DEF] THEN RESTRICT_DEF_TAC 
    THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC CON_MONO
    THEN EXISTS_TAC 
         "(!! n ::(\n. n IN A). (dataOK(V,E) (OK:^OK_Type) n x) 
                                 AND (comOK (V,E) OK n r))"
    THEN REPEAT CONJ_TAC
    THENL
    [ %-- 1 --%
      (CONV_TAC o RAND_CONV o RAND_CONV o ABS_CONV o MK_ABS_CONV)  "n:*A" 
      THEN CONF_TAC 1 
      THEN BETA_TAC THEN IMP_RES_TAC dataOK_CONF
      THEN RESTRICT_DEF_TAC THEN ASM_REWRITE_TAC[] ;
      %-- 2 --%
      (CONV_TAC o RAND_CONV o RAND_CONV o ABS_CONV o MK_ABS_CONV)  "n:*A"
      THEN MATCH_MP_TAC CON_BY_sWF      
      THEN EXISTS_TAC "R:^WRel" THEN BETA_TAC
      THEN ASM_REWRITE_TAC[] 
      THEN REWRITE_TAC [SYM (SPEC_ALL preOK)]
      THEN RESTRICT_DEF_TAC THEN ASM_REWRITE_TAC [IN_DEF] ;
      %-- 3 --%
      DEL_ALL_TAC THEN REWRITE_TAC [IN_DEF] THEN pred_JADE_TAC 3
    ]) ;;

let OK_Round_Decom = prove_thm
  (`OK_Round_Decom`,
   "sWF A R /\ ~(A={}) /\ ~(V={}) /\ FINITE A /\ FINITE V /\
    NetConfOk (V,E) (Pr:^Uprog) (OK:^OK_Type) x  /\
    NetRegConfOk (V,E) Pr OK r /\
    STABLE Pr J /\
    (!n. n IN A ==> MDS1 Pr J (V,E) (A,R) OK n x r) /\
    (!n a. n IN A /\ a IN V ==> MDC1 Pr J (V,E) (A,R) OK n a x r) /\
    (!n a b. n IN A /\ a IN V /\ b IN V /\ E a b ==>
                  MDC2 Pr J (V,E) (A,R) OK n a b x r)
    ==>
    MDC Pr J (V,E) A OK x",
    REWRITE_TAC [MDS1_DEF; MDC1_DEF; MDC2_DEF]
    THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC OK_decom
    THEN RESTRICT_DEF_TAC THEN ASM_REWRITE_TAC [SYM (SPEC_ALL IN_DEF)]
    THEN REPEAT STRIP_TAC
    THEN (CONV_TAC o RATOR_CONV o RAND_CONV o ONCE_REWRITE_CONV) 
         [ (GEN_ALL o SYM o SPEC_ALL) pAND_UNIT]
    THEN MATCH_MP_TAC CON_STABLE_SHIFT
    THEN ASM_REWRITE_TAC[] THEN CONJ_TAC
    THENL [IMP_RES_TAC preOK_CONF THEN ASM_REWRITE_TAC[]; ALL_TAC ]
    THEN MATCH_MP_TAC CON_SPIRAL
    THEN SUBGOAL_THEN "UNITY (Pr:^Uprog)" ASSUME_TAC
    THENL [ UNDISCH_ALL_TAC 
            THEN REWRITE_TAC [SYM(SPEC_ALL MEMBER_NOT_EMPTY)]
            THEN REPEAT STRIP_TAC THEN RES_TAC
            THEN IMP_RES_TAC CON_IMP_UNITY ;
            ALL_TAC ]
    THEN CONJ_TAC
    THENL
    [ %-- 1 --%
      MATCH_MP_TAC dataOK_split THEN ASM_REWRITE_TAC[]
      THEN CONJ_TAC 
      THENL
      [ %-- 1.1 --%
        UNDISCH_ALL_TAC 
        THEN REWRITE_TAC [SYM(SPEC_ALL MEMBER_NOT_EMPTY)]
        THEN REPEAT STRIP_TAC THEN RES_TAC
        THEN IMP_RES_TAC CON_IMP_STABLE ;
        %-- 1.2 --%
        REPEAT STRIP_TAC THEN RES_TAC ] ;
      %-- 2 --%
      (CONV_TAC o RATOR_CONV o RAND_CONV o ONCE_REWRITE_CONV) 
         [ (GEN_ALL o SYM o SPEC_ALL) pAND_UNIT]
      THEN MATCH_MP_TAC CON_STABLE_SHIFT  
      THEN REPEAT CONJ_TAC   
      THENL
      [ %-- 2.1 --%
        IMP_RES_TAC dataOK_CONF THEN ASM_REWRITE_TAC[] ;
        %-- 2.2 --%
        MATCH_MP_TAC comOK_split THEN ASM_REWRITE_TAC[pAND_ASSOC]
        THEN REPEAT STRIP_TAC THEN RES_TAC ;
        %-- 2.3 --%
        UNDISCH_ALL_TAC 
        THEN REWRITE_TAC [SYM(SPEC_ALL MEMBER_NOT_EMPTY)]
        THEN REPEAT STRIP_TAC THEN RES_TAC
        THEN IMP_RES_TAC CON_IMP_STABLE 
      ]
   ]) ;;

%---------------------------------------------------------------------------------
   We continue by refining MDC1 and MDC2. The convergence to OK n a x in MDC1
   can be delegated to P.a alone. In the same way the convergence to OK n a (r b) 
   in MDC2 can be delegated to P.a.

   Let us introduce MDC3 and MDS2 as below. We will show that MDC1 can be
   implemented by these two.
---------------------------------------------------------------------------------%

let MDC3_DEF = new_definition
  (`MDC3_DEF`,
   "MDC3 (P:*Vert->^Uprog) J (V,E) (A,(R:^WRel)) (OK:^OK_Type) n a x r =             
    CON (P a) (J AND (preOK (V,E) (A,R) OK n x r)) TT (OK n a x)") ;;

let MDS2_DEF = new_definition
  (`MDS2_DEF`,
   "MDS2 (Pr:^Uprog) J (V,E) (A,(R:^WRel)) (OK:^OK_Type) n x r =  
    STABLE Pr (J AND (preOK (V,E) (A,R) OK n x r))") ;;

let MDC1_Decom = prove_thm
  (`MDC1_Decom`,
   "FINITE V /\ a IN V /\
    Proper (P:*Vert->^Uprog) V /\ Disjunct P V /\ 
    MDS2 (Net P V) J (V,E) (A,R:^WRel) (OK:^OK_Type) n x r /\
    MDC3 P J (V,E) (A,R) OK n a x r 
    ==>
    MDC1 (Net P V) J (V,E) (A,R) OK n a x r",
    REWRITE_TAC [MDC1_DEF; MDC3_DEF; MDS2_DEF; Proper_DEF; Disjunct_DEF; Net_DEF]
    THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC CON_PAR_Fold_ioP
    THEN EXISTS_TAC "a:*Vert" THEN ASM_REWRITE_TAC[]) ;;

%---------------------------------------------------------------------------------
   As said, the convergence to OK n a (r b) in MDC2 can be delegated to P.a. 
   
   Let us introduce MDC4 as below. We will show that MDC2 can be implemented 
   by MDC4 and MDS1.
---------------------------------------------------------------------------------%

let MDC4_DEF = new_definition
  (`MDC4_DEF`,
   "MDC4 (P:*Vert->^Uprog) J (V,E) (A,(R:^WRel)) (OK:^OK_Type) n a b x r =    
      CON (P a) (J AND (preOK (V,E) (A,R) OK n x r) AND (dataOK (V,E) OK n x))
                 TT
                 (OK n a (r b))") ;;


let MDC2_Decom = prove_thm
  (`MDC2_Decom`,
   "FINITE V /\ a IN V /\
    Proper (P:*Vert->^Uprog) V /\ Disjunct P V /\ 
    MDS1 (Net P V) J (V,E) (A,R:^WRel) (OK:^OK_Type) n x r /\
    MDC4 P J (V,E) (A,R) OK n a b x r 
    ==>
    MDC2 (Net P V) J (V,E) (A,R) OK n a b x r",
    REWRITE_TAC [MDC2_DEF; MDC4_DEF; MDS1_DEF; Proper_DEF; Disjunct_DEF; Net_DEF]
    THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC CON_PAR_Fold_ioP
    THEN EXISTS_TAC "a:*Vert" THEN ASM_REWRITE_TAC[]) ;;

%---------------------------------------------------------------------------------
    The Net stability of preOK and dataOK in MDS1 can be implemented by 
    enforcing their stability at node level.

    We will refine MDS1 by MDS3 defined below.
---------------------------------------------------------------------------------%

let STABLE_PAR_Fold_IMG = prove
  ("!(P:*->^Uprog) V p. 
       FINITE V /\ (!a. a IN V ==> STABLE (P a) p) 
       ==> 
       STABLE (Fold $PAR (IMAGE P V)) p",
    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC STABLE_PAR_Fold 
    THEN REWRITE_TAC [IN_IMAGE] THEN REPEAT STRIP_TAC
    THENL [ MATCH_MP_TAC IMAGE_FINITE THEN ASM_REWRITE_TAC[] ;
            RES_TAC THEN ASM_REWRITE_TAC[] ]) ;;

let MDS3_DEF = new_definition
  (`MDS3_DEF`,
   "MDS3 (P:*Vert->^Uprog) J (V,E) (A,(R:^WRel)) (OK:^OK_Type) n a x r =  
    STABLE (P a) 
           (J AND ((preOK (V,E) (A,R) OK n x r) AND (dataOK(V,E)OK n x)))") ;;

let MDS1_Decom = prove_thm
  (`MDS1_Decom`,
   "FINITE V /\
    (!a. a IN V ==> 
         MDS3 (P:*Vert->^Uprog) J (V,E) (A,(R:^WRel)) (OK:^OK_Type) n a x r)
    ==>
    MDS1 (Net P V) J (V,E) (A,R) OK n x r",
    REWRITE_TAC [MDS3_DEF; MDS1_DEF; Net_DEF] 
    THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC STABLE_PAR_Fold_IMG 
    THEN ASM_REWRITE_TAC[]) ;;

%---------------------------------------------------------------------------------
    In same way, the Net stability of preOK in MDS2 can be implemented by 
    enforcing its stability at node level.

    We will refine MDS2 by MDS4 defined below.
---------------------------------------------------------------------------------%

let MDS4_DEF = new_definition
  (`MDS4_DEF`,
   "MDS4 (P:*Vert->^Uprog) J (V,E) (A,(R:^WRel)) (OK:^OK_Type) n a x r =  
    STABLE (P a) (J AND ((preOK (V,E) (A,R) OK n x r)))") ;;

let MDS2_Decom = prove_thm
  (`MDS2_Decom`,
   "FINITE V /\
    (!a. a IN V ==> 
         MDS4 (P:*Vert->^Uprog) J (V,E) (A,(R:^WRel)) (OK:^OK_Type) n a x r)
    ==>
    MDS2 (Net P V) J (V,E) (A,R) OK n x r",
    REWRITE_TAC [MDS4_DEF; MDS2_DEF; Net_DEF] 
    THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC STABLE_PAR_Fold_IMG 
    THEN ASM_REWRITE_TAC[]) ;;

%---------------------------------------------------------------------------------
   The node level stability of preOK and dataOK in MDS3 can be implemented by 
   enforcing the node level stability of preOK (MDS4) and the node level stability 
   of preOK and OK x simultaneously.

   We will show that MDS3 can be implemented by MDS4 and MDS5 defined below.
---------------------------------------------------------------------------------%

let MDS5_DEF = new_definition
  (`MDS5_DEF`,
   "MDS5 (P:*Vert->^Uprog) J (V,E) (A,R:^WRel) (OK:^OK_Type) n a x r  =
      STABLE (P a) (J AND (preOK (V,E) (A,R) OK n x r) AND (OK n a x))") ;;

let MDS3_Decom_lemma1 = prove
  ("Disjunct (P:*Vert->^Uprog) V /\ Proper P V /\ 
    a IN V /\ b IN V /\ ~(a=b) /\
    ConfOk (V,E) P (OK:^OK_Type) x   
    ==>
    STABLE (P a) (OK n b x)",
    REWRITE_TAC [Disjunct_DEF; Proper_DEF; ConfOk]
    THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC CONF_Local
    THEN ONCE_REWRITE_TAC [ioP_SYM]
    THEN EXISTS_TAC "(P:*Vert->^Uprog)  b" THEN REPEAT CONJ_TAC
    THEN RES_TAC THEN ASM_REWRITE_TAC[]) ;;


let MDS3_Decom = prove_thm
  (`MDS3_Decom`,
   "Disjunct (P:*Vert->^Uprog) V /\ Proper P V /\ 
    ConfOk (V,E) P (OK:^OK_Type) x /\
    n IN A /\ a IN V /\
    MDS4 P J (V,E) (A,R) OK n a x r /\
    (!m b. m IN A /\ b IN V  ==> MDS5 P J (V,E) (A,R) OK m b x r)
    ==>
    MDS3 P J (V,E) (A,R) OK n a x r",
    REWRITE_TAC [MDS5_DEF; MDS4_DEF; MDS3_DEF; dataOK; qAND_shift_dataOK; 
                 SYM (SPEC_ALL pAND_ASSOC)]
    THEN REPEAT STRIP_TAC 
    THEN SUBGOAL_THEN "~((\a:*Vert. a IN V)=FF)"  ASSUME_TAC
    THENL [ REWRITE_TAC [Non_FF_lemma] THEN BETA_TAC
            THEN EXISTS_TAC "a:*Vert" THEN ASM_REWRITE_TAC[] ;
            ALL_TAC ]
    THEN (CONV_TAC o RAND_CONV o RAND_CONV o RAND_CONV o 
          ABS_CONV o MK_ABS_CONV) "a:*Vert"
    THEN FIRST_ASSUM 
         (\thm. REWRITE_TAC [MATCH_MP qAND_LEFT_DISTR_AND thm] 
                THEN DEL_TAC (concl thm))
    THEN (CONV_TAC o RAND_CONV o RAND_CONV o ABS_CONV o MK_ABS_CONV) "i:*Vert"
    THEN MATCH_MP_TAC STABLE_GEN_CONJ
    THEN RESTRICT_DEF_TAC 
    THEN REPEAT STRIP_TAC 
    THEN ASM_CASES_TAC "a=(x':*Vert)"
    THENL
    [ %-- 1 --%
      RES_TAC THEN ASM_REWRITE_TAC[] ;
      %-- 2 --%
      MATCH_MP_TAC STABLE_SIMPLE_CONJ THEN CONJ_TAC
      THENL
      [ ASM_REWRITE_TAC[] ;
        MATCH_MP_TAC MDS3_Decom_lemma1 THEN ASM_REWRITE_TAC[]
      ]
   ]) ;;

%---------------------------------------------------------------------------------
  Now we continue with the refinement of the convergence part (MDC3 and 4).

  The convergence in MDC3 can be reduced to an ENSURES property. In addition,
  it also implies several stability requirements, but it turns out that these
  follows from MDS4 and 5.

  Let MDE1 defines the following ENSURES property. We will show that MDC3 can
  be implemedted by MDS4, MDS5, and MDE1.
---------------------------------------------------------------------------------%

let MDE1_DEF = new_definition
  (`MDE1_DEF`,
   "MDE1 (P:*Vert->^Uprog) J (V,E) (A,R) (OK:^OK_Type) n a x r  =
    ENSURES (P a) (J AND (preOK (V,E) (A,R) OK n x r)) (OK n a x)") ;;


let MDC3_Decom = prove_thm
  (`MDC3_Decom`,
   "ConfOk (V,E) (P:*Vert->^Uprog) (OK:^OK_Type) x /\ a IN V /\
    MDS4 P J (V,E) (A,R) OK n a x r /\
    MDS5 P J (V,E) (A,R) OK n a x r /\
    MDE1 P J (V,E) (A,R) OK n a x r
    ==>
    MDC3 P J (V,E) (A,R) OK n a x r",
    REWRITE_TAC [ConfOk; MDE1_DEF; MDS4_DEF; MDS5_DEF; MDC3_DEF]
    THEN REPEAT STRIP_TAC 
    THEN MATCH_MP_TAC CON_ENSURES_LIFT
    THEN RES_TAC THEN ASM_REWRITE_TAC [CONF_TT; pAND_UNIT]
    THEN ONCE_REWRITE_TAC 
         [prove("(p:^Pred) AND (q AND r) = q AND (r AND p)", pred_CYAN_TAC)]
    THEN ASM_REWRITE_TAC []) ;;

%---------------------------------------------------------------------------------
  The convergence in MDC4 can be reduced to an ENSURES property and the
  corresponding stability requirements.

  Let MDE2 defines the following ENSURES property and MDS6 the following STABLE
  property.  We will show that MDC4 can be implemedted by MDS3, MDS6, and MDE2.

---------------------------------------------------------------------------------%

let MDS6_DEF = new_definition
  (`MDS6_DEF`,
   "MDS6 (P:*Vert->^Uprog) J (V,E) (A,R:^WRel) (OK:^OK_Type) n a b x r  =
    STABLE (P a)
           (J AND (preOK (V,E) (A,R) OK n x r) AND (dataOK (V,E) OK n x)
                   AND (OK n a (r b)))") ;;

let MDE2_DEF = new_definition
  (`MDE2_DEF`,
   "MDE2 (P:*Vert->^Uprog) J (V,E) (A,R) (OK:^OK_Type) n a b x r  =
    ENSURES (P a) (J AND (preOK (V,E) (A,R) OK n x r) AND (dataOK (V,E) OK n x))
                  (OK n a (r b))") ;;

let MDC4_Decom = prove_thm
  (`MDC4_Decom`,
   "RegConfOk (V,E) (P:*Vert->^Uprog) (OK:^OK_Type) r /\ 
    a IN V /\ b IN V /\ E a b /\
    MDS3 P J (V,E) (A,R) OK n a x r /\
    MDS6 P J (V,E) (A,R) OK n a b x r  /\
    MDE2 P J (V,E) (A,R) OK n a b x r 
    ==>
    MDC4 P J (V,E) (A,R) OK n a b x r",
    REWRITE_TAC [MDE2_DEF; MDS6_DEF; RegConfOk; MDS3_DEF; MDC4_DEF]
    THEN REPEAT STRIP_TAC 
    THEN MATCH_MP_TAC CON_ENSURES_LIFT
    THEN RES_TAC THEN ASM_REWRITE_TAC [CONF_TT; pAND_UNIT]
    THEN ONCE_REWRITE_TAC 
         [prove("(p:^Pred) AND (q AND (r AND s)) = q AND r AND s AND p", 
          pred_CYAN_TAC)]
    THEN ASM_REWRITE_TAC []) ;;

%---------------------------------------------------------------------------------
  Let ok be a predicate of type *A->*Vert->*value->bool. For a given value v,
  ok n a v means that the value v is an acceptable result for node a at round n.

  We can define OK in terms of ok:

     OK = MPS ok

  where MPS (Make State Predicate) is defined as below. Notice that, for example,
  (MPS ok) n a x is a predicate that only depends on the variable x.a.
---------------------------------------------------------------------------------%

let ok_Type = ":*A->*Vert->*value->bool" ;;

let MSP = new_definition
  (`MSP`,
   "MSP (ok:^ok_Type) n a x (s:^State) = ok n a (s(x a))") ;;

%---------------------------------------------------------------------------------
   if OK is expressed in terms of MSP ok, then the confinement requirements

        ConfOk (V,E) P (MSP ok) x
        RegConfOk (V,E) P (MSP ok) r 

   can be satisfies by insisting that the variable x.a and r.b.a are among write 
   variables of P.a
---------------------------------------------------------------------------------%

let WRITE_ConfOk = prove_thm
  (`WRITE_ConfOk`,
   "(!a. a IN V ==> (WRITE (P a)) (x a)) 
      ==> 
     ConfOk (V,E) (P:*Vert->^Uprog)  (MSP (ok:^ok_Type)) x",
    REWRITE_TAC [ConfOk; MSP; CONF_DEF; Pj_EQ]
    THEN REPEAT STRIP_TAC
    THEN RES_TAC THEN RES_TAC
    THEN ASM_REWRITE_TAC[]) ;;

let WRITE_RegConfOk = prove_thm
  (`WRITE_RegConfOk`,
   "(!a b. a IN V /\ b IN V /\ E a b ==> (WRITE (P a)) (r b a)) 
      ==> 
     RegConfOk (V,E) (P:*Vert->^Uprog)  (MSP (ok:^ok_Type)) r",
    REWRITE_TAC [RegConfOk; MSP; CONF_DEF; Pj_EQ]
    THEN REPEAT STRIP_TAC
    THEN RES_TAC THEN RES_TAC
    THEN ASM_REWRITE_TAC[]) ;;

%---------------------------------------------------------------------------------
  A function Res is called a resolvent for a round-decomposable problem if for
  any function f:*Vert->*value, round n, and node a:

     if given that for any node b, the value of f.b is acceptable for b at any
     round preceeding n, then applying Res.a to f results a value that is
     acceptable for node a at round n.
---------------------------------------------------------------------------------%

let Resolve_DEF = new_definition
  (`Resolve_DEF`,
   "Resolve (V,E:^CRel) (A,R:^WRel) Res (ok:^ok_Type) =
       (!f n a. n IN A /\ a IN V /\
             (!m b. m IN A /\ b IN V /\ R m n /\ E b a ==> ok m b (f b)) 
              ==>
              ok n a (Res a f))") ;;

%---------------------------------------------------------------------------------
   Given OK is expressed as MSP ok, the following theorem tells us how to 
   establish OK for a link register from dataOK.
---------------------------------------------------------------------------------%

let dataOK_regOK = prove_thm
  (`dataOK_regOK`,
   "a IN V ==> |== (((dataOK (V,E) (MSP (ok:^ok_Type)) n x) AND
                    (\s:^State. (s o (r (b:*Vert))) a = (s o x) a))
                    IMP ((MSP ok) n a (r b)))",
    REWRITE_TAC [ dataOK; o_THM]
    THEN pred_JADE_TAC 0 THEN UNDISCH_ALL_TAC
    THEN REWRITE_TAC [MSP] THEN REPEAT STRIP_TAC
    THEN RES_TAC THEN ASM_REWRITE_TAC[]) ;; 

%---------------------------------------------------------------------------------
   Given a resolvent, and that OK is expressed as MSP ok, the following theorem 
   tells us how to establish OK from preOK.
---------------------------------------------------------------------------------%

let Resolve_OK = prove_thm
  (`Resolve_OK`,
   "Resolve (V,E:^CRel) (A,R:^WRel) Res (ok:^ok_Type) /\ n IN A /\ a IN V
    ==>
    |==  (((preOK (V,E) (A,R) (MSP ok) n x r) AND 
           (\s:^State. (s o x) a = Res a (s o (r a))))
          IMP
          ((MSP ok) n a x))",
    REWRITE_TAC [preOK; o_THM; dataOK; comOK;  Resolve_DEF]
    THEN pred_JADE_TAC 0 THEN UNDISCH_ALL_TAC
    THEN REWRITE_TAC [MSP] THEN REPEAT STRIP_TAC
    THEN ASM_REWRITE_TAC[]
    THEN FIRST_ASSUM (\thm. MATCH_MP_TAC thm THEN DEL_TAC (concl thm))
    THEN ASM_REWRITE_TAC[] THEN REPEAT STRIP_TAC
    THEN REWRITE_TAC[o_THM] THEN RES_TAC) ;;

%---------------------------------------------------------------------------------
   Above theorem can be used to simplify the ENSURES requirement in MDS1.
   MDS1 requires tha P.a must ensures that x.a is OK. Using above theorem we know
   that this can be established by computing the function Res.a and applies its
   result to x.a, given that Res is a resolvent function.

   We will show that MDE1 can be implemented by MDS4 and MDE3. The definition of
   MDE3 is given below. Note that MDE3 has the form

        ENSURES Pr TT q 

   This can be implemented by the existence of an action a satisfying

       {TT} a {q}
---------------------------------------------------------------------------------%

let ENSURES_PSP_STABLE = prove
  ("!(Pr:^Uprog) p q J.
       ENSURES Pr p q /\ STABLE Pr J 
       ==> 
       ENSURES Pr (J AND p) (J AND q)",
     let th = prove("(J:^Pred) AND q =
                     (((J AND q) OR (p AND FF)) OR (FF AND q))", pred_CYAN_TAC)
     in
     REPEAT GEN_TAC THEN REWRITE_TAC [STABLE]
     THEN SUBST1_TAC th THEN REPEAT STRIP_TAC
     THEN IMP_RES_TAC ENSURES_PSP) ;;

let ENSURES_PSP_STABLE_TT = prove
  ("!(Pr:^Uprog) q J.
       ENSURES Pr TT q /\ STABLE Pr J 
       ==> 
       ENSURES Pr J (J AND q)",
    let th = prove("(J:^Pred) = J AND TT", pred_CYAN_TAC) in
    REPEAT STRIP_TAC 
    THEN (CONV_TAC o RATOR_CONV o RAND_CONV o ONCE_REWRITE_CONV) [th]
    THEN IMP_RES_TAC ENSURES_PSP_STABLE) ;;

let lemma = prove
  ("|== (((P:^Pred) AND s) IMP q)
    ==>
    |== (((J AND P) AND s) IMP ((J AND P) AND q))",
    pred_JADE_TAC 3) ;;

let Q = 
    "((J:^Pred) AND (preOK (V,E) (A,R) (MSP (ok:^ok_Type)) n x r))
      AND (MSP ok n a x)" ;;

let Q' = 
    "((J:^Pred) AND (preOK (V,E) (A,R) (MSP (ok:^ok_Type)) n x r))
      AND (\s. (s o x) (a:*Vert) = Res a (s o (r a)))" ;;

let MDE3_DEF = new_definition
  (`MDE3_DEF`,
   "MDE3 (P:*Vert->^Uprog) (a:*Vert) Res x r =
    ENSURES (P a) TT (\s. (s o x) a = Res a (s o ((r:*Vert->^VarVec) a)))") ;;

let MDE1_Decom = prove_thm
  (`MDE1_Decom`,
   "Resolve (V,E) (A,R) Res (ok:^ok_Type) /\ 
    n IN A /\ a IN V /\
    MDS4 (P:*Vert->^Uprog) J (V,E) (A,R) (MSP ok) n a x r /\
    MDE3 P a Res x r
    ==>
    MDE1 P J (V,E) (A,R) (MSP ok) n a x r",
    REWRITE_TAC [MDE1_DEF; MDE3_DEF; MDS4_DEF]
    THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC ENSURES_CONSQ_WEAK
    THEN EXISTS_TAC Q THEN CONJ_TAC
    THENL [ ALL_TAC; DEL_ALL_TAC THEN pred_JADE_TAC 1 ]
    THEN MATCH_MP_TAC ENSURES_CONSQ_WEAK
    THEN EXISTS_TAC Q' THEN CONJ_TAC
    THENL     
    [ MATCH_MP_TAC ENSURES_PSP_STABLE_TT THEN ASM_REWRITE_TAC[] ;
      MATCH_MP_TAC (GEN_ALL lemma)
      THEN MATCH_MP_TAC Resolve_OK THEN ASM_REWRITE_TAC[] ]
    ) ;;


%---------------------------------------------------------------------------------
   In the same way we can use theorem dataOK_regOK to simplify the ENSURES 
   requirement in MDS2.
   MDS2 requires tha P.a must ensures that the link register r.b.a is OK. 
   Using the theorem we know that this can be established by cimply copying the
   value of x.a to the link register.

   We will show that MDE2 can be implemented by MDS3 and MDE4. The definition of
   MDE3 is given below. MDS3 can be implemented by MDS4 and 5 and MDE3 has the
   form

        ENSURES Pr TT q 

   which is easy to implement.
---------------------------------------------------------------------------------%

  
let lemma = prove
  ("|== (((D:^Pred) AND s) IMP q)
    ==>
    |== (((J AND (P AND D)) AND s) IMP ((J AND (P AND D)) AND q))",
    pred_JADE_TAC 3) ;;

let Q = 
    "((J:^Pred) AND ((preOK (V,E) (A,R) (MSP (ok:^ok_Type)) n x r) 
                     AND (dataOK (V,E) (MSP ok) n x)))
      AND (MSP ok n a (r b))" ;;

let Q' = 
    "((J:^Pred) AND ((preOK (V,E) (A,R) (MSP (ok:^ok_Type)) n x r) 
                     AND (dataOK (V,E) (MSP ok) n x)))
      AND (\s. (s o (r (b:*Vert))) (a:*Vert) = (s o x) a)" ;;

let MDE4_DEF = new_definition
  (`MDE4_DEF`,
   "MDE4 (P:*Vert->^Uprog) (a:*Vert) (b:*Vert) x r =
    ENSURES (P a) TT (\s. (s o (r b)) a = (s o x) a)") ;;


let MDE2_Decom = prove_thm
  (`MDE2_Decom`,
   "a IN V /\
    MDS3 (P:*Vert->^Uprog) J (V,E) (A,R) (MSP (ok:^ok_Type)) n a x r /\
    MDE4 P a b x r
    ==>
    MDE2 P J (V,E) (A,R) (MSP ok) n a b x r",
    REWRITE_TAC [MDE2_DEF; MDE4_DEF; MDS3_DEF]
    THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC ENSURES_CONSQ_WEAK
    THEN EXISTS_TAC Q THEN CONJ_TAC
    THENL [ ALL_TAC; DEL_ALL_TAC THEN pred_JADE_TAC 1 ]
    THEN MATCH_MP_TAC ENSURES_CONSQ_WEAK
    THEN EXISTS_TAC Q' THEN CONJ_TAC
    THENL     
    [ MATCH_MP_TAC ENSURES_PSP_STABLE_TT THEN ASM_REWRITE_TAC[] ;
      MATCH_MP_TAC (GEN_ALL lemma)
      THEN IMP_RES_TAC dataOK_regOK THEN ASM_REWRITE_TAC[] ]
    ) ;;


%---------------------------------------------------------------------------------
     Let us now combine all these results into one theorem.
---------------------------------------------------------------------------------%

let MDS3_TAC =
    MATCH_MP_TAC MDS3_Decom THEN ASM_REWRITE_TAC[]
    THEN CONJ_TAC 
      THENL
      [ MATCH_MP_TAC WRITE_ConfOk THEN REPEAT STRIP_TAC
        THEN FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[] ;
        FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[] ] ;;

let MD_SpecDecom = prove_thm
  (`MD_SpecDecom`,
   "sWF A R /\ ~(A={}) /\ ~(V={}) /\ FINITE A /\ FINITE V /\
    Resolve (V,E) (A,R) Res (ok:^ok_Type) /\
    Proper (P:*Vert->^Uprog) V /\ Disjunct P V /\
    (!a. a IN V ==> (WRITE (P a)) (x a)) /\
    (!a b. a IN V /\ b IN V /\ E a b ==> (WRITE (P a)) (r b a)) /\
    (!a. a IN V ==> STABLE (P a) J) /\
    (!n a. n IN A /\ a IN V ==> MDS4 P J (V,E) (A,R) (MSP ok) n a x r) /\
    (!n a. n IN A /\ a IN V ==> MDS5 P J (V,E) (A,R) (MSP ok) n a x r) /\
    (!n a b. n IN A /\ a IN V /\ b IN V /\ E a b ==>
             MDS6 P J (V,E) (A,R) (MSP ok) n a b x r) /\
    (!a. a IN V ==> MDE3 P a Res x r) /\
    (!a b. a IN V /\ b IN V /\ E a b ==> MDE4 P a b x r)
    ==>
    MDC (Net P V) J (V,E) A (MSP ok) x",
    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC OK_Round_Decom
    THEN ASM_REWRITE_TAC[] THEN REPEAT STRIP_TAC
    THENL
    [ %-- NetConfOk --%
      MATCH_MP_TAC ConfOk_Net THEN ASM_REWRITE_TAC[]
      THEN MATCH_MP_TAC WRITE_ConfOk THEN REPEAT STRIP_TAC
      THEN FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[] ;
      %-- NetRegConfOk --%
      MATCH_MP_TAC RegConfOk_Net THEN ASM_REWRITE_TAC[] 
      THEN MATCH_MP_TAC WRITE_RegConfOk THEN REPEAT STRIP_TAC
      THEN FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[] ;
      %-- STABLE J --%
      REWRITE_TAC [Net_DEF] 
      THEN MATCH_MP_TAC STABLE_PAR_Fold 
      THEN REWRITE_TAC [IN_IMAGE] THEN REPEAT STRIP_TAC
      THENL [ MATCH_MP_TAC IMAGE_FINITE THEN ASM_REWRITE_TAC[] ;
              RES_TAC THEN ASM_REWRITE_TAC[]] ;
      %-- MDS1 --%
      MATCH_MP_TAC MDS1_Decom THEN ASM_REWRITE_TAC[]
      THEN REPEAT STRIP_TAC THEN MDS3_TAC ;
      %-- MDC1 --%
      MATCH_MP_TAC MDC1_Decom THEN ASM_REWRITE_TAC[]
      THEN CONJ_TAC
      THENL
      [ %-- MDS2 --%
        MATCH_MP_TAC MDS2_Decom THEN ASM_REWRITE_TAC[]
        THEN REPEAT STRIP_TAC THEN FIRST_ASSUM MATCH_MP_TAC
        THEN ASM_REWRITE_TAC[] ;
        %-- MDC3 --%
        MATCH_MP_TAC MDC3_Decom THEN ASM_REWRITE_TAC[]
        THEN REPEAT CONJ_TAC 
        THENL [ MATCH_MP_TAC WRITE_ConfOk THEN REPEAT STRIP_TAC
                THEN FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[] ;
                FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[] ;
                FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[] ;
                ALL_TAC ]
        THEN MATCH_MP_TAC MDE1_Decom THEN ASM_REWRITE_TAC[]
        THEN REPEAT CONJ_TAC 
        THEN FIRST_ASSUM MATCH_MP_TAC
        THEN ASM_REWRITE_TAC[] 
        THEN EXISTS_TAC "n:*A" THEN ASM_REWRITE_TAC[] ] ;
      %-- MDC2 --%
      MATCH_MP_TAC MDC2_Decom THEN ASM_REWRITE_TAC[]
      THEN CONJ_TAC
      THENL
      [ %-- MDS1 --%
        MATCH_MP_TAC MDS1_Decom THEN ASM_REWRITE_TAC[]
        THEN REPEAT STRIP_TAC THEN MDS3_TAC ;
        %-- MDC4 --%
        MATCH_MP_TAC MDC4_Decom THEN ASM_REWRITE_TAC[]
        THEN REPEAT CONJ_TAC
        THENL
        [ %-- RegConfOk --%
          MATCH_MP_TAC WRITE_RegConfOk THEN REPEAT STRIP_TAC
          THEN FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[] ;
          %-- MDS3 --%
          MDS3_TAC ;
          %-- MDS6 --%
          FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[] ;
          %-- MDE2 --%
          MATCH_MP_TAC MDE2_Decom THEN ASM_REWRITE_TAC[] 
          THEN CONJ_TAC
          THENL
          [ MDS3_TAC ;
            FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[]
            THEN EXISTS_TAC "n:*A" THEN ASM_REWRITE_TAC[] ]
        ]
      ]]) ;;

%---------------------------------------------------------------------------------
  Let us construct for each P.a a program consisting of two actions:

      Comp: applying the resolvent to compute the new value of x.a
      Copy: copying the x.a to all connected link registers

  We name the program RsP.
---------------------------------------------------------------------------------%

let Comp_DEF = new_definition
  (`Comp_DEF`,
   "Comp Res (a:*Vert) (x:^VarVec) (r:*Vert->^VarVec) (s:^State) (t:^State) = 
        (!v. (v=x a) ==> (t v = Res a (s o (r a)))) /\
        (!v.~(v=x a) ==> (t v = s v))") ;;

let Copy_DEF = new_definition
  (`Copy_DEF`,
   "Copy (a:*Vert) b (x:^VarVec) (r:*Vert->^VarVec) (s:^State) (t:^State) = 
      (!v. (v=r b a) ==> (t v  = (s o x) a)) /\
      (!v.~(v=r b a) ==> (t v = s v))") ;;

let RsP = new_definition
  (`RsP`,
   "(RsP Res (V,E:^CRel) x r a):^Uprog =
        (CHF ({Comp Res a x r} UNION 
              (GSPEC (\b. (Copy a b x r,(b IN V /\ E a b))))),
         TT,
         CHF ({x a} UNION (GSPEC (\b. (r a b,(b IN V /\ E b a))))
                    UNION (GSPEC (\b. (r b a,(b IN V /\ E a b))))),
         CHF ({x a} UNION (GSPEC (\b. (r b a,(b IN V /\ E a b))))))") ;;

%---------------------------------------------------------------------------------
   It is necessary that the variable vector x restricted on V is injective.
   That is, for distinct vertices a and b, x.a and x.b are distinct variables.

   It is also necessary that r and x are disjunct. That is, for any vertices
   a, b, and c, x.a and r.b.c are two distinct variables.

   Also, the link regeister vector r has to be injective. That is, for any
   vertices a,b,c,d: if a,b<>c,d then r.a.b and r.b.c are distinct variables.
---------------------------------------------------------------------------------%

let Inject_DEF = new_definition
  (`Inject_DEF`,
   "Inject (f:*->**) = !a b. ~(b = a) ==> ~(f a = f b)") ;;

let Inject2_DEF = new_definition
  (`Inject2_DEF`,
   "Inject2 (g:*A->*B->*C) = !a b c d. ~((a=c)/\(b=d)) ==> ~(g a b = g c d)") ;;

let Distinct_DEF = new_definition
  (`Distinct_DEF`,
   "Distinct (f:*A->*D) (g:*B->*C->*D) = !a b c. ~(g b c = f a)") ;;

%---------------------------------------------------------------------------------
                  RsP forms a write-disjunct network. 
---------------------------------------------------------------------------------%

let RsP_Disjunct = prove_thm
  (`RsP_Disjunct`,
   "Inject x /\ Inject2 r /\ Distinct x r
    ==>
    Disjunct ((RsP Res (V,E:^CRel) x r):*Vert->^Uprog) V",
    REWRITE_TAC [Inject_DEF; Inject2_DEF; Distinct_DEF; Disjunct_DEF; RsP; ioP; WRITE]
    THEN REWRITE_TAC [SYM (SPEC_ALL IN_DEF); IN_UNION; IN_SING]
    THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
    THEN REPEAT STRIP_TAC
    THENL
    [ %-- 1 --%
      SUBGOAL_THEN "(x:^VarVec) b = x a" ASSUME_TAC
      THENL [ EVERY_ASSUM (\thm. SUBST1_TAC (SYM thm) ? ALL_TAC) THEN REFL_TAC; 
              ALL_TAC]
      THEN RES_TAC ;
      %-- 2 --%
      SUBGOAL_THEN "(r:*Vert->^VarVec) b' b= x (a:*Vert)" ASSUME_TAC
      THENL [ EVERY_ASSUM (\thm. SUBST1_TAC (SYM thm) ? ALL_TAC) THEN REFL_TAC; 
              ALL_TAC]
      THEN RES_TAC ;
      %-- 3 --%
      SUBGOAL_THEN "(r:*Vert->^VarVec) b' a= x (b:*Vert)" ASSUME_TAC
      THENL [ EVERY_ASSUM (\thm. SUBST1_TAC (SYM thm) ? ALL_TAC) THEN REFL_TAC; 
              ALL_TAC]
      THEN RES_TAC ;
      %-- 4 --%
      SUBGOAL_THEN "(r:*Vert->^VarVec) b' a= r b'' b" ASSUME_TAC
      THENL [ EVERY_ASSUM (\thm. SUBST1_TAC (SYM thm) ? ALL_TAC) THEN REFL_TAC; 
              ALL_TAC]
      THEN RULE_ASSUM_TAC (REWRITE_RULE [CONJUNCT1 (SPEC_ALL DE_MORGAN_THM)])
      THEN RES_TAC ]) ;;

%---------------------------------------------------------------------------------
   Prove that RsP a satisfies MDE3 and 4. These are quite easy to prove.
---------------------------------------------------------------------------------%

let ENSURES_TT = prove
  ("!a. UNITY (Pr:^Uprog) /\ PROG Pr a /\ HOA(TT,a,p) 
        ==>
        ENSURES Pr TT p",
    REWRITE_TAC [ENSURES; HOA_DEF] THEN REDUCE_UNLESS_TAC
    THENL 
    [ ASM_REWRITE_TAC[] ;
      EXISTS_TAC "a:^Action" THEN ASM_REWRITE_TAC[]
      THEN REPEAT STRIP_TAC THEN RES_TAC]) ;;

let RsP_sat_MDE3 = prove_thm
  (`RsP_sat_MDE3`,
   "Distinct x r /\ 
    Proper ((RsP Res (V,E:^CRel) x r):*Vert->^Uprog) V /\
    a IN V
    ==>
    MDE3 (RsP Res (V,E:^CRel) x r) a Res x r",
    REWRITE_TAC [MDE3_DEF; Distinct_DEF; Proper_DEF]
    THEN STRIP_TAC THEN RES_TAC
    THEN MATCH_MP_TAC ENSURES_TT THEN ASM_REWRITE_TAC[]
    THEN REWRITE_TAC [RsP; o_DEF; PROG]
    THEN REWRITE_TAC [SYM (SPEC_ALL IN_DEF); IN_UNION; IN_SING]
    THEN EXISTS_TAC "(Comp Res a (x:^VarVec) r):^Action"
    THEN REWRITE_TAC [Comp_DEF; o_DEF; HOA_DEF]
    THEN pred_JADE_TAC 0 
    THEN SUBGOAL_THEN "!b. (t:^State) ((r:*Vert->^VarVec) a b) = s (r a b)"
         (\thm. REWRITE_TAC [thm])
    THENL [ GEN_TAC THEN FIRST_ASSUM MATCH_MP_TAC 
            THEN ASM_REWRITE_TAC[] ; ALL_TAC ]
    THEN FIRST_ASSUM MATCH_MP_TAC THEN REFL_TAC) ;;


let RsP_sat_MDE4 = prove_thm
  (`RsP_sat_MDE4`,
   "Distinct x r /\
    Proper ((RsP Res (V,E:^CRel) x r):*Vert->^Uprog) V /\
    a IN V /\ b IN V /\ E a b
    ==>
    MDE4 (RsP Res (V,E:^CRel) x r) a b x r",
    REWRITE_TAC [MDE4_DEF; Distinct_DEF; Proper_DEF]
    THEN STRIP_TAC THEN RES_TAC
    THEN MATCH_MP_TAC ENSURES_TT THEN ASM_REWRITE_TAC[]
    THEN REWRITE_TAC [RsP; o_DEF; PROG]
    THEN REWRITE_TAC [SYM (SPEC_ALL IN_DEF); IN_UNION] THEN  BETA_TAC
    THEN EXISTS_TAC "(Copy a b (x:^VarVec) r):^Action"
    THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
    THEN REWRITE_TAC [prove("(p\/q)/\r=(p/\r)\/(q/\r)", TAUT_TAC)]
    THEN DISJ2_TAC
    THEN REWRITE_TAC [Copy_DEF; o_DEF; HOA_DEF]
    THEN pred_JADE_TAC 0 
    THENL
    [ %-- 1 --%
      EXISTS_TAC  "b:*Vert" THEN ASM_REWRITE_TAC[] ;
      %-- 2 --%
      SUBGOAL_THEN "(t:^State) ((r:*Vert->^VarVec) b a) = (s:^State) (x a)" 
              SUBST1_TAC
      THENL [ FIRST_ASSUM MATCH_MP_TAC THEN REFL_TAC ; ALL_TAC ]
      THEN CONV_TAC SYM_CONV THEN FIRST_ASSUM MATCH_MP_TAC
      THEN (CONV_TAC o ONCE_DEPTH_CONV) SYM_CONV THEN ASM_REWRITE_TAC[] ]
    ) ;;

%---------------------------------------------------------------------------------
   Prove that RsP a satisfies MDS4, 5 and 6. These requires somewhat lengthy 
   proofs. 

   First let us show that the action Comp respect the requirements of those MDS.
   We show the same for Copy. Then we can conclude that RsP does indeed satisfy
   the required stability.
---------------------------------------------------------------------------------%

Set_Assumptions
   [ (`inject_x`, "Inject (x:^VarVec)");
     (`distinct_xr`, "Distinct (x:^VarVec) (r:*Vert->^VarVec)");
     (`Trans_R`, "TRANS (R:^WRel)");
     (`Resolve_Res`, "Resolve (V,E) (A,R) Res (ok:^ok_Type)")] ;;

LEMMA (`Comp_lemma1`,
     "a IN V /\ n IN A 
      ==>
      HOA ((\s. !m b. m IN A /\ b IN V /\ E b a /\ R m n ==> ok m b (s(r a b))),
          (Comp Res a x r):^Action,
          MSP (ok:^ok_Type) n a x)",
     `Comp applies the resolution Res`,
     RULE_ASSUM_TAC (REWRITE_RULE [Resolve_DEF])
     THEN REWRITE_TAC [MSP; Resolve_DEF; HOA_DEF; Comp_DEF; o_DEF]
     THEN BETA_TAC THEN REPEAT STRIP_TAC
     THEN XRULE_ASSUM_TAC (SPEC "(x:^VarVec) a")
     THEN XRULE_ASSUM_TAC (REWRITE_RULE[])
     THEN DEL_TAC "T" THEN ASM_REWRITE_TAC[]
     THEN FIRST_ASSUM (\thm. MATCH_MP_TAC thm THEN DEL_TAC (concl thm))
     THEN BETA_TAC THEN ASM_REWRITE_TAC[]
     THEN REPEAT STRIP_TAC THEN RES_TAC) ;;

LEMMA (`Comp_lemma2`,
    "a IN V /\ b IN V /\ ~(a=b) 
     ==>
     HOA (MSP (ok:^ok_Type) n b x,
          (Comp Res a x r):^Action,
          MSP ok n b x)",
     `Comp.a leaves x.b untouched`,
     REWRITE_TAC [MSP; HOA_DEF; Comp_DEF; o_DEF]
     THEN BETA_TAC THEN REPEAT STRIP_TAC
     THEN XRULE_ASSUM_TAC (REWRITE_RULE [Inject_DEF])
     THEN XRULE_ASSUM_TAC (SPEC "(x:^VarVec) b")
     THEN RES_TAC THEN RES_TAC
     THEN ASM_REWRITE_TAC[]) ;;

LEMMA (`Comp_lemma3`,
    "a IN V /\ b IN V /\ c IN V 
     ==>
     HOA (MSP (ok:^ok_Type) n b (r c),
          (Comp Res a x r):^Action,
          MSP ok n b (r c))",
     `Comp leaves r untouched`,
     REWRITE_TAC [MSP; HOA_DEF; Comp_DEF; o_DEF]
     THEN pred_JADE_TAC 0
     THEN XRULE_ASSUM_TAC (REWRITE_RULE [Distinct_DEF])
     THEN XRULE_ASSUM_TAC (SPECL ["a:*Vert"; "c:*Vert"; "b:*Vert"])
     THEN RES_TAC THEN ASM_REWRITE_TAC[]) ;;

let lemma4_TAC = 
    ASM_CASES_TAC "a=a':*Vert"
       THENL
       [ %-- 1.1 --%
         FIRST_ASSUM (SUBST1_TAC o SYM)
         THEN MATCH_MP_TAC (QUAFF_RULE 2 (REWRITE_RULE [HOA_DEF] (RECALL `Comp_lemma1`)))
         THEN EXISTS_TAC "s:^State"
         THEN ASM_REWRITE_TAC[]
         THEN REPEAT STRIP_TAC
         THEN RULE_ASSUM_TAC (REWRITE_RULE [MSP; TRANS_DEF]) 
         THEN XRULE_ASSUM_TAC (SPEC "m:*A")
         THEN OLD_RES_TAC THEN OLD_RES_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[] ;
         %-- 1.2 --%
         MATCH_MP_TAC (QUAFF_RULE 2 (REWRITE_RULE [HOA_DEF] (RECALL `Comp_lemma2`)))
         THEN EXISTSL_TAC ["a:*Vert"; "s:^State"]
         THEN ASM_REWRITE_TAC[] THEN OLD_RES_TAC ] ;;

LEMMA (`Comp_lemma4`,
    "a IN V /\ n IN A 
     ==>
     HOA (preOK (V,E) (A,R) (MSP (ok:^ok_Type)) n x r,
          (Comp Res a x r):^Action,
          preOK (V,E) (A,R) (MSP ok) n x r)",
     `using lemma1, 2 and 3`,
     REWRITE_TAC [preOK; dataOK; comOK; HOA_DEF]
     THEN pred_JADE_TAC 0
     THENL
     [ %-- 1 --%
       lemma4_TAC ;
       %-- 2 --%
       MATCH_MP_TAC (QUAFF_RULE 2 (REWRITE_RULE [HOA_DEF] (RECALL `Comp_lemma3`)))
       THEN EXISTSL_TAC ["a:*Vert"; "s:^State"] 
       THEN ASM_REWRITE_TAC[] 
       THEN XRULE_ASSUM_TAC (SPEC "i:*A") THEN OLD_RES_TAC
       THEN OLD_RES_TAC ]) ;;

let lemma5_TAC =
    UNDISCH_ALL_TAC THEN REWRITE_TAC [preOK; comOK; HOA_DEF]
    THEN pred_JADE_TAC 0
    THEN MATCH_MP_TAC (QUAFF_RULE 2 (REWRITE_RULE [HOA_DEF] (RECALL `Comp_lemma1`))) 
    THEN EXISTS_TAC "s:^State"
    THEN ASM_REWRITE_TAC[]
    THEN REPEAT STRIP_TAC
    THEN RULE_ASSUM_TAC (REWRITE_RULE [MSP; TRANS_DEF]) 
    THEN XRULE_ASSUM_TAC (SPEC "m:*A")
    THEN OLD_RES_TAC 
    THEN FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[] ;;

LEMMA (`Comp_lemma5`,
    "n IN A /\ a IN V
     ==>
     HOA ((preOK (V,E) (A,R) (MSP (ok:^ok_Type)) n x r) AND ((MSP ok) n a x),
          (Comp Res a x r):^Action,
          (preOK (V,E) (A,R) (MSP ok) n x r) AND ((MSP ok) n a x))",
     `from lemma1 and 4`,
     REWRITE_TAC [HOA_DEF]
     THEN pred_JADE_TAC 0
     THENL
     [ %-- 1 --%
       MATCH_MP_TAC (QUAFF_RULE 2 (REWRITE_RULE [HOA_DEF] (RECALL `Comp_lemma4`))) 
       THEN EXISTSL_TAC ["a:*Vert"; "s:^State"] 
       THEN ASM_REWRITE_TAC[] ;
       %-- 2 --%
       lemma5_TAC ]) ;;

LEMMA (`Comp_lemma6`,
    "n IN A /\ a IN V /\ b IN V /\ E a b
     ==>
     HOA ((preOK (V,E) (A,R) (MSP (ok:^ok_Type)) n x r) AND (dataOK (V,E) (MSP ok) n x) 
          AND ((MSP ok) n a ( r b)),
          (Comp Res a x r):^Action,
          (preOK (V,E) (A,R) (MSP ok) n x r) AND (dataOK (V,E) (MSP ok) n x) 
          AND ((MSP ok) n a (r b)))",
     `lemma 4 and 5`,
     REWRITE_TAC [dataOK; HOA_DEF]
     THEN pred_JADE_TAC 0
     THENL
     [ %-- 1 --%
       MATCH_MP_TAC (QUAFF_RULE 2 (REWRITE_RULE [HOA_DEF] (RECALL `Comp_lemma4`))) 
       THEN EXISTSL_TAC ["a:*Vert"; "s:^State"] 
       THEN ASM_REWRITE_TAC[] ;
       %-- 2 --%
       UNDISCH_ALL_TAC THEN REWRITE_TAC [preOK; dataOK; comOK]
       THEN pred_JADE_TAC 0
       THEN lemma4_TAC ;
       %-- 3 --%
       MATCH_MP_TAC (QUAFF_RULE 2 (REWRITE_RULE [HOA_DEF] (RECALL `Comp_lemma3`)))
       THEN EXISTSL_TAC ["a:*Vert"; "s:^State"] 
       THEN ASM_REWRITE_TAC[] ]) ;;

let Comp_sat_MDS4 = (QUAFF_RULE 2 (GEN_RECALL `Comp_lemma4`));;
let Comp_sat_MDS5 = (QUAFF_RULE 2 (GEN_RECALL `Comp_lemma5`));;
let Comp_sat_MDS6 = (QUAFF_RULE 2 (GEN_RECALL `Comp_lemma6`));;

%---------------------------------------------------------------------------------
    Let now show that Copy also meet the requirement of MDS4, 5, and 6.
---------------------------------------------------------------------------------%

Set_Assumptions
   [ (`inject_x`, "Inject (x:^VarVec)");
     (`inject_r`, "Inject2 (r:*Vert->^VarVec)"); 
     (`distinct_xr`, "Distinct (x:^VarVec) (r:*Vert->^VarVec)");
     (`Trans_R`, "TRANS (R:^WRel)");
     (`Resolve_Res`, "Resolve (V,E) (A,R) Res (ok:^ok_Type)")] ;;

LEMMA (`Copy_lemma1`,
  "n IN A /\ a IN V /\ b IN V /\ E a b
   ==>
   HOA (preOK (V,E) (A,R) (MSP (ok:^ok_Type)) n x r,
        (Copy a b x r):^Action,
        preOK (V,E) (A,R) (MSP ok) n x r)",
   `definition of Copy, preOK, etc`,
    REWRITE_TAC [preOK; dataOK; comOK; MSP; Copy_DEF; o_DEF; HOA_DEF]
    THEN pred_JADE_TAC 0
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [Inject2_DEF; Distinct_DEF])
    THENL
    [ %-- 1 --%
      SUBGOAL_THEN "(t:^State) ((x:^VarVec) a') = s (x a')" SUBST1_TAC
      THENL [ FIRST_ASSUM MATCH_MP_TAC 
              THEN (CONV_TAC o ONCE_DEPTH_CONV) SYM_CONV THEN ASM_REWRITE_TAC[] ;
              ALL_TAC ]
      THEN RES_TAC ;
      %-- 2 --%
      ASM_CASES_TAC "((a':*Vert)=a) /\ ((b':*Vert)=b)"
      THENL
      [ %-- 2.1. --%
        UNDISCH_ALL_TAC THEN REPEAT STRIP_TAC
        THEN ASM_REWRITE_TAC[]
        THEN SUBGOAL_THEN "(t:^State) ((r:*Vert->^VarVec) b a) = (s:^State) (x a)" SUBST1_TAC
        THENL [ FIRST_ASSUM MATCH_MP_TAC THEN REFL_TAC; ALL_TAC ]
        THEN RES_TAC ;
        %-- 2.2 --%
        SUBGOAL_THEN "(t:^State) ((r:*Vert->^VarVec) b' a') = s (r b' a')" SUBST1_TAC
        THENL [ FIRST_ASSUM MATCH_MP_TAC THEN FIRST_ASSUM MATCH_MP_TAC
                THEN ONCE_REWRITE_TAC [CONJ_SYM] THEN ASM_REWRITE_TAC[] ;
                ALL_TAC ]
        THEN RES_TAC ]
    ]) ;;

LEMMA (`Copy_lemma2`,
    "n IN A /\ a IN V /\ b IN V /\ E a b
     ==>
     HOA ((preOK (V,E) (A,R) (MSP (ok:^ok_Type)) n x r) AND ((MSP ok) n a x),
          (Copy a b x r):^Action,
          (preOK (V,E) (A,R) (MSP ok) n x r) AND ((MSP ok) n a x))",
     `use also lemma1`,
     REWRITE_TAC [HOA_DEF]
     THEN pred_JADE_TAC 0
     THENL
     [ %-- 1 --%
       MATCH_MP_TAC (QUAFF_RULE 2 (REWRITE_RULE [HOA_DEF] (RECALL `Copy_lemma1`)))
       THEN EXISTSL_TAC ["a:*Vert"; "b:*Vert"; "s:^State"]
       THEN ASM_REWRITE_TAC[] ;
       %-- 2 --%
       UNDISCH_ALL_TAC
       THEN REWRITE_TAC [preOK; dataOK; comOK; MSP; Copy_DEF; o_DEF; Distinct_DEF]
       THEN pred_JADE_TAC 0
       THEN SUBGOAL_THEN "(t:^State) ((x:^VarVec) a) = s (x a)" SUBST1_TAC
       THENL [ FIRST_ASSUM MATCH_MP_TAC 
              THEN (CONV_TAC o ONCE_DEPTH_CONV) SYM_CONV THEN ASM_REWRITE_TAC[] ;
              ALL_TAC ]
       THEN ASM_REWRITE_TAC[] ]
     ) ;;


LEMMA (`Copy_lemma3`,
    "n IN A /\ a IN V /\ b IN V /\ b' IN V /\ E a b /\ E a b'
     ==>
     HOA ((preOK (V,E) (A,R) (MSP (ok:^ok_Type)) n x r) AND (dataOK (V,E) (MSP ok) n x) 
          AND ((MSP ok) n a (r b')),
          (Copy a b x r):^Action,
          (preOK (V,E) (A,R) (MSP ok) n x r) AND (dataOK (V,E) (MSP ok) n x) 
          AND ((MSP ok) n a (r b')))",
     `use lemma 1`,
     REWRITE_TAC [dataOK; HOA_DEF]
     THEN pred_JADE_TAC 0 
     THENL
     [ %-- 1 --%
       MATCH_MP_TAC (QUAFF_RULE 2 (REWRITE_RULE [HOA_DEF] (RECALL `Copy_lemma1`)))
       THEN EXISTSL_TAC ["a:*Vert"; "b:*Vert"; "s:^State"]
       THEN ASM_REWRITE_TAC[] ;
       %-- 2 --%
       UNDISCH_ALL_TAC
       THEN REWRITE_TAC [preOK; dataOK; comOK; MSP; Copy_DEF; o_DEF; Distinct_DEF]
       THEN pred_JADE_TAC 0
       THEN SUBGOAL_THEN "(t:^State) ((x:^VarVec) a') = s (x a')" SUBST1_TAC
       THENL [ FIRST_ASSUM MATCH_MP_TAC 
              THEN (CONV_TAC o ONCE_DEPTH_CONV) SYM_CONV THEN ASM_REWRITE_TAC[] ;
              ALL_TAC ]
       THEN RES_TAC ;
       %-- 3 --%
       UNDISCH_ALL_TAC 
       THEN REWRITE_TAC [preOK; dataOK; comOK; MSP; Copy_DEF; o_DEF; Distinct_DEF;
                         Inject2_DEF]
       THEN pred_JADE_TAC 0
       THEN ASM_CASES_TAC "(b':*Vert)=b"
       THENL
       [ %-- 2.1. --%
         ASM_REWRITE_TAC[]
         THEN SUBGOAL_THEN "(t:^State) ((r:*Vert->^VarVec) b a) = (s:^State) (x a)" 
              SUBST1_TAC
         THENL [ FIRST_ASSUM MATCH_MP_TAC THEN REFL_TAC; ALL_TAC ]
         THEN RES_TAC ;
        %-- 2.2 --%
        SUBGOAL_THEN "(t:^State) ((r:*Vert->^VarVec) b' a) = s (r b' a)" SUBST1_TAC
        THENL [ FIRST_ASSUM MATCH_MP_TAC THEN FIRST_ASSUM MATCH_MP_TAC
                THEN ASM_REWRITE_TAC[] ; ALL_TAC ]
        THEN ASM_REWRITE_TAC[] ]
      ]) ;;

let Copy_sat_MDS4 = (QUAFF_RULE 2 (GEN_RECALL `Copy_lemma1`));;
let Copy_sat_MDS5 = (QUAFF_RULE 2 (GEN_RECALL `Copy_lemma2`));;
let Copy_sat_MDS6 = (QUAFF_RULE 2 (GEN_RECALL `Copy_lemma3`));;

%---------------------------------------------------------------------------------
    Now let us prove that RsP satisfies MDS4, 5, and 6.
---------------------------------------------------------------------------------%

let RsP_sat_MDS_TAC = 
    RESTRICT_DEF_TAC
    THEN REWRITE_TAC [(SYM o SPEC_ALL) IN_DEF; IN_SING; IN_UNION]
    THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
    THEN REWRITE_TAC [pAND_UNIT; NOT_FF]
    THEN REWRITE_TAC [ prove ("(p:^Pred) AND TT = p", pred_CYAN_TAC)]
    THEN REWRITE_TAC [ prove ("(p:^Pred) OR FF = p", pred_CYAN_TAC)]
    THEN REPEAT STRIP_TAC ;;

let RsP_sat_MDS4 = prove_thm
  (`RsP_sat_MDS4`,
   "Resolve (V,E) (A,R) Res (ok:^ok_Type) /\
    TRANS R /\ Inject (x:^VarVec) /\ Inject2 (r:*Vert->^VarVec) /\ Distinct x r /\
    n IN A /\ a IN V 
    ==>
    MDS4 (RsP Res (V,E) x r) TT (V,E) (A,R) (MSP ok) n a x r",
    REWRITE_TAC [MDS4_DEF; RsP; STABLE; UNLESS; PROG]
    THEN RsP_sat_MDS_TAC
    THENL 
    [ %-- 1 --%
      ASM_REWRITE_TAC[] THEN MATCH_MP_TAC Comp_sat_MDS4
      THEN ASM_REWRITE_TAC[] ;
      %-- 2 --%
      ASM_REWRITE_TAC[] THEN MATCH_MP_TAC Copy_sat_MDS4
      THEN EXISTS_TAC "Res:*Vert->(*Vert->*value)->*value"
      THEN ASM_REWRITE_TAC[]]) ;;

let RsP_sat_MDS5 = prove_thm
  (`RsP_sat_MDS5`,
   "Resolve (V,E) (A,R) Res (ok:^ok_Type) /\
    TRANS R /\ Inject (x:^VarVec) /\ Inject2 (r:*Vert->^VarVec) /\ Distinct x r /\
    n IN A /\ a IN V 
    ==>
    MDS5 (RsP Res (V,E) x r) TT (V,E) (A,R) (MSP ok) n a x r",
    REWRITE_TAC [MDS5_DEF; RsP; STABLE; UNLESS; PROG]
    THEN RsP_sat_MDS_TAC
    THENL 
    [ %-- 1 --%
      ASM_REWRITE_TAC[] THEN MATCH_MP_TAC Comp_sat_MDS5
      THEN ASM_REWRITE_TAC[] ;
      %-- 2 --%
      ASM_REWRITE_TAC[] THEN MATCH_MP_TAC Copy_sat_MDS5
      THEN EXISTS_TAC "Res:*Vert->(*Vert->*value)->*value"
      THEN ASM_REWRITE_TAC[]]) ;;

let RsP_sat_MDS6 = prove_thm
  (`RsP_sat_MDS6`,
   "Resolve (V,E) (A,R) Res (ok:^ok_Type) /\
    TRANS R /\ Inject (x:^VarVec) /\ Inject2 (r:*Vert->^VarVec) /\ Distinct x r /\
    n IN A /\ a IN V /\ b IN V /\ E a b
    ==>
    MDS6 (RsP Res (V,E) x r) TT (V,E) (A,R) (MSP ok) n a b x r",
    REWRITE_TAC [MDS6_DEF; RsP; STABLE; UNLESS; PROG]
    THEN RsP_sat_MDS_TAC
    THENL 
    [ %-- 1 --%
      ASM_REWRITE_TAC[] THEN MATCH_MP_TAC Comp_sat_MDS6
      THEN ASM_REWRITE_TAC[] ;
      %-- 2 --%
      ASM_REWRITE_TAC[] THEN MATCH_MP_TAC Copy_sat_MDS6
      THEN EXISTS_TAC "Res:*Vert->(*Vert->*value)->*value"
      THEN ASM_REWRITE_TAC[]]) ;;

%---------------------------------------------------------------------------------
    The following theorem states that RsP satisfies the specification MDC. It
    implements thus a minimun distance computation.
---------------------------------------------------------------------------------%

let RsP_sat_MDC = prove_thm
  (`RsP_sat_MDC`,
   "sWF A R /\ ~(A={}) /\ ~(V={}) /\ TRANS R /\ FINITE A /\ FINITE V /\
    Resolve (V,E) (A,R) Res (ok:^ok_Type) /\
    Proper (RsP Res (V,E) x r) V /\
    Inject (x:^VarVec) /\ Inject2 r /\ Distinct x r 
    ==>
    MDC (Net (RsP Res (V,E) x r) V) TT (V,E) A (MSP ok) x",
    STRIP_TAC
    THEN MATCH_MP_TAC MD_SpecDecom
    THEN ASM_REWRITE_TAC[] THEN REPEAT STRIP_TAC
    THENL
    [ %-- 1 --%
      MATCH_MP_TAC RsP_Disjunct THEN ASM_REWRITE_TAC[] ;
      %-- 2 --%
      REWRITE_TAC [RsP; WRITE; SYM (SPEC_ALL IN_DEF); IN_UNION; IN_SING] ;
      %-- 3 --%
      REWRITE_TAC [RsP; WRITE; SYM (SPEC_ALL IN_DEF); IN_UNION]
      THEN DISJ2_TAC THEN CONV_TAC SET_SPEC_CONV
      THEN EXISTS_TAC "b:*Vert" THEN ASM_REWRITE_TAC[] ;
      %-- 4 --%
      REWRITE_TAC [STABLE_TT] ;
      %-- 5 --%
      MATCH_MP_TAC RsP_sat_MDS4 THEN ASM_REWRITE_TAC[] ;
      %-- 6 --%
      MATCH_MP_TAC RsP_sat_MDS5 THEN ASM_REWRITE_TAC[] ;
      %-- 7 --%
      MATCH_MP_TAC RsP_sat_MDS6 THEN ASM_REWRITE_TAC[] ;
      %-- 8 --%
      MATCH_MP_TAC RsP_sat_MDE3 THEN ASM_REWRITE_TAC[] ;
      %-- 9 --%
      MATCH_MP_TAC RsP_sat_MDE4 THEN ASM_REWRITE_TAC[]
    ]) ;;


close_theory();;




