let PAR_TAC =
       MATCH_MP_TAC UNLESS_PAR 
       ORELSE MATCH_MP_TAC UNL_PAR 
       ORELSE MATCH_MP_TAC INV_PAR
       ORELSE MATCH_MP_TAC (QUAFF_RULE 3 ENSURES_PAR)
       ORELSE MATCH_MP_TAC (QUAFF_RULE 3 ENS_PAR)
       ORELSE MATCH_MP_TAC (QUAFF_RULE 5 SINGH_thm)
       ORELSE MATCH_MP_TAC (QUAFF_RULE 5 SINGH_SIMPLE_thm)
       THEN REPEAT CONJ_TAC ;;
let INV_STR_TAC =
    MATCH_MP_TAC UNL_INV_STR 
    ORELSE MATCH_MP_TAC ENS_INV_STR 
    ORELSE MATCH_MP_TAC (QUAFF_RULE 5 LTO_INV_STR)
    THEN REPEAT CONJ_TAC ;;
let SR_TAC p' q' (asml,g) =
    let J = rand(rator(rator g)) in
    let p = rand(rator g) in
    let q = rand g in
    (REWRITE_TAC [UNL; ENS; LTO]
     THEN SUBGOAL_THEN "(^p AND ^J) = (^p' AND ^J)" SUBST1_TAC
     THENL
     [ ALL_TAC ;
       SUBGOAL_THEN "(^q AND ^J) = (^q' AND ^J)" SUBST1_TAC
       THENL [ ALL_TAC ;
               REWRITE_TAC (map (SYM o SPEC_ALL) [UNL; ENS; LTO])]])
    (asml,g) ;; 
let Uop_Type PredT UprogT =
    % make the type of UNITY operators, eg, LEADSTO %
    mk_type (`fun`,[UprogT; 
       mk_type (`fun`,[PredT;
          mk_type (`fun`,[PredT;":bool"])])]) ;;

let UPop_Type PredT UprogT =
    % make the type of paremeterized UNITY operators, eg, LTO %
    mk_type (`fun`,[UprogT;
       mk_type (`fun`,[PredT;
          mk_type (`fun`,[PredT;
             mk_type (`fun`,[PredT; ":bool"])])])]) ;;

let Uop_find OP trm = 
    % find all occurrences of a subterm of the form OP Pr p q%
    find_terms (\tr. rator(rator(rator tr))=OP ? false) trm ;;

let UP_cons OP J trm  =
    % given trm = OP Pr p q, form OP Pr J p q %
    list_mk_comb (OP,[rand(rator(rator trm)); J; rand(rator trm); rand trm]) ;;

let mk_U2UP_subs str1 str2 PredT UprogT J trm =
    % make str1 --> str2, eg, UNLESS --> UNL, substitution list %
    map 
    (\tr. (UP_cons (mk_const (str2,UPop_Type PredT UprogT)) J tr),tr)
    (Uop_find (mk_const (str1,Uop_Type PredT UprogT)) trm) ;;

let U2UP_subs PredT UprogT J trm =
    % subsitute all occurrences of UNITY operators with parameterized
      UNITY operators %
    let subs_lst = 
        append (mk_U2UP_subs `UNLESS` `UNL` PredT UprogT J trm)
       (append (mk_U2UP_subs `ENSURES` `ENS` PredT UprogT J trm)
               (mk_U2UP_subs `LEADSTO` `LTO` PredT UprogT J trm))  in 
    subst subs_lst trm ;;

let Lsubs p thm =
    % if thm = |- op P a b, and if |- a=p, then substitute a with p %
    (let eq_thm = prove(mk_eq(rand(rator(concl thm)),p), 
                 pred_CYAN_TAC) in
     SUBS [eq_thm] thm) 
     ? thm ;;

let Rsubs q thm =
    % if thm = |- op P a b, and if |- b=q, then substitute b with q %
    (let eq_thm = prove(mk_eq(rand(concl thm),q) ,pred_CYAN_TAC) in
     SUBS [eq_thm] thm) 
     ? thm ;;
        
let Lsubs_TAC (asml,g) =
    RULE_ASSUM_TAC (Lsubs (rand(rator g))) (asml,g);;

let Rsubs_TAC (asml,g) =
    RULE_ASSUM_TAC (Rsubs (rand g)) (asml,g);;


let U2UP_LIFT PredT UprogT J thm =
   let lifted_term = U2UP_subs PredT UprogT J (concl (SPEC_ALL thm)) in
   (GEN_ALL o prove) 
     (lifted_term,
      REWRITE_TAC [UNL; ENS; LTO] 
      THEN REPEAT STRIP_TAC 
      THENL [ALL_TAC ; ASM_REWRITE_TAC[]]
      THEN OLD_IMP_RES_TAC (QUAFF_RULE 8 thm)
      THEN Lsubs_TAC
      THEN Rsubs_TAC
      THEN ONCE_ASM_REWRITE_TAC[]) ;;
