%-------------------------------------------------------------------------------
   This theory defines a convergence operator in terms of progress by reach.
   Various basic properties of convergence are given here.

   Author:

   Wishnu Prasetya
   Dept. of Computer Science
   Utrecht University
   the Netherlands
  
   email: wishnu@cs.ruu.nl
-------------------------------------------------------------------------------%

loadf `UU_UNITY` ;;
load_theory `reach` ;;
new_theory `con` ;;
new_parent `dag`;;

let Autoload thy =    
    let defs = map fst (definitions thy) in
        map (\name. autoload_theory(`definition`,thy,name)) defs;
    let thms = map fst (theorems thy) in
        map (\name. autoload_theory(`theorem`,thy,name)) thms; 
    delete_cache thy; ();;

Autoload `reach` ;;
Autoload `dag` ;;

loadf `OLD_RES`;;
loadf `reach_SUP`;;

map (delete_cache o fst) (cached_theories()) ;;

%-------------------------------------------------------------------------------
                definition of convergence in terms of REACH
-------------------------------------------------------------------------------%

let CON = new_definition
  (`CON`,
   "CON (Pr:^Uprog) J p q =
      (WRITE Pr) CONF q /\
      (?q'. REACH Pr J p (q' AND q) /\ STABLE Pr (q' AND q AND J))") ;;

%-------------------------------------------------------------------------------
   Convergence implies progress, properness of the program, and stability
-------------------------------------------------------------------------------%

let CON_IMP_REACH = prove_thm
  (`CON_IMP_REACH`,
   "!(Pr:^Uprog) J. (CON Pr J) SUBREL  (REACH Pr J)",
    REWRITE_TAC [SUBREL; CON]
    THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC REACH_MONO
    THEN EXISTS_TAC "q' AND (y:^Pred)"
    THEN ASM_REWRITE_TAC[]
    THEN pred_JADE_TAC 3) ;;

let CON_IMP_UNITY = prove_thm
  (`CON_IMP_UNITY`,
   "!(Pr:^Uprog) J p q. CON Pr J p q ==> UNITY Pr",
    REWRITE_TAC [CON] THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC REACH_IMP_UNITY) ;;

let CON_IMP_STABLE = prove_thm
  (`CON_IMP_STABLE`,
   "!(Pr:^Uprog) J p q. CON Pr J p q ==> STABLE Pr J",
    REWRITE_TAC [CON] THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC REACH_IMP_STABLE) ;;

%-------------------------------------------------------------------------------
       the theorem below shows how to implement convergence with REACH           
-------------------------------------------------------------------------------%

let CON_REACH_LIFT = prove_thm
  (`CON_REACH_LIFT`,
   "!(Pr:^Uprog) J p q.
        REACH Pr J p q /\ STABLE Pr (q AND J) ==> CON Pr J p q",
    REWRITE_TAC [CON]
    THEN REPEAT STRIP_TAC
    THENL
    [ IMP_RES_TAC REACH_IMP_CONF ;
      EXISTS_TAC "TT:^Pred"
      THEN ASM_REWRITE_TAC [pAND_UNIT] ]) ;;

%-------------------------------------------------------------------------------
    the theorem below shows how to implement convergence with ENSURES
-------------------------------------------------------------------------------%

let CON_ENSURES_LIFT = prove_thm
  (`CON_ENSURES_LIFT`,
   "!(Pr:^Uprog) J p q.
       (WRITE Pr) CONF p /\ (WRITE Pr) CONF q /\
        STABLE Pr J /\ STABLE Pr (q AND J) /\
        ENSURES Pr (p AND J) q
        ==>
        CON Pr J p q",
    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC CON_REACH_LIFT
    THEN ASM_REWRITE_TAC[] 
    THEN MATCH_MP_TAC (REWRITE_RULE [SUBREL; B_ENS] REACH_ENS_LIFT)
    THEN ASM_REWRITE_TAC[]) ;;

%-------------------------------------------------------------------------------
                         monotonicity of convergence
-------------------------------------------------------------------------------%

let lemma = prove
  ("|==(q IMP r) ==> ((q AND r) = (q:^Pred))",
    pred_JADE_TAC 1
    THEN CONV_TAC FUN_EQ_CONV
    THEN GEN_TAC THEN BETA_TAC
    THEN EQ_TAC THEN REPEAT STRIP_TAC
    THEN RES_TAC THEN ASM_REWRITE_TAC[]) ;;

let CON_MONO = prove_thm
  (`CON_MONO`,
   "!(Pr:^Uprog) J p q r.
         (WRITE Pr) CONF r /\ CON Pr J p q /\ |== (q IMP r) ==> CON Pr J p r",
    REWRITE_TAC [CON]
    THEN REPEAT STRIP_TAC
    THENL
    [ ASM_REWRITE_TAC[] ;
      EXISTS_TAC "q' AND (q:^Pred)"
      THEN REWRITE_TAC [pAND_ASSOC]
      THEN SUBST1_TAC (SYM(ISPECL ["q:^Pred"; "r:^Pred"; "J:^Pred"]
                       pAND_ASSOC))
      THEN FIRST_ASSUM (\thm. REWRITE_TAC [ MATCH_MP lemma thm])
      THEN ASM_REWRITE_TAC[]]) ;;

let lemma1 = prove
  ("|==(((q:^Pred) AND J) IMP r) 
    ==> (((q' AND q) AND (r AND J)) = (q' AND (q AND J)))",
    pred_JADE_TAC 1
    THEN CONV_TAC FUN_EQ_CONV
    THEN GEN_TAC THEN BETA_TAC
    THEN EQ_TAC THEN REPEAT STRIP_TAC
    THEN RES_TAC THEN ASM_REWRITE_TAC[]) ;;

let lemma2 = prove
  ("|==(((q:^Pred) AND J) IMP r) ==> |==(((q' AND q) AND J) IMP ((q' AND q) AND r))",
    pred_JADE_TAC 1) ;;

let CON_gMONO = prove_thm
  (`CON_gMONO`,
   "!(Pr:^Uprog) J p q r.
         (WRITE Pr) CONF r /\ CON Pr J p q /\ |== ((q AND J) IMP r) 
         ==> 
         CON Pr J p r",
    REWRITE_TAC [CON]
    THEN REPEAT STRIP_TAC
    THEN ASM_REWRITE_TAC[] 
    THEN EXISTS_TAC "q' AND (q:^Pred)"
    THEN IMP_RES_TAC lemma1 THEN ASM_REWRITE_TAC[] 
    THEN IMP_RES_TAC lemma2
    THEN IMP_RES_TAC REACH_IMP_CONF
    THEN MATCH_MP_TAC REACH_gMONO
    THEN EXISTS_TAC "q' AND (q:^Pred)" THEN ASM_REWRITE_TAC[] 
    THEN CONF_TAC 3) ;;

let CON_ANTIMONO = prove_thm
  (`CON_ANTIMONO`,
   "!(Pr:^Uprog) J p q r.
         (WRITE Pr) CONF p /\ CON Pr J q r /\ |== (p IMP q) ==> CON Pr J p r",
    REWRITE_TAC [CON]
    THEN REPEAT STRIP_TAC
    THENL 
    [ ASM_REWRITE_TAC[] ;
      EXISTS_TAC "q':^Pred" 
      THEN ASM_REWRITE_TAC[]
      THEN MATCH_MP_TAC REACH_ANTIMONO
      THEN EXISTS_TAC "q:^Pred" THEN ASM_REWRITE_TAC[]]) ;;

let CON_gANTIMONO = prove_thm
  (`CON_gANTIMONO`,
   "!(Pr:^Uprog) J p q r.
         (WRITE Pr) CONF p /\ CON Pr J q r /\ |== ((p AND J) IMP q) 
         ==> 
         CON Pr J p r",
    REWRITE_TAC [CON]
    THEN REPEAT STRIP_TAC
    THEN ASM_REWRITE_TAC[] 
    THEN EXISTS_TAC "(q':^Pred)"
    THEN ASM_REWRITE_TAC[] 
    THEN IMP_RES_TAC REACH_gANTIMONO) ;;

let lemma1 = prove
  ("|== ((J:^Pred) IMP (p EQUAL r)) ==> |== ((r AND J) IMP p)",
    pred_JADE_TAC 1) ;;

let lemma2 = prove
  ("|== ((J:^Pred) IMP (q EQUAL s)) ==> |== ((q AND J) IMP s)",
    pred_JADE_TAC 1) ;;

let CON_SUBST = prove_thm
  (`CON_SUBST`,
   "!(Pr:^Uprog) J p q r s.
        (WRITE Pr) CONF r /\ (WRITE Pr) CONF s /\
         |== (J IMP (p EQUAL r)) /\ |== (J IMP (q EQUAL s)) /\
         CON Pr J p q
         ==>
         CON Pr J r s",
    REPEAT STRIP_TAC
    THEN IMP_RES_TAC lemma1
    THEN IMP_RES_TAC lemma2
    THEN IMP_RES_TAC CON_gMONO
    THEN IMP_RES_TAC CON_gANTIMONO) ;;

%-------------------------------------------------------------------------------
                         reflexivity of convergence
-------------------------------------------------------------------------------%

let CON_REFL = prove_thm
  (`CON_REFL`,
   "!(Pr:^Uprog) J p.
         UNITY Pr /\ (WRITE Pr) CONF p /\ 
         STABLE Pr J /\ STABLE Pr (p AND J) ==> CON Pr J p p",
    REWRITE_TAC [CON]
    THEN REPEAT STRIP_TAC
    THENL 
    [ ASM_REWRITE_TAC[] ;
      EXISTS_TAC "TT:^Pred" 
      THEN ASM_REWRITE_TAC[pAND_UNIT]
      THEN MATCH_MP_TAC REACH_REFL 
      THEN ASM_REWRITE_TAC[] ]) ;;

%-------------------------------------------------------------------------------
                 spiralling law and transitivity of convergence
-------------------------------------------------------------------------------%

let PROVE_STABLE_TAC n =
    UNDISCH_ALL_TAC
    THEN REWRITE_TAC [STABLE] 
    THEN REDUCE_UNLESS_TAC
    THEN REPEAT_FIN n (OLD_RES_TAC THEN ASM_REWRITE_TAC[]) ;;

let CON_SPIRAL = prove_thm
  (`CON_SPIRAL`,
   "!(Pr:^Uprog) J p q r.
         CON Pr J p q /\ CON Pr J q r ==> CON Pr J p (q AND r)",
    REWRITE_TAC [CON]
    THEN REPEAT STRIP_TAC
    THENL
    [ CONF_TAC 2 ;
      EXISTS_TAC "q' AND (q'':^Pred)" THEN CONJ_TAC
      THENL
      [ %--- REACH part ---%
        REACH_TRANS_TAC "q' AND (q:^Pred)"
        THEN SUBST1_TAC (SYM(ISPEC "q' AND (q:^Pred)" pAND_REFL))
        THEN SUBST1_TAC (prove
             ("((q' AND q'') AND (q AND r)) =
               (((q'' AND r) AND (q' AND q)) OR (FF:^Pred))",
              pred_CYAN_TAC))
        THEN XRULE_ASSUM_TAC (REWRITE_RULE [STABLE])
        THEN XRULE_ASSUM_TAC (REWRITE_RULE [SYM(SPEC_ALL pAND_ASSOC)])
        THEN REACH_PSP_TAC 
        THENL [IMP_RES_TAC REACH_IMP_CONF ; CONF_TAC 1 ; ALL_TAC]
        THEN REACH_ANTIMONO_TAC "q:^Pred"
        THENL [ IMP_RES_TAC REACH_IMP_CONF ; pred_JADE_TAC 2] ;
        %--- STABLE part ---%
        PROVE_STABLE_TAC 4 ]
    ]) ;;

let CON_TRANS = prove_thm
  (`CON_TRANS`,
   "!(Pr:^Uprog) J p q r.
         CON Pr J p q /\ CON Pr J q r ==> CON Pr J p r",        
    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC CON_MONO
    THEN EXISTS_TAC "q AND (r:^Pred)"
    THEN REPEAT CONJ_TAC
    THENL
    [ XRULE_ASSUM_TAC (REWRITE_RULE [CON])
      THEN ASM_REWRITE_TAC[] ;
      IMP_RES_TAC CON_SPIRAL ;
      pred_JADE_TAC 3 ]) ;;

%-------------------------------------------------------------------------------
                        conjunction of convergences
-------------------------------------------------------------------------------%

let CON_SIMPLE_CONJ = prove_thm
  (`CON_SIMPLE_CONJ`,
   "!(Pr:^Uprog) J p q r s.
         CON Pr J p q /\ CON Pr J r s ==> CON Pr J (p AND r) (q AND s)",   
    REWRITE_TAC [CON]
    THEN REPEAT STRIP_TAC
    THENL
    [ CONF_TAC 1 ;
      EXISTS_TAC "q' AND (q'':^Pred)" THEN CONJ_TAC
      THENL
      [ %--- REACH part ---%
        XRULE_ASSUM_TAC (REWRITE_RULE [SYM(SPEC_ALL pAND_ASSOC)])
        THEN SUBST1_TAC (prove
             ("((q' AND q'') AND (q AND s)) =
               ((q' AND q) AND (q'' AND (s:^Pred)))", pred_CYAN_TAC))
        THEN IMP_RES_TAC REACH_STABLE_SIMPLE_CONJ ;
        %--- STABLE part ---%
        PROVE_STABLE_TAC 4 ]
     ]) ;;

let MK_ABS_CONV x f = SYM(BETA_CONV "(\^x. ^f) ^x") ;;

let qAND_AND_qAND = prove
   ("!W (f:**->^Pred) g.
        ((!!i::W. f i) AND (!!i::W. g i)) = (!!i::W. (f i) AND (g i))",
     pred_JADE_TAC 0  
     THEN CONV_TAC FUN_EQ_CONV
     THEN BETA_TAC THEN GEN_TAC 
     THEN EQ_TAC THEN REPEAT STRIP_TAC
     THEN RES_TAC) ;;
     
let CON_CONJ = prove_thm
  (`CON_CONJ`,
   "!(W:**->bool) (Pr:^Uprog) J P Q.
        CHF_FINITE W /\ UNITY Pr /\ STABLE Pr J /\
        (!i :: W. CON Pr J (P i) (Q i))
        ==>
        CON Pr J (!!i::W. P i) (!!i::W. Q i)",
    REWRITE_TAC [CON]
    THEN RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
    THENL
    [ CONF_TAC 1
      THEN RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
      THEN RES_TAC ;
      (XRULE_ASSUM_TAC o CONV_RULE o ONCE_DEPTH_CONV) RIGHT_AND_EXISTS_CONV
      THEN (XRULE_ASSUM_TAC o CONV_RULE o ONCE_DEPTH_CONV) RIGHT_IMP_EXISTS_CONV
      THEN (XRULE_ASSUM_TAC o CONV_RULE) SKOLEM_CONV 
      THEN UNDISCH_ALL_TAC THEN REPEAT STRIP_TAC
      THEN EXISTS_TAC "!!i:**::W. (q' i):^Pred" THEN CONJ_TAC
      THENL
      [ %--- REACH part ---%
        REWRITE_TAC [qAND_AND_qAND]
        THEN (CONV_TAC o RAND_CONV o RAND_CONV o ABS_CONV) (MK_ABS_CONV "i:**")
        THEN MATCH_MP_TAC REACH_STABLE_CONJ
        THEN RESTRICT_DEF_TAC
        THEN REWRITE_TAC [pAND_ASSOC]
        THEN REPEAT STRIP_TAC
        THEN OLD_RES_TAC THEN ASM_REWRITE_TAC[] ;
        %--- STABLE part ---%
        PROVE_STABLE_TAC 5 ]
   ]) ;;


%-------------------------------------------------------------------------------
                     Sifting the stable predicate.
-------------------------------------------------------------------------------%


let CON_STABLE_SHIFT = prove_thm
  (`CON_STABLE_SHIFT`,
   "!(Pr:^Uprog) J a p q.
        (WRITE Pr) CONF a /\ CON Pr (J AND a) p q /\ STABLE Pr J ==>
        CON Pr J (p AND a) q",
     REWRITE_TAC [CON] 
     THEN REPEAT STRIP_TAC 
     THENL [ ASM_REWRITE_TAC[] ; ALL_TAC ]  
     THEN EXISTS_TAC "(q':^Pred) AND a"
     THEN CONJ_TAC
     THENL
     [ %-- reach part --%
       (CONV_TAC o RATOR_CONV o RAND_CONV o RAND_CONV o ONCE_REWRITE_CONV)  
       [SYM (ISPEC "a:^Pred" pAND_REFL)]
        THEN REWRITE_TAC [SYM (SPEC_ALL pAND_ASSOC)]   
        THEN MATCH_MP_TAC REACH_STABLE_SHIFT
        THEN SUBST1_TAC
             (prove("((q' AND a) AND q) = ((q' AND q) AND (a:^Pred))",
                     pred_CYAN_TAC))
        THEN ASM_REWRITE_TAC[]
        THEN MATCH_MP_TAC REACH_STABLE_PSP
        THEN IMP_RES_TAC REACH_IMP_STABLE
        THEN ASM_REWRITE_TAC
             [prove("(a AND (J AND a)) = (J AND (a:^Pred))",
                     pred_CYAN_TAC)] ;
        %-- stable part --%
        ASM_REWRITE_TAC
             [prove("((q' AND a) AND (q AND J)) = 
                     ((q':^Pred) AND (q AND (J AND a)))",
                     pred_CYAN_TAC)]
     ]) ;;

%-------------------------------------------------------------------------------
              Round decomposition in terms of dag
-------------------------------------------------------------------------------%

let Dom = ":*A->bool" ;;
let MonoRel = ":*A->*A->bool" ;;

let lemma1 = prove
  ("!(A:^Dom). ((TT:^Pred) = (!!i::A. K TT i))",
    REWRITE_TAC [K_THM] THEN pred_CYAN_TAC) ;;

let lemma2 = prove
  ("FDAG (A:^Dom) U ==> CHF_FINITE A", 
    REWRITE_TAC [FDAG; FNW] THEN TAUT_TAC) ;;

let lemma3 = prove
  ("A a /\
    (!y::A:^Dom. CON Pr J (!!x::(\x. TC U x y). Q x) (Q y))
    ==> UNITY (Pr:^Uprog) /\ STABLE Pr J",
   RESTRICT_DEF_TAC
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC (REWRITE_RULE [SUBREL] CON_IMP_REACH)
   THENL
   [ IMP_RES_TAC REACH_IMP_UNITY ;
     IMP_RES_TAC REACH_IMP_STABLE ]) ;;

let lemma4 = prove
  ("FDAG (A:^Dom) U ==> NW A (TC U)",
    REWRITE_TAC [FDAG; FNW]
    THEN REPEAT STRIP_TAC THEN IMP_RES_TAC TC_PRESERVES_NW) ;; 

  
let CON_BY_WF = prove_thm
  (`CON_BY_WF`,
   "!(A:^Dom) U (Pr:^Uprog) Pr J Q.
         FDAG A U /\ (?a. A a) /\
         (!y::A. CON Pr J (!!x::(\x. TC U x y). Q x) (Q y))
         ==>
         CON Pr J TT (!!y::A. Q y)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC lemma2
   THEN SUBGOAL_THEN "UNITY (Pr:^Uprog) /\ STABLE Pr J"
        ASSUME_TAC
   THENL [ IMP_RES_TAC lemma3 THEN ASM_REWRITE_TAC[]; ALL_TAC ]
   THEN SUBST1_TAC (SPEC "A:^Dom" lemma1)
   THEN MATCH_MP_TAC CON_CONJ
   THEN ASM_REWRITE_TAC[]
   THEN RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
   THEN CONV_TAC (MK_ABS_CONV "x:*A")
   THEN MATCH_MP_TAC (QUAFF_RULE 3 FDAG_WF_IND)
   THEN EXISTS_TAC "A:^Dom" THEN EXISTS_TAC "U:^MonoRel"
   THEN BETA_TAC THEN ASM_REWRITE_TAC[] 
   THEN REWRITE_TAC[K_THM] 
   THEN XRULE_ASSUM_TAC (QUAFF_RULE 3)
   THEN REPEAT STRIP_TAC
   THEN MATCH_MP_TAC CON_TRANS
   THEN EXISTS_TAC "(!!i:*A:: \j. TC U j x'. (Q i):^Pred)"
   THEN CONJ_TAC
   THENL [ ALL_TAC ; RES_TAC ]
   THEN SUBST1_TAC (SPEC "\j:*A. TC U j x'" lemma1)
   THEN MATCH_MP_TAC CON_CONJ
   THEN ASM_REWRITE_TAC[]
   THEN RESTRICT_DEF_TAC THEN ASM_REWRITE_TAC [K_THM]
   THEN IMP_RES_TAC lemma4
   THEN XRULE_ASSUM_TAC (REWRITE_RULE [NW])
   THEN MATCH_MP_TAC SUBSET_CHF_FIN
   THEN EXISTS_TAC "A:^Dom" 
   THEN BETA_TAC THEN ASM_REWRITE_TAC[]
   THEN REPEAT STRIP_TAC THEN RES_TAC) ;;

%-------------------------------------------------------------------------------
          Round decomposition in terms of well-founded relation
-------------------------------------------------------------------------------%

let lemma1 = prove
   ("CHF_FINITE (CHF V) = FINITE (V:(*)set)",
     REWRITE_TAC [CHF_FINITE; set_ISO_DEF]) ;;

let lemma2 = prove
   ("(!!i:*A::W. TT) = (TT:^Pred)", pred_CYAN_TAC) ;;

let lemma3 = 
    GEN_ALL (REWRITE_RULE [lemma2] (BETA_RULE
    (ISPECL ["W:*A->bool"; "Pr:^Uprog"; "J:^Pred"; "\i:*A. (TT:^Pred)"]
     CON_CONJ))) ;;

let lemma4 = prove
  ("x IN A /\ (U:^MonoRel) x y = x IN (A INTER {t | U t y})",
    REWRITE_TAC [IN_INTER]
    THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
    THEN REFL_TAC) ;;
    
let CON_BY_sWF = prove_thm
  (`CON_BY_sWF`,
   "!(A:(*A)set) U (Pr:^Uprog) Pr J Q.
         sWF A U /\ ~(A={}) /\ FINITE A /\
         (!y::(\y. y IN A). CON Pr J (!!x::(\x. x IN A /\ U x y). Q x) (Q y))
         ==>
         CON Pr J TT (!!y::(\y. y IN A). Q y)",
   REWRITE_TAC [SYM (SPEC_ALL MEMBER_NOT_EMPTY)]
   THEN RESTRICT_DEF_TAC
   THEN REPEAT STRIP_TAC
   THEN SUBGOAL_THEN "UNITY (Pr:^Uprog)" ASSUME_TAC
   THENL [ RES_TAC THEN UNDISCH_ALL_TAC
           THEN REWRITE_TAC [CON] THEN REPEAT STRIP_TAC 
           THEN IMP_RES_TAC REACH_IMP_UNITY ;
           ALL_TAC ]
   THEN SUBGOAL_THEN "STABLE (Pr:^Uprog) J" ASSUME_TAC
   THENL [ RES_TAC THEN UNDISCH_ALL_TAC
           THEN REWRITE_TAC [CON] THEN REPEAT STRIP_TAC 
           THEN IMP_RES_TAC REACH_IMP_STABLE ;
           ALL_TAC ]
   THEN MATCH_MP_TAC lemma3
   THEN REWRITE_TAC [IN_DEF]
   THEN (CONV_TAC o DEPTH_CONV) ETA_CONV THEN ASM_REWRITE_TAC [lemma1]
   THEN RESTRICT_DEF_TAC
   THEN (CONV_TAC o RAND_CONV o ABS_CONV o RAND_CONV) (MK_ABS_CONV "x:*A")
   THEN REWRITE_TAC [SYM (SPEC_ALL IN_DEF)]
   THEN XRULE_ASSUM_TAC (REWRITE_RULE [sWF_EQU_IND; ADMIT_sWF_Ind])
   THEN FIRST_ASSUM (\thm. MATCH_MP_TAC thm THEN DEL_TAC (concl thm))
   THEN BETA_TAC THEN REPEAT STRIP_TAC 
   THEN MATCH_MP_TAC CON_TRANS
   THEN EXISTS_TAC "!! x':*A:: \x'. x' IN A /\ U x' (y:*A). (Q x'):^Pred"
   THEN CONJ_TAC
   THENL [ ALL_TAC ; RES_TAC ]
   THEN ASM_CASES_TAC "(\x':*A. x' IN A /\ U x' (y:*A)) = FF"
   THENL [ ASM_REWRITE_TAC [qAND_EMPTY]
           THEN MATCH_MP_TAC CON_REFL
           THEN ASM_REWRITE_TAC [CONF_TT; pAND_UNIT] ;
           ALL_TAC ]
   THEN MATCH_MP_TAC (GEN_ALL lemma3)
   THEN RESTRICT_DEF_TAC THEN ASM_REWRITE_TAC [lemma4; IN_DEF]
   THEN (CONV_TAC o DEPTH_CONV) ETA_CONV
   THEN REWRITE_TAC [lemma1] 
   THEN IMP_RES_TAC INTER_FINITE THEN ASM_REWRITE_TAC[]
   ) ;;

close_theory() ;;

%---------------------------------------------------------------------------------
   Author:

     Wishnu Prasetya
     Dept. of Computer Science
     Utrecht University
     the Netherlands

     email: wishnu@cs.ruu.nl
---------------------------------------------------------------------------------% 
