loadf `predicate` ;;
load_theory `TDC` ;;
new_theory `dag`;;

let Autoload thy =    
    let defs = map fst (definitions thy) in
        map (\name. autoload_theory(`definition`,thy,name)) defs;
    let thms = map fst (theorems thy) in
        map (\name. autoload_theory(`theorem`,thy,name)) thms; 
    delete_cache thy; ();;

Autoload `relation` ;;
Autoload `TDC` ;;
Autoload `WF`;;
loadf `OLD_RES`;;
let MonoRel = ":*A->*A->bool" ;;
let Dom = ":*A->bool" ;;

%----------------------------------------------------------------------------------

   * A set A and relation U forms a network if U is restricted in A.

   * A network (A,U) is finite if A is finite (consequently so is U)

   * a finite dag is a finite, and anti-reflexive (thus no cycle) network

----------------------------------------------------------------------------------%

let NW = new_definition
  (`NW`,
   "NW (A:^Dom) U = !x y. U x y ==> A x /\ A y") ;;

let FNW = new_definition
  (`FNW`,
   "FNW (A:^Dom) U = NW A U /\ CHF_FINITE A") ;;

let FDAG = new_definition
  (`FDAG`,
   "FDAG (A:^Dom) U = FNW A U /\ ANTI_REFL (TC U)") ;;

%----------------------------------------------------------------------------------
                a finite dag admits a well founded induction
----------------------------------------------------------------------------------%

let MK_ABS_CONV x f = SYM(BETA_CONV "(\^x. ^f) ^x") ;;

let MK_2ABS_CONV x y f = 
    SYM((DEPTH_CONV BETA_CONV) "(\^x ^y. ^f) ^x ^y") ;;

let lemma = prove("(!x:*. p \/ q ==> r) ==> (!x:*. q ==> r)", 
       REPEAT STRIP_TAC THEN RES_TAC) ;;

let MY_TAC =  
   UNDISCH_TAC "?x.((x = a) \/ V x) /\ (U:^MonoRel) x y"
   THEN REPEAT STRIP_TAC
   THENL %2%
   [ OLD_RES_TAC
     THEN UNDISCH_TAC "?x.((x = a) \/ V x) /\ (U:^MonoRel) x x'"
     THEN REPEAT STRIP_TAC
     THENL %2%
     [ UNDISCH_TAC "(U:^MonoRel) x'' x'" THEN ASM_REWRITE_TAC[] ;
       EXISTS_TAC "x'':*A" THEN ASM_REWRITE_TAC[] THEN OLD_RES_TAC ] ;
       ALL_TAC ] ;;

let FNW_WF_lemma1 = prove
  ("!(A:^Dom) U.
        CHF_FINITE A /\ NW A U /\ TRANS U /\ (!x. ~U x x)
        ==>
        !P::CHF_FINITE. 
            (!x. P x ==> A x) /\ (!y. P y ==> (?x. P x /\ U x y)) 
             ==> (!x. ~P x)",
    RESTRICT_DEF_TAC
    THEN REWRITE_TAC [NW; TRANS_DEF] 
    THEN REPEAT_FIN 4 STRIP_TAC
    THEN (CONV_TAC o RAND_CONV) (MK_ABS_CONV "x:^Dom")
    THEN MATCH_MP_TAC CHF_INDUCTION
    THEN REWRITE_TAC [FF_DEF]
    THEN BETA_TAC THEN BETA_TAC
    THEN REWRITE_TAC[]
    THEN REPEAT STRIP_TAC
    THEN UNDISCH_TAC 
         "(!x:*A. V x ==> A x) /\ (!y. V y ==> (?x. V x /\ U x y))
                  ==> (!x. ~V x)"
    THEN REWRITE_TAC[]
    THEN REWRITE_TAC [ NOT_IMP]
    THEN (CONV_TAC o DEPTH_CONV) NOT_FORALL_CONV
    THEN REPEAT STRIP_TAC 
    THEN OLD_RES_TAC
    THENL %4%
    [ %-- 1 --%
      MY_TAC THEN EXISTS_TAC "x':*A" THEN ASM_REWRITE_TAC[] ;
      %-- 2 --%
      RES_TAC
      THENL 
      [ UNDISCH_TAC "(U:^MonoRel) x' x" THEN ASM_REWRITE_TAC[] ;
        EXISTS_TAC "x':*A" THEN ASM_REWRITE_TAC[] ] ;
      %-- 3 --%
      MY_TAC THEN EXISTS_TAC "x':*A" THEN ASM_REWRITE_TAC[] ;
      %-- 4 --%
      EXISTS_TAC "x:*A" THEN ASM_REWRITE_TAC[]]
   ) ;;
 

let SUBSET_CHF_FIN = prove_thm
  (`SUBSET_CHF_FIN`,
   "!(A:^Dom) B.
       CHF_FINITE A /\ (!x. B x ==> A x) ==> CHF_FINITE B",
    REWRITE_TAC [CHF_FINITE]  
    THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC (QUAFF_RULE 3 SUBSET_FINITE)
    THEN EXISTS_TAC "SPEC (A:^Dom)"
    THEN ASM_REWRITE_TAC [SUBSET_DEF; IN_DEF; CHF_SPEC_lemma]) ;;


let FNW_WF_lemma3 = prove
  ("!(A:^Dom) U. 
       CHF_FINITE A /\ NW A U /\
      (!P::CHF_FINITE. 
            (!x. P x ==> A x) /\ (!y. P y ==> (?x. P x /\ U x y)) 
             ==> (!x. ~P x))
       ==>
       (!P. (!y. A y /\ P y ==> (?x. P x /\ U x y)) ==> (!x. A x ==> ~P x))",
    RESTRICT_DEF_TAC
    THEN REPEAT STRIP_TAC
    THEN XRULE_ASSUM_TAC (SPEC "(\s:*A. P s /\ A s)")
    THEN XRULE_ASSUM_TAC (BETA_RULE)
    THEN SUBGOAL_THEN "!x:*A. ~(P x /\ A x)" ASSUME_TAC
    THENL
    [ FIRST_ASSUM (MATCH_MP_TAC o (QUAFF_RULE 3))
      THEN REPEAT STRIP_TAC
      THENL 
      [ MATCH_MP_TAC SUBSET_CHF_FIN
        THEN BETA_TAC
        THEN EXISTS_TAC "A:^Dom" THEN ASM_REWRITE_TAC[]
        THEN REPEAT STRIP_TAC ;
        XRULE_ASSUM_TAC (REWRITE_RULE [NW])
        THEN RES_TAC
        THEN EXISTS_TAC "x':*A"
        THEN ASM_REWRITE_TAC[] THEN RES_TAC ] ;
      RES_TAC
      ]) ;;
  
let FNW_WF_lemma4 = prove
  ("!(A:^Dom) U. 
       (!P. (!y. A y /\ P y ==> (?x. P x /\ U x y)) ==> (!x. A x ==> ~P x))
       ==> 
       (!P. (!y::A. (!x. U x y ==> P x) ==> P y) ==> (!x::A. P x))",
    RESTRICT_DEF_TAC
    THEN REPEAT GEN_TAC
    THEN CONV_TAC CONTRAPOS_CONV
    THEN (CONV_TAC o DEPTH_CONV) NOT_FORALL_CONV
    THEN REWRITE_TAC [NOT_IMP]
    THEN (CONV_TAC o DEPTH_CONV) NOT_FORALL_CONV
    THEN REPEAT STRIP_TAC
    THEN EXISTS_TAC "(\s:*A. ~P s)"
    THEN BETA_TAC THEN REPEAT STRIP_TAC
    THENL 
    [ ONCE_REWRITE_TAC [CONJ_SYM]
      THEN REWRITE_TAC [SYM (SPEC_ALL NOT_IMP)]
      THEN RES_TAC
      THEN XRULE_ASSUM_TAC (CONV_RULE CONTRAPOS_CONV)
      THEN RES_TAC
      THEN XRULE_ASSUM_TAC (CONV_RULE NOT_FORALL_CONV)
      THEN ASM_REWRITE_TAC[] ;
      EXISTS_TAC "x:*A" THEN ASM_REWRITE_TAC[]]) ;;


let TC_PRESERVES_NW = prove_thm
  (`TC_PRESERVES_NW`,
   "!(A:^Dom) U. NW A U ==> NW A (TC U)",
   REWRITE_TAC [NW]
   THEN GEN_TAC THEN GEN_TAC THEN STRIP_TAC
   THEN GEN_TAC THEN GEN_TAC
   THEN (CONV_TAC o RAND_CONV) (MK_2ABS_CONV "x:*A" "y:*A")
   THEN REWRITE_TAC [TC_DEF; SUBREL; TRANS_DEF] 
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN BETA_TAC THEN ASM_REWRITE_TAC[]
   THEN TAUT_TAC) ;;


let FDAG_WF_IND = prove_thm
  (`FDAG_WF_IND`,
   "!(A:^Dom) U.
         FDAG A U ==>
         (!P. (!y::A. (!x. TC U x y ==> P x) ==> P y) ==> (!x::A. P x))",
    REWRITE_TAC [FDAG; FNW; ANTI_REFL]
    THEN REPEAT_FIN 3 STRIP_TAC
    THEN MATCH_MP_TAC FNW_WF_lemma4
    THEN MATCH_MP_TAC FNW_WF_lemma3
    THEN REPEAT CONJ_TAC
    THENL
    [ ASM_REWRITE_TAC[] ;
      IMP_RES_TAC TC_PRESERVES_NW ;
      MATCH_MP_TAC FNW_WF_lemma1
      THEN ASM_REWRITE_TAC [] THEN REPEAT CONJ_TAC
      THENL [IMP_RES_TAC TC_PRESERVES_NW ; REWRITE_TAC [TC_TRANS]]]) ;;

close_theory () ;;
