%---------------------------------------------------------------------------------
   This file contains ML tools (tactics, conversions, etc) to support the 
   theory of operator REACH (defined in mk_reach.ml and stored in reach.th).

   Author:

     Wishnu Prasetya
     Dept. of Computer Science
     Utrecht University
     the Netherlands

     email: wishnu@cs.ruu.nl
---------------------------------------------------------------------------------% 

%---------------------------------------------------------------------------------
           MK_2ABS_CONV x y f = |- f = (\x y. f) x y
           MK_3ABS_CONV, analogous.
---------------------------------------------------------------------------------% 

let MK_2ABS_CONV x y f =
    SYM ((DEPTH_CONV BETA_CONV) "(\^x ^y. ^f) ^x ^y") ;;

let MK_3ABS_CONV x y z f =
    SYM ((DEPTH_CONV BETA_CONV) "(\^x ^y ^z. ^f) ^x ^y ^z") ;;

let REACH_ENS_LIFT_TAC =
    let thm = REWRITE_RULE [SUBREL; B_ENS] REACH_ENS_LIFT in
    MATCH_MP_TAC (QUAFF_RULE 3 thm)
    THEN ASM_REWRITE_TAC[] THEN REPEAT CONJ_TAC ;;

let REACH_TRANS_TAC trm =
    MATCH_MP_TAC (REWRITE_RULE [TRANS_DEF] REACH_TRANS)
    THEN EXISTS_TAC trm 
    THEN ASM_REWRITE_TAC[] THEN REPEAT CONJ_TAC ;;

let REACH_LDISJ_TAC =
    MATCH_MP_TAC (REWRITE_RULE [LDISJ_DEF] REACH_LDISJ)
    THEN ASM_REWRITE_TAC[] THEN REPEAT CONJ_TAC ;;

let REACH_INDUCT1_TAC = 
    REPEAT GEN_TAC
    THEN MATCH_MP_TAC REACH_INDUCT1
    THEN REWRITE_TAC [SUBREL; B_ENS; TRANS_DEF; LDISJ_DEF]
    THEN BETA_TAC 
    THEN REPEAT STRIP_TAC ;;

let REACH_INDUCT3_TAC = 
   REPEAT GEN_TAC
   THEN MATCH_MP_TAC REACH_INDUCT3 THEN BETA_TAC
   THEN REPEAT STRIP_TAC ;;

let REACH_MONO_TAC p =
    MATCH_MP_TAC REACH_MONO
    THEN EXISTS_TAC p
    THEN ASM_REWRITE_TAC[] THEN REPEAT CONJ_TAC ;;

let REACH_ANTIMONO_TAC p =
    MATCH_MP_TAC REACH_ANTIMONO
    THEN EXISTS_TAC p
    THEN ASM_REWRITE_TAC[] THEN REPEAT CONJ_TAC ;;

let REACH_GEN_LDISJ_TAC f =
   let thm = (GEN_ALL o BETA_RULE o (ISPEC f) o (GEN "f:*ind->^Pred") o
              SPEC_ALL)  REACH_GEN_LDISJ  in
   MATCH_MP_TAC thm
   THEN ASM_REWRITE_TAC[] THEN REPEAT CONJ_TAC ;;

let REACH_SIMPLE_DISJ_TAC =
    MATCH_MP_TAC REACH_SIMPLE_DISJ
    THEN ASM_REWRITE_TAC[]
    THEN REPEAT CONJ_TAC ;;

let REACH_CANCEL_TAC r =
    MATCH_MP_TAC REACH_CANCEL
    THEN EXISTS_TAC r THEN ASM_REWRITE_TAC[]
    THEN REPEAT CONJ_TAC ;;

let REACH_PSP_TAC =
    MATCH_MP_TAC REACH_PSP
    THEN ASM_REWRITE_TAC [] THEN REPEAT CONJ_TAC ;;

%---------------------------------------------------------------------------------
     Written by:
     Wishnu Prasetya
     Dept. of Computer Science
     Utrecht University
     the Netherlands

     email: wishnu@cs.ruu.nl
---------------------------------------------------------------------------------%