%----------------------------------------------------------------------------------

  This theory provides a number of basic properties (as in CM88) of the ENSURES
  operator.

  Main results:

    - monotonicity of ENSURES
    - ENSURES includes IMP
    - conjunctivity and PSP law for ENSURES
    - imposibility to progress to FF

  Author: Wishnu Prasetya
  Dept. of Computer Science
  Utrecht University
  the Netherlands
  email: wishnu@cs.ruu.nl
  
  requires libraries: WF (and relation, predicate, and sets)

----------------------------------------------------------------------------------% 

load_library `WF` ;;
load_theory `unless`;;
new_theory `ensures`;;
loadf `my_misc` ;;
loadf `OLD_RES` ;;

let Autoload thy =     
    let defs = map fst (definitions thy) in
        map (\name. autoload_theory(`definition`,thy,name)) defs;
    let thms = map fst (theorems thy) in
        map (\name. autoload_theory(`theorem`,thy,name)) thms; 
    delete_cache thy; ();;

Autoload `unless` ;;
Autoload `def_UNITY` ;;

let State = ":*var->*value" ;;
let Pred = ":(*var->*value)->bool" ;;
let Action = ":(*var->*value)->(*var->*value)->bool" ;;
let ACTIONS = ":((*var->*value)->(*var->*value)->bool)->bool" ;;
let VARS = ":*var->bool" ;;
let Uprog = ":^ACTIONS # ^Pred # ^VARS # ^VARS" ;;

let ENSURES_TAC =
    RESTRICT_DEF_TAC
    THEN REWRITE_TAC [ENSURES] THEN REPEAT STRIP_TAC  
    THEN ASM_REWRITE_TAC[] ;;

let HOA_TAC =
  XRULE_ASSUM_TAC RESTRICT_DEF_RULE
  THEN RESTRICT_DEF_TAC
  THEN UNDISCH_ALL_TAC 
  THEN REWRITE_TAC [HOA_DEF] THEN REPEAT STRIP_TAC ;;

%-----------------------------------------------------------------------------
    An ENSURES expression is only valid for a UNITY program
-----------------------------------------------------------------------------%

let ENSURES_IMP_UNITY = prove_thm
  (`ENSURES_IMP_UNITY`,
    "!(Pr:^Uprog) p q. ENSURES Pr p q ==> UNITY Pr",
    REWRITE_TAC[ENSURES]
    THEN TAUT_TAC) ;;

%-----------------------------------------------------------------------------
    reflexivity of ENSURES
-----------------------------------------------------------------------------%

let ENSURES_REFL = prove_thm
   (`ENSURES_REFL`, "!Pr:^Uprog:: UNITY. !p. ENSURES Pr p p",
   ENSURES_TAC THENL 
   [    MATCH_ACCEPT_TAC UNLESS_REFL ;
        UNDISCH_TAC "UNITY (x:^Uprog)"
        THEN SUBST1_TAC (SPEC "x:^Uprog" ID_UPROG)
        THEN REWRITE_TAC [UNITY; ALWAYS_ENABLED; PROG; HOA_DEF]
        THEN RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
        THEN EXISTS_TAC "A:^Action"
        THEN ASM_REWRITE_TAC[]
        THEN pred_JADE_TAC 5 ]) ;;

%-----------------------------------------------------------------------------
    monotonicity ENSURES
-----------------------------------------------------------------------------%

let ENSURES_CONSQ_WEAK = prove_thm
  (`ENSURES_CONSQ_WEAK`,
   "!(Pr:^Uprog) p q r.
         ENSURES  Pr p q /\ (|== (q  IMP r)) ==> (ENSURES Pr p r)",
   ENSURES_TAC
   THENL [    IMP_RES_TAC UNLESS_CONSQ_WEAK ;           
              HOA_TAC
              THEN EXISTS_TAC "x:^Action" THEN ASM_REWRITE_TAC[]
              THEN UNDISCH_ALL_TAC THEN pred_JADE_TAC 1
              THEN ASM_CASES_TAC "(q:^Pred) s"
              THEN REPEAT RES_TAC] );;

%-----------------------------------------------------------------------------
    It is not possible to progress (by ENSURES) to FF unless from FF itself.
-----------------------------------------------------------------------------%

let ENSURES_IMPOS = prove_thm
  (`ENSURES_IMPOS`,
   "!(Pr:^Uprog) p. ENSURES Pr p FF ==> (|== (NOT p))",
    ENSURES_TAC 
    THEN XRULE_ASSUM_TAC (SUBS [SPEC "Pr:^Uprog" ID_UPROG])
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [UNITY; ALWAYS_ENABLED; PROG])
    THEN HOA_TAC
    THEN UNDISCH_ALL_TAC THEN pred_JADE_TAC 1
    THEN RULE_ASSUM_TAC (\thm. if (concl thm) = "!s. ?t. (x:^Action) s t"
                               then (SPEC "s:^State" thm) else thm)
    THEN UNDISCH_TAC "?t. (x:^Action) s t" THEN STRIP_TAC
    THEN RES_TAC) ;;

%-----------------------------------------------------------------------------
    The Progress Safety Progress (PSP) principle for ENSURES
-----------------------------------------------------------------------------%

let ENSURES_PSP = prove_thm 
   (`ENSURES_PSP`,
   "!(Pr:^Uprog) p q p' q'.
          UNLESS Pr p q  /\  ENSURES Pr p' q' ==>
          ENSURES Pr (p AND p') (((p AND q') OR (p' AND q)) OR (q AND q'))",
    ENSURES_TAC THENL
    [     IMP_RES_TAC UNLESS_CONJ  ;
          XRULE_ASSUM_TAC (REWRITE_RULE [UNLESS])
          THEN HOA_TAC
          THEN EXISTS_TAC "x:^Action"
          THEN ASM_REWRITE_TAC [] THEN UNDISCH_ALL_TAC
          THEN pred_JADE_TAC 0 THEN OLD_RES_TAC
          THEN XRULE_ASSUM_TAC (REWRITE_RULE [prove("(p==>F) = ~p",TAUT_TAC)])
          THEN XRULE_ASSUM_TAC SPEC_ALL
          THEN OLD_RES_TAC THEN ASM_REWRITE_TAC[] ]) ;;

%-----------------------------------------------------------------------------
    ENSURES Conjunction law
-----------------------------------------------------------------------------%


let ENSURES_CONJ = prove_thm 
   (`ENSURES_CONJ`,
   "!(Pr:^Uprog) p q p' q'.
    ENSURES Pr p q /\ ENSURES Pr p' q' ==>
    ENSURES Pr (p AND p') (((p AND q') OR (p' AND q)) OR (q AND q'))",
     REPEAT STRIP_TAC
     THEN ASM_TAC (CONJUNCT1 o CONJUNCT2 o REWRITE_RULE[ENSURES])
     THEN IMP_RES_TAC ENSURES_PSP) ;;

%-----------------------------------------------------------------------------
    some disjunctivity property
-----------------------------------------------------------------------------%

let ENSURES_DISJ = prove_thm 
   (`ENSURES_DISJ`,
    "!(Pr:^Uprog) p q r.
           ENSURES Pr p q ==> ENSURES Pr (p OR r) (q OR r)",
    ENSURES_TAC THENL
    [     ASSUME_TAC (SPECL ["Pr:^Uprog" ; "r:^Pred" ] UNLESS_REFL)
          THEN IMP_RES_TAC UNLESS_SIMPLE_DISJ ;
          HOA_TAC
          THEN EXISTS_TAC "x:^Action" THEN ASM_REWRITE_TAC[]
          THEN UNDISCH_ALL_TAC THEN pred_JADE_TAC 0
          THEN XRULE_ASSUM_TAC (REWRITE_RULE [(CONJUNCT2 o SPEC_ALL) DE_MORGAN_THM])
          THENL [ XRULE_ASSUM_TAC CONJUNCT1 ; XRULE_ASSUM_TAC CONJUNCT2]
          THEN RES_TAC THEN ASM_REWRITE_TAC[] ]) ;;

%-----------------------------------------------------------------------------
    ENSURES includes IMP, and some corollaries
-----------------------------------------------------------------------------%

let ENSURES_IMP_LIFT = prove_thm
  (`ENSURES_IMP_LIFT`,
   "!Pr:^Uprog::UNITY. !p q. (|== (p IMP q)) ==> ENSURES Pr p q",
    RESTRICT_DEF_TAC
    THEN (REPEAT STRIP_TAC)
    THEN OLD_IMP_RES_TAC (RESTRICT_SHIFT ENSURES_REFL)
    THEN IMP_RES_TAC ENSURES_CONSQ_WEAK) ;;

let ENSURES_cor2 = prove_thm
  (`ENSURES_cor2`,
   "!(Pr:^Uprog) p q r.
       ENSURES Pr (p OR q) r ==> ENSURES Pr p (q OR r)",
    ENSURES_TAC THENL
    [     IMP_RES_TAC UNLESS_cor4 ;
          HOA_TAC
          THEN EXISTS_TAC "x:^Action" THEN ASM_REWRITE_TAC[]
          THEN UNDISCH_ALL_TAC THEN pred_JADE_TAC 0
          THEN XRULE_ASSUM_TAC (REWRITE_RULE [(CONJUNCT2 o SPEC_ALL) DE_MORGAN_THM])
          THEN XRULE_ASSUM_TAC CONJUNCT2
          THEN RES_TAC THEN ASM_REWRITE_TAC[] ]) ;;

let ENSURES_cor3 = prove_thm
  (`ENSURES_cor3`,
   "!(Pr:^Uprog) p q r.
       ENSURES Pr p (q OR r) ==> ENSURES Pr (p AND (NOT q)) (q OR r)",
    ENSURES_TAC THENL
    [     ASM_REWRITE_TAC [SYM(SPEC_ALL UNLESS_cor3)] ;
          HOA_TAC
          THEN EXISTS_TAC "x:^Action" THEN ASM_REWRITE_TAC[]
          THEN UNDISCH_ALL_TAC THEN pred_JADE_TAC 0
          THEN XRULE_ASSUM_TAC (REWRITE_RULE [(CONJUNCT2 o SPEC_ALL) DE_MORGAN_THM])
          THEN XRULE_ASSUM_TAC CONJUNCT2
          THEN OLD_RES_TAC THEN ASM_REWRITE_TAC[] ]) ;;

%-----------------------------------------------------------------------------
    Compositionality of ENSURES wrt PAR
-----------------------------------------------------------------------------%

let ENSURES_PAR = prove_thm (`ENSURES_PAR`,
   "!(Pr:^Uprog)::UNITY. !p q Qr.
    UNLESS Pr p q /\ ENSURES Qr p q ==> ENSURES (Pr PAR Qr) p q",
    REWRITE_TAC [ENSURES; PROG; pOR_DEF] 
    THEN RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
    THENL [ IMP_RES_TAC UNITY_PAR ;
            IMP_RES_TAC UNLESS_PAR ;
            REWRITE_TAC [PAR; pOR_DEF; PROG]
            THEN BETA_TAC
            THEN EXISTS_TAC "x':^Action"
            THEN ASM_REWRITE_TAC[] ]) ;;

close_theory() ;;
