%----------------------------------------------------------------------------------

  This theory provides a number of basic properties (as in CM88) of the LEADSTO
  operator.

  Main results are:

    - lifting ENSURES and ==> to LEADSTO
    - reflexivity, transitivity, disjunctivity, and monotonicity of LEADSTO
    - LEADSTO induction theorems
    - Cancelation Law
    - Progess Sagfety Progress (PSP) Law
    - Bounded Progress Law
    - Completion Law

  Author: Wishnu Prasetya
  Dept. of Computer Science
  Utrecht University
  the Netherlands
  email: wishnu@cs.ruu.nl

  requires libraries: WF (and relation, predicate, and sets)

----------------------------------------------------------------------------------% 

load_library `WF` ;;
load_theory `ensures` ;;
new_theory `leadsto`;;
loadf `OLD_RES` ;;
loadf `my_misc` ;;

let Autoload thy =     
    let defs = map fst (definitions thy) in
        map (\name. autoload_theory(`definition`,thy,name)) defs;
    let thms = map fst (theorems thy) in
        map (\name. autoload_theory(`theorem`,thy,name)) thms; 
    delete_cache thy; ();;

Autoload `relation` ;;
Autoload `def_UNITY` ;;
Autoload `unless` ;;
Autoload `ensures` ;;
Autoload `WF` ;;
Autoload `TDC` ;;

let State = ":*var->*value" ;;
let Pred = ":(*var->*value)->bool" ;;
let Action = ":(*var->*value)->(*var->*value)->bool" ;;
let ACTIONS = ":((*var->*value)->(*var->*value)->bool)->bool" ;;
let VARS = ":*var->bool" ;;
let Uprog = ":^ACTIONS # ^Pred # ^VARS # ^VARS" ;;

let MK_2ABS_CONV x y f =
    SYM ((DEPTH_CONV BETA_CONV) "(\^x ^y. ^f) ^x ^y") ;;

%----------------------------------------------------------------------------------
       an object satisfying an leadsto property is a UNITY program.
----------------------------------------------------------------------------------%   

let LEADSTO_IMP_UNITY = prove_thm
  (`LEADSTO_IMP_UNITY`,
   "!(Pr:^Uprog) p q. LEADSTO Pr p q ==> UNITY Pr",
    REWRITE_TAC [LEADSTO; TDC_DEF]
    THEN REPEAT STRIP_TAC
    THEN CONV_TAC (MK_2ABS_CONV "p:^Pred" "q:^Pred")
    THEN POP_ASSUM MATCH_MP_TAC
    THEN REWRITE_TAC [SUBREL; TRANS_DEF; LDISJ_DEF; ENSURES]
    THEN RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
    THEN RES_TAC) ;;

%----------------------------------------------------------------------------------
                 The relation LEADSTO includes ENSURES
----------------------------------------------------------------------------------% 

let LEADSTO_ENS_LIFT_thm = prove
  ("!Pr:^Uprog. (ENSURES Pr) SUBREL (LEADSTO Pr)",
   REWRITE_TAC [LEADSTO] 
   THEN REWRITE_TAC [TDC_LIFT]) ;;

let LEADSTO_ENS_LIFT = prove_thm
  (`LEADSTO_ENS_LIFT`,
   "!(Pr:^Uprog) p q. ENSURES Pr p q ==> LEADSTO Pr p q",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC 
        (REWRITE_RULE [SUBREL] LEADSTO_ENS_LIFT_thm)) ;;

let LEADSTO_ENS_LIFT_TAC = MATCH_MP_TAC LEADSTO_ENS_LIFT ;;

%----------------------------------------------------------------------------------
                 The relation LEADSTO is transitive
----------------------------------------------------------------------------------% 

let LEADSTO_Ri_TRANS_thm = prove
  ("!Pr:^Uprog. Ri_TR (ENSURES Pr) (LEADSTO Pr)",
   REWRITE_TAC [LEADSTO; TDC_Ri_TRANS] ) ;;

let LEADSTO_TRANS_thm = prove
  ("!Pr:^Uprog. TRANS (LEADSTO Pr)",
   REWRITE_TAC [LEADSTO; TDC_TRANS]) ;;

let LEADSTO_TRANS = prove_thm
  (`LEADSTO_TRANS`,
   "!(p:^Pred) r q Pr.
        LEADSTO Pr p r /\ LEADSTO Pr r q ==> LEADSTO Pr p q",
   REWRITE_TAC [REWRITE_RULE [TRANS_DEF] LEADSTO_TRANS_thm]) ;;

let LEADSTO_TRANS_TAC trm =
    MATCH_MP_TAC LEADSTO_TRANS
    THEN EXISTS_TAC trm THEN REPEAT STRIP_TAC ;;

%----------------------------------------------------------------------------------
                 The relation LEADSTO is left-disjunctive
----------------------------------------------------------------------------------% 

let LEADSTO_LDISJ_thm = prove
  ("!Pr:^Uprog. LDISJ (LEADSTO Pr)",
   REWRITE_TAC [LEADSTO] THEN REWRITE_TAC [TDC_LDISJ]) ;;

let LEADSTO_DISJ = prove_thm (`LEADSTO_DISJ`,
   "!W q (Pr:^Uprog).
       ~(W=FF) /\ (!p:: W. LEADSTO Pr p q) ==> LEADSTO Pr (??p::W. p) q",
   REWRITE_TAC 
   [ Non_FF_lemma ;
     (REWRITE_RULE [LDISJ_DEF] LEADSTO_LDISJ_thm)]) ;;

%----------------------------------------------------------------------------------
                 The relation LEADSTO is reflexive
----------------------------------------------------------------------------------% 

let LEADSTO_REFL = prove_thm 
  (`LEADSTO_REFL`,
    "!Pr:^Uprog::UNITY. !p. LEADSTO Pr p p",
    RESTRICT_DEF_TAC
    THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC (RESTRICT_SHIFT ENSURES_REFL)
    THEN XRULE_ASSUM_TAC (SPEC "p:^Pred")
    THEN IMP_RES_TAC (RESTRICT_SHIFT LEADSTO_ENS_LIFT)) ;;

let LEADSTO_REFL_TAC =
    MATCH_MP_TAC (QUAFF_RULE 2 LEADSTO_REFL)
    THEN REPEAT STRIP_TAC ;;

%----------------------------------------------------------------------------------
                 The relation LEADSTO includes IMP
----------------------------------------------------------------------------------% 

let MK_LT_ABS_CONV trm =
    let p = rand(rator(rator trm)) in
    let q = rand(rator trm) in
    SYM ((DEPTH_CONV BETA_CONV) "(\^p ^q. ^trm) ^p ^q") ;;

let LEADSTO_IMPLICATION = prove_thm (`LEADSTO_IMPLICATION`,
    "!Pr:^Uprog::UNITY. 
     !p q. (|== (p IMP q)) ==> (LEADSTO Pr p q)",
    RESTRICT_DEF_TAC
    THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC (RESTRICT_SHIFT ENSURES_REFL)
    THEN XRULE_ASSUM_TAC (SPEC "p:^Pred")
    THEN IMP_RES_TAC (RESTRICT_SHIFT ENSURES_CONSQ_WEAK)
    THEN IMP_RES_TAC (RESTRICT_SHIFT LEADSTO_ENS_LIFT)) ;;

let LEADSTO_IMPLICATION_TAC =
    MATCH_MP_TAC (QUAFF_RULE 2 LEADSTO_IMPLICATION)
    THEN REPEAT CONJ_TAC ;;

%----------------------------------------------------------------------------------
                      LEADSTO monotonicity
----------------------------------------------------------------------------------% 

let LEADSTO_CONSQ_WEAK = prove_thm (`LEADSTO_CONSQ_WEAK`,
   "!(Pr:^Uprog) p q r. 
     LEADSTO Pr p q /\ (|== (q IMP r)) ==> LEADSTO Pr p r",
     REPEAT STRIP_TAC
     THEN IMP_RES_TAC LEADSTO_IMP_UNITY
     THEN IMP_RES_TAC (RESTRICT_SHIFT LEADSTO_IMPLICATION)
     THEN IMP_RES_TAC LEADSTO_TRANS) ;;

let LEADSTO_CONSQ_WEAK_TAC trm = 
    MATCH_MP_TAC LEADSTO_CONSQ_WEAK
    THEN EXISTS_TAC trm 
    THEN REPEAT CONJ_TAC ;;

%----------------------------------------------------------------------------------
                      LEADSTO anti-monotonic 
----------------------------------------------------------------------------------% 

let LEADSTO_ANTE_STRONG = prove_thm (`LEADSTO_ANTE_STRONG`,
   "!(Pr:^Uprog) q r p.
     (LEADSTO Pr q r) /\ (|== (p IMP q)) ==> LEADSTO Pr p r",
     RESTRICT_DEF_TAC
     THEN REPEAT STRIP_TAC
     THEN IMP_RES_TAC LEADSTO_IMP_UNITY
     THEN IMP_RES_TAC (RESTRICT_SHIFT LEADSTO_IMPLICATION)
     THEN IMP_RES_TAC LEADSTO_TRANS) ;;

let LEADSTO_ANTE_STRONG_TAC trm = 
    MATCH_MP_TAC LEADSTO_ANTE_STRONG
    THEN EXISTS_TAC trm 
    THEN REPEAT CONJ_TAC ;;

%----------------------------------------------------------------------------------
                      LEADSTO general disjunction
----------------------------------------------------------------------------------% 

let lemma1 = 
   (BETA_RULE o (ISPECL 
    ["W:*ind->bool"; "\p. LEADSTO (Pr:^Uprog) p (??j:*ind::W. Q j)"]))
    FORALL_RES_CHANGE ;;

let lemma2 = 
   (BETA_RULE o (SPEC "\p:^Pred. ?i:*ind. W i /\ (p = P i)") o
    RESTRICT_DEF_RULE) LEADSTO_DISJ ;;

let LEADSTO_GEN_DISJ = prove_thm 
  (`LEADSTO_GEN_DISJ`,
   "!(Pr:^Uprog) (W:*ind->bool) (P:*ind->(^Pred)) Q.
       ~(W=FF) /\ (!i:*ind:: W. LEADSTO Pr (P i) (Q i))   ==>
       LEADSTO Pr (??i::W. P i) (??i::W. Q i)",
    RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
    THEN SUBGOAL_THEN
         "!i:*ind. W i ==> LEADSTO  (Pr:^Uprog) (P i) (??j::W. Q j)"
         (\thm. ASSUME_TAC thm)
    THENL 
    [ %-- 1  --%
      ASSUME_TAC (ISPECL ["W:*ind->bool"; "Q:*ind->^Pred"] qOR_LUB1)
      THEN XRULE_ASSUM_TAC RESTRICT_SHIFT 
      THEN REPEAT STRIP_TAC THEN RES_TAC
      THEN IMP_RES_TAC (RESTRICT_SHIFT LEADSTO_CONSQ_WEAK)  ;
      %-- 2 --%
      XRULE_ASSUM_TAC (REWRITE_RULE [RESTRICT_DEF_RULE lemma1])
      THEN IMP_RES_TAC (ISPECL ["W:*ind->bool"; "P:*ind->^Pred"] Non_FF_DOM_CHANGE)
      THEN XRULE_ASSUM_TAC (SPEC "P:*ind->^Pred")
      THEN IMP_RES_TAC lemma2
      THEN XRULE_ASSUM_TAC (REWRITE_RULE [(SYM o SPEC_ALL) qOR_RES_CHANGE])
      THEN ASM_REWRITE_TAC[] ]) ;;

let LEADSTO_GEN_DISJ_thm = prove
  ("!(Pr:^Uprog) (W:*ind->bool). DISJ W (LEADSTO Pr)",
    REWRITE_TAC [DISJ_DEF; SYM(SPEC_ALL Non_FF_lemma) ;
                 LEADSTO_GEN_DISJ]) ;;
    
%----------------------------------------------------------------------------------
                      LEADSTO simple disjunction
----------------------------------------------------------------------------------% 

let LEADSTO_SIMPLE_DISJ = prove_thm (`LEADSTO_SIMPLE_DISJ`,
   "!p1 p2 q (Pr:^Uprog). 
        LEADSTO Pr p1 q  /\ LEADSTO Pr p2 q  
        ==> 
        LEADSTO Pr (p1 OR p2) q",
    REPEAT STRIP_TAC
    THEN SUBST1_TAC (SYM(ISPEC "q:^Pred" pOR_REFL))
    THEN MATCH_MP_TAC SIMPLE_DISJ
    THEN ASM_REWRITE_TAC [LEADSTO_GEN_DISJ_thm]) ;;
   
%----------------------------------------------------------------------------------
              a more general left-disjunction theorem for LEADSTO 
----------------------------------------------------------------------------------% 

let LEADSTO_DISJ_LEFT = prove_thm 
  (`LEADSTO_DISJ_LEFT`,
   "!(Pr:^Uprog) (W:*ind->bool) (P:*ind->(^Pred)) q.
       ~(W=FF) /\ (!i:*ind:: W. LEADSTO Pr (P i) q)   ==>
       LEADSTO Pr (??i::W. P i) q",
    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC GEN_LDISJ
    THEN ASM_REWRITE_TAC [LEADSTO_GEN_DISJ_thm]
    THEN ASM_REWRITE_TAC [SYM(SPEC_ALL Non_FF_lemma)]) ;;

%----------------------------------------------------------------------------------
              a right-disjunction theorem for LEADSTO 
----------------------------------------------------------------------------------% 

let LEADSTO_DISJ_RIGHT = prove_thm 
  (`LEADSTO_DISJ_RIGHT`,
  "!(Pr:^Uprog)(W:*ind->bool) p (Q:*ind->(^Pred)).
       ~(W=FF) /\ (!i:*ind:: W. LEADSTO Pr p (Q i))   ==>
       LEADSTO Pr p (??i::W. Q i)",
    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC GEN_RDISJ
    THEN ASM_REWRITE_TAC [LEADSTO_GEN_DISJ_thm]
    THEN ASM_REWRITE_TAC [SYM(SPEC_ALL Non_FF_lemma)]) ;;

%----------------------------------------------------------------------------------
              a simpler general disjunction theorem for LEADSTO 
----------------------------------------------------------------------------------% 

let LEADSTO_SIMPLE_GEN_DISJ = prove_thm (`LEADSTO_SIMPLE_GEN_DISJ`,
   "!(Pr:^Uprog) p1 q1 p2 q2.
        LEADSTO Pr p1 q1 /\ LEADSTO Pr p2 q2   
        ==>
        LEADSTO Pr (p1 OR p2) (q1 OR q2)",
    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC SIMPLE_DISJ
    THEN ASM_REWRITE_TAC [LEADSTO_GEN_DISJ_thm]) ;;

%----------------------------------------------------------------------------------
                        the Cancelation Law
----------------------------------------------------------------------------------% 

let LEADSTO_CANCEL = prove_thm 
  (`LEADSTO_CANCEL`,
   "!(Pr:^Uprog) p b r q. 
         LEADSTO Pr p (q OR b) /\ LEADSTO Pr b r  
         ==> 
         LEADSTO Pr p (q OR r)",
     RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
     THEN MATCH_MP_TAC CANCEL
     THEN EXISTS_TAC "b:^Pred"
     THEN ASM_REWRITE_TAC[LEADSTO_TRANS_thm; LEADSTO_GEN_DISJ_thm]
     THEN IMP_RES_TAC LEADSTO_IMP_UNITY
     THEN IMP_RES_TAC (RESTRICT_SHIFT LEADSTO_REFL)
     THEN ASM_REWRITE_TAC[]) ;;

let LEADSTO_CANCEL_TAC trm = 
    MATCH_MP_TAC LEADSTO_CANCEL THEN EXISTS_TAC trm 
    THEN REPEAT CONJ_TAC ;;

%----------------------------------------------------------------------------------
                  some Induction theorems for LEADSTO
----------------------------------------------------------------------------------% 

let LEADSTO_INDUCT_thm1 = prove_thm 
  (`LEADSTO_INDUCT_thm1`,
   "!X (Pr:^Uprog). 
       (ENSURES Pr) SUBREL X /\ TRANS X /\ LDISJ X
       ==> 
       (!p q. LEADSTO Pr p q ==> X p q)",
    REWRITE_TAC [LEADSTO; TDC_DEF]
    THEN REPEAT STRIP_TAC
    THEN POP_ASSUM MATCH_MP_TAC
    THEN ASM_REWRITE_TAC[]) ;;

let LEADSTO_INDUCT_thm2 = prove_thm 
  (`LEADSTO_INDUCT_thm2`,
   "!X (Pr:^Uprog). 
       (ENSURES Pr) SUBREL X /\ Ri_TR (ENSURES Pr) X /\ LDISJ X
       ==> 
       (!p q. LEADSTO Pr p q ==> X p q)",
    REWRITE_TAC [LEADSTO; TDC_EQU_Ri_TDC; Ri_TDC_DEF]
    THEN REPEAT STRIP_TAC
    THEN POP_ASSUM MATCH_MP_TAC
    THEN ASM_REWRITE_TAC[]) ;;

let LEADSTO_INDUCT_thm3 = prove_thm 
  (`LEADSTO_INDUCT_thm3`,
   "!X (Pr:^Uprog). 
       (ENSURES Pr) SUBREL ((LEADSTO Pr) rINTER X) /\
        TRANS ((LEADSTO Pr) rINTER X) /\
        LDISJ ((LEADSTO Pr) rINTER X)
       ==> 
       (!p q. LEADSTO Pr p q ==> X p q)",
    REWRITE_TAC [LEADSTO; SUBREL] 
    THEN REPEAT GEN_TAC THEN STRIP_TAC
    THEN MATCH_MP_TAC (REWRITE_RULE [SUBREL] TDC_INDUCT2)
    THEN ASM_REWRITE_TAC[]) ;;

let Type_Of_X = ":^Pred->^Pred->^Pred->^Pred->bool";;

let LT_DOUBLE_INDUCT = prove_thm
  (`LT_DOUBLE_INDUCT`,
   "!(X:^Type_Of_X) (Pr:^Uprog). 
       (!p q r s. X p q r s  = X r s p q)                   /\
       (!p q r s. ENSURES Pr p q ==> X p q r s)             /\
       (!W q r s. 
            ~(W=FF) /\ (!p :: W. LEADSTO Pr p q /\  X p q r s) 
            ==> X (?? p :: W. p) q r s)                           /\
       (!p q r u v w.  
            ENSURES Pr p q /\ ENSURES Pr u v /\
            LEADSTO Pr q r /\ LEADSTO Pr v w /\
            X p r v w /\ X u w q r /\ X q r v w 
             ==> X p r u w)  
        ==>
       (!p q r s. LEADSTO Pr p q /\ LEADSTO Pr r s ==> X p q r s)",
    REWRITE_TAC [LEADSTO; TDC_EQU_Ri_TDC; Non_FF_lemma;
                 Ri_TDC_DOUBLE_INDUCT]) ;;

let LEADSTO_INDUCT1_TAC = 
   REPEAT GEN_TAC
   THEN MATCH_MP_TAC LEADSTO_INDUCT_thm1
   THEN REWRITE_TAC [SUBREL; TRANS_DEF; LDISJ_DEF]
   THEN BETA_TAC 
   THEN REPEAT STRIP_TAC ;;

let LEADSTO_INDUCT2_TAC = 
   REPEAT GEN_TAC
   THEN MATCH_MP_TAC LEADSTO_INDUCT_thm2
   THEN REWRITE_TAC [SUBREL; Ri_TR_DEF; LDISJ_DEF]
   THEN BETA_TAC 
   THEN REPEAT STRIP_TAC ;;

let LEADSTO_INDUCT3_TAC = 
   REPEAT GEN_TAC
   THEN MATCH_MP_TAC LEADSTO_INDUCT_thm3
   THEN REWRITE_TAC [SUBREL; TRANS_DEF; LDISJ_DEF; rINTER]
   THEN BETA_TAC 
   THEN REPEAT STRIP_TAC 
   THENL [ IMP_RES_TAC LEADSTO_ENS_LIFT ; ALL_TAC ;
           IMP_RES_TAC LEADSTO_TRANS ; ALL_TAC ;
           MATCH_MP_TAC LEADSTO_DISJ 
           THEN ASM_REWRITE_TAC [Non_FF_lemma]
           THEN CONJ_TAC
           THENL [ EXISTS_TAC "x:^Pred" THEN ASM_REWRITE_TAC[] ;
                   UNDISCH_ALL_TAC
                   THEN RESTRICT_DEF_TAC
                   THEN REPEAT STRIP_TAC THEN RES_TAC ] ;
           ALL_TAC] ;;

let LT_DOUBLE_INDUCT_TAC = 
   REPEAT GEN_TAC
   THEN MATCH_MP_TAC LT_DOUBLE_INDUCT
   THEN BETA_TAC 
   THEN REPEAT STRIP_TAC ;;

%----------------------------------------------------------------------------------
                       the Impossibility Law
----------------------------------------------------------------------------------% 

let FF_EVERY_WHERE_NOT = prove_thm
  (`FF_EVERY_WHERE_NOT`,
   "!p:^Pred. (p=FF) = (|== (NOT p))",
    pred_JADE_TAC 0 
    THEN CONV_TAC (ONCE_DEPTH_CONV FUN_EQ_CONV)
    THEN BETA_TAC
    THEN EQ_TAC 
    THEN REPEAT STRIP_TAC 
    THEN ASM_REWRITE_TAC[]
    THEN RES_TAC) ;;

let LEADSTO_IMPOS_lemma = prove
  ("LEADSTO (Pr:^Uprog) p q ==> (\p q. (q=FF) ==> (|== (NOT p))) p q",
    LEADSTO_INDUCT1_TAC
    THENL
    [ %--- ENSURES case ---%
      RULE_ASSUM_TAC
           (\thm. if is_eq(concl thm) then thm
                  else REWRITE_RULE [ASSUME "(y:^Pred)=FF"] thm)
      THEN IMP_RES_TAC ENSURES_IMPOS ;
      %--- transitivity ---%
      XRULE_ASSUM_TAC (REWRITE_RULE [FF_EVERY_WHERE_NOT])
      THEN REPEAT RES_TAC ;
      %--- disjunctivity ---%
      UNDISCH_ALL_TAC 
      THEN pred_JADE_TAC 5 ]
   ) ;;

let LEADSTO_IMPOS = prove_thm
  (`LEADSTO_IMPOS`,
   "!(Pr:^Uprog) p.
       LEADSTO (Pr:^Uprog) p FF ==> |== (NOT p)",
   REWRITE_TAC [
      (REWRITE_RULE[] o BETA_RULE o SPEC "FF:^Pred" o GEN "q:^Pred")
       LEADSTO_IMPOS_lemma]) ;;

%----------------------------------------------------------------------------------
                                the PSP Law
----------------------------------------------------------------------------------% 

let LEADSTO_PSP = prove_thm(`LEADSTO_PSP`,
    "!(Pr:^Uprog) p q r s.
           LEADSTO Pr p q /\ UNLESS Pr r s ==> 
           LEADSTO Pr (p AND r) ((q AND r) OR s)",
     REPEAT STRIP_TAC
     THEN IMP_RES_TAC LEADSTO_IMP_UNITY
     THEN CONV_TAC (MK_2ABS_CONV "p:^Pred" "q:^Pred")
     THEN UNDISCH_TAC "LEADSTO (Pr:^Uprog) p q"
     THEN LEADSTO_INDUCT1_TAC 
     THENL 
     [ %-- 1 --%
       LEADSTO_TRANS_TAC "((r AND y) OR (x AND s)) OR (s AND (y:^Pred))"
       THENL
       [ LEADSTO_ENS_LIFT_TAC
         THEN SUBST1_TAC (ISPECL ["x:^Pred"; "r:^Pred"] pAND_SYM)
         THEN IMP_RES_TAC ENSURES_PSP ;
         LEADSTO_IMPLICATION_TAC
         THENL [ ASM_REWRITE_TAC[] ; pred_JADE_TAC 3] ] ;
       %-- 2 --%
       RULE_ASSUM_TAC (ONCE_REWRITE_RULE [pOR_SYM])
       THEN IMP_RES_TAC LEADSTO_CANCEL
       THEN ASM_REWRITE_TAC [
            prove("(z AND r) OR (s:^Pred) = (s OR (s OR (z AND r)))",
                  pred_CYAN_TAC)] ;
       %-- 3 --%
       REWRITE_TAC [(BETA_RULE o ISPECL ["W:^Pred->bool"; "\p:^Pred. p"])
                    qAND_RIGHT_DISTR_OR]
       THEN MATCH_MP_TAC 
            ((BETA_RULE o 
              ISPECL ["Pr:^Uprog"; "W:^Pred->bool"; "\p:^Pred. p AND r"])
              LEADSTO_DISJ_LEFT)
       THEN ASM_REWRITE_TAC [Non_FF_lemma]
       THEN EXISTS_TAC "x:^Pred"
       THEN ASM_REWRITE_TAC [] ]) ;;

let LEADSTO_PSP_TAC = 
    MATCH_MP_TAC LEADSTO_PSP THEN REPEAT CONJ_TAC ;;

%----------------------------------------------------------------------------------
                       the  Bounded Progress Law
----------------------------------------------------------------------------------% 

let Metric = ":^State -> *Dom" ;;
let Rel = ":*Dom->*Dom->bool" ;;

let LEADSTO_WF_INDUCT = prove_thm
  (`LEADSTO_WF_INDUCT`,
   "!(U:^Rel)::ADMIT_WF_INDUCTION.
    !(Pr:^Uprog) p q. !(M:^Metric).
        (!m. LEADSTO Pr (p AND (\s. M s = m)) ((p AND (\s. U (M s) m)) OR q))
        ==> 
        LEADSTO Pr p q",
    RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC (ISPECL ["x:^Rel"; "M:^Metric"] BOUNDED_REACH)
    THEN ASM_REWRITE_TAC [LEADSTO_TRANS_thm; LEADSTO_LDISJ_thm;
                          LEADSTO_GEN_DISJ_thm]
    THEN MATCH_MP_TAC (QUAFF_RULE 1 LEADSTO_REFL)
    THEN XRULE_ASSUM_TAC SPEC_ALL
    THEN IMP_RES_TAC LEADSTO_IMP_UNITY
    THEN ASM_REWRITE_TAC[]) ;;

%----------------------------------------------------------------------------------
                     Impossibility to decrease forever  
----------------------------------------------------------------------------------% 

let EVENTUALLY_FALSE = prove_thm (`EVENTUALLY_FALSE`,
   "!U::ADMIT_WF_INDUCTION.
      !(Pr:^Uprog)(M:^Metric) p. 
      (!m. LEADSTO Pr ((NOT p) AND (\s. M s = m)) (\s. U (M s) m))
      ==> 
      LEADSTO Pr TT p",
    RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
    THEN SUBGOAL_THEN
         "!m:*Dom. LEADSTO (Pr:^Uprog) 
                   ((NOT p) AND (\s. M s = m)) ((\s. x(M s)m) OR p)"
         ASSUME_TAC 
    THENL
      [ %-- prove the lemma introduced by SUBGOAL_THEN --%
        GEN_TAC 
        THEN MATCH_MP_TAC LEADSTO_TRANS
        THEN EXISTS_TAC "\s:^State. (x:^Rel) (M s) m"
        THEN ASM_REWRITE_TAC[]
        THEN MATCH_MP_TAC (QUAFF_RULE 3 LEADSTO_IMPLICATION)
        THEN XRULE_ASSUM_TAC (SPEC "m:*Dom")
        THEN IMP_RES_TAC LEADSTO_IMP_UNITY
        THEN ASM_REWRITE_TAC[] THEN pred_JADE_TAC 2 ;
        ALL_TAC ]
    THEN MATCH_MP_TAC (ISPECL ["x:^Rel"; "M:^Metric"] BOUNDED_ALWAYS_REACH)
    THEN ASM_REWRITE_TAC [LEADSTO_TRANS_thm; LEADSTO_LDISJ_thm;
                          LEADSTO_GEN_DISJ_thm]
    THEN MATCH_MP_TAC (QUAFF_RULE 1 LEADSTO_REFL)
    THEN XRULE_ASSUM_TAC SPEC_ALL
    THEN IMP_RES_TAC LEADSTO_IMP_UNITY
    THEN ASM_REWRITE_TAC[]) ;;

%----------------------------------------------------------------------------------
    Now we are going to prove the most difficult law: the Completion Law.
    First we will need to prove several lemmas.
----------------------------------------------------------------------------------% 

let LT_COMPLETION = prove
  ("LEADSTO (Pr:^Uprog)  p q /\ LEADSTO Pr r s ==>
       (\p q r s.
       !b.  LEADSTO Pr p q /\ LEADSTO Pr r s /\ 
            UNLESS Pr q b  /\ UNLESS Pr s b
            ==> LEADSTO Pr (p AND r) ((q AND s) OR b)
        )p q r s",
    LT_DOUBLE_INDUCT_TAC
    THENL
    [ %-- 1 --%
      EQ_TAC THEN REPEAT STRIP_TAC 
      THEN  ONCE_REWRITE_TAC [pAND_SYM] THEN RES_TAC ;
      %-- 2 --%
      SUBST1_TAC (SYM(ISPEC "((q:^Pred) AND s) OR b" pOR_REFL))
      THEN LEADSTO_CANCEL_TAC "(p:^Pred) AND s"
      THENL
      [ %-- 2.1 --%
        LEADSTO_ANTE_STRONG_TAC "(r:^Pred) AND (p OR q)"
        THENL
        [ %-- 2.1.1 --%
          REWRITE_TAC [ prove
             ("(((q AND s) OR b) OR (p AND s)) = ((s AND (p OR q)) OR (b:^Pred))",
              pred_CYAN_TAC)]
          THEN LEADSTO_PSP_TAC
          THENL 
          [ %-- 2.1.1.1 --%
            ASM_REWRITE_TAC[] ;
            %-- 2.1.1.2 --%
            MATCH_MP_TAC UNLESS_CONSQ_WEAK
            THEN EXISTS_TAC
                 "(((NOT p) AND b) OR ((NOT q) AND q)) OR (q AND (b:^Pred))"
            THEN CONJ_TAC
            THENL
            [ XRULE_ASSUM_TAC (REWRITE_RULE[ENSURES]) 
              THEN MATCH_MP_TAC UNLESS_DISJ
              THEN ASM_REWRITE_TAC[] ;
              pred_JADE_TAC 3 ]
          ] ;
          %-- 2.1.2 --%
          pred_JADE_TAC 3 ] ;
       %-- 2.2 --%
        LEADSTO_PSP_TAC
        THEN ASM_REWRITE_TAC[]
      ] ;
      %-- 3 --%
      SUBGOAL_THEN 
           "!p. W p ==> LEADSTO (Pr:^Uprog) ((\i.i AND r)p) ((q AND s) OR b)"
           (\thm. ASSUME_TAC thm)
      THENL
      [ BETA_TAC THEN XRULE_ASSUM_TAC (QUAFF_RULE 2)
        THEN REPEAT STRIP_TAC THEN OLD_RES_TAC THEN RES_TAC ;
        ALL_TAC ]
      THEN IMP_RES_TAC (RESTRICT_SHIFT LEADSTO_DISJ_LEFT)
      THEN XRULE_ASSUM_TAC BETA_RULE
      THEN ASM_REWRITE_TAC 
           [REWRITE_RULE [I_THM] 
            (ISPECL ["W:^Pred->bool"; "I:^Pred->^Pred"] qAND_RIGHT_DISTR_OR)] ;
      %-- 4 --%
      LEADSTO_TRANS_TAC "(((p AND v) OR (u AND q)) OR (q AND (v:^Pred)))"
      THENL
      [ %-- 4.1 --%
        LEADSTO_ENS_LIFT_TAC
        THEN MATCH_MP_TAC ENSURES_CONJ 
        THEN ASM_REWRITE_TAC[] ;
        %-- 4.2 --%
        MATCH_MP_TAC LEADSTO_SIMPLE_DISJ THEN CONJ_TAC
        THENL 
        [ MATCH_MP_TAC LEADSTO_SIMPLE_DISJ THEN CONJ_TAC
          THENL
          [ OLD_RES_TAC ;
            SUBST1_TAC (ISPECL ["r:^Pred"; "w:^Pred"] pAND_SYM)
            THEN OLD_RES_TAC] ;
          OLD_RES_TAC
        ]
     ]]) ;;


let LEADSTO_COMPL_lemma = prove
  ("!(Pr:^Uprog) p q r s b.
        LEADSTO Pr p q /\ LEADSTO Pr r s /\
        UNLESS Pr q b /\ UNLESS Pr s b
       ==> LEADSTO Pr (p AND r) ((q AND s) OR b)",
    RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC (BETA_RULE LT_COMPLETION)
    THEN ASM_REWRITE_TAC[]) ;;

%----------------------------------------------------------------------------------
              Here is a simple case of Completion Law
----------------------------------------------------------------------------------% 

let LEADSTO_SIMPLE_COMPLETION = prove_thm
  (`LEADSTO_SIMPLE_COMPLETION`,
   "!(Pr:^Uprog) p q r s b. 
        LEADSTO Pr p (q OR b) /\ LEADSTO Pr r (s OR b) /\
        UNLESS Pr q b /\ UNLESS Pr s b ==>
         LEADSTO Pr (p AND r) ((q AND s) OR b)",
    REPEAT STRIP_TAC
    THEN SUBGOAL_THEN 
         "((q AND s) OR b) = (((q OR b) AND (s OR b)) OR (b:^Pred))"
         (\thm. ONCE_REWRITE_TAC [thm])
    THENL [ pred_CYAN_TAC; ALL_TAC]
    THEN (ASSUME_TAC o SPECL ["Pr:^Uprog"; "b:^Pred"]) UNLESS_REFL
    THEN SUBGOAL_THEN 
         "UNLESS Pr (q OR b) b /\ UNLESS (Pr:^Uprog) (s OR b) b"
         (\thm. ASSUME_TAC thm)
    THENL [ OLD_IMP_RES_TAC UNLESS_SIMPLE_DISJ
            THEN XRULE_ASSUM_TAC (REWRITE_RULE [pOR_REFL])
            THEN ASM_REWRITE_TAC[] ; 
            ALL_TAC ]
    THEN MATCH_MP_TAC LEADSTO_COMPL_lemma
    THEN ASM_REWRITE_TAC[]) ;;

let AND_SPLIT = prove
  ("!X V a. (!i:**::\x. (x = a) \/ (V x). X i) = ((!i::V. X i) /\ (X a))",
    REPEAT GEN_TAC THEN RESTRICT_DEF_TAC 
    THEN BETA_TAC THEN EQ_TAC
    THEN REPEAT STRIP_TAC THEN RES_TAC
    THEN ASM_REWRITE_TAC[] 
    THEN XRULE_ASSUM_TAC (REWRITE_RULE[] o SPEC "a:**")
    THEN ASM_REWRITE_TAC[]);;

let pAND_SPLIT = 
   (ONCE_REWRITE_RULE [DISJ_SYM] o ONCE_REWRITE_RULE [EQ_SYM_EQ]) 
    qAND_DOM_SPLIT ;;

let UNLESS_CONJ_lemma = prove
  ("(!i:**::V. UNLESS (Pr:^Uprog) (Q i) b) ==> UNLESS Pr (!!i::V.Q i) b",
    RESTRICT_DEF_TAC 
    THEN REWRITE_TAC [UNLESS; HOA_DEF]
    THEN pred_JADE_TAC 0
    THEN ASM_CASES_TAC "(b:^Pred) t"
    THEN ASM_REWRITE_TAC[] THEN REPEAT STRIP_TAC
    THEN REPEAT OLD_RES_TAC) ;;

let COMPLETION = prove
  ("!Pr:^Uprog. 
    !(W:**->bool). CHF_FINITE W ==>
    (\V. !P Q b. UNITY Pr /\ 
         (!i :: V. LEADSTO Pr (P i) ((Q i) OR b)) /\
         (!i :: V. UNLESS Pr (Q i) b) ==>
         LEADSTO Pr (!! i :: V. P i) ((!! i :: V. Q i) OR b)) W",
   GEN_TAC THEN STRIP_TAC 
   THEN MATCH_MP_TAC CHF_INDUCTION
   THEN BETA_TAC THEN REPEAT STRIP_TAC 
   THENL 
   [ %-- 1 --%
     REWRITE_TAC [qAND_EMPTY ; pOR_ZERO]
     THEN IMP_RES_TAC (RESTRICT_SHIFT LEADSTO_REFL)
     THEN ASM_REWRITE_TAC[] ;
     %-- 2 --%
     ONCE_REWRITE_TAC [EQ_SYM_EQ]
     THEN REWRITE_TAC [pAND_SPLIT]
     THEN UNDISCH_ALL_TAC
     THEN REWRITE_TAC [(BETA_RULE o ISPEC
          "\i:**. LEADSTO (Pr:^Uprog) (P i) ((Q i) OR b)") AND_SPLIT]
     THEN REWRITE_TAC [(BETA_RULE o ISPEC
          "\i:**. UNLESS (Pr:^Uprog) (Q i) b") AND_SPLIT]
     THEN REPEAT STRIP_TAC
     THEN RES_TAC
     THEN IMP_RES_TAC UNLESS_CONJ_lemma
     THEN IMP_RES_TAC (RESTRICT_SHIFT LEADSTO_SIMPLE_COMPLETION) ]) ;;

%----------------------------------------------------------------------------------
                      Here is the Completion Law
----------------------------------------------------------------------------------% 

let LEADSTO_COMPLETION = prove_thm
  (`LEADSTO_COMPLETION`,
   "!Pr:^Uprog::UNITY. !(W:**->bool)::CHF_FINITE.
    !P Q b.
       (!i :: W. LEADSTO Pr (P i) ((Q i) OR b)) /\
       (!i :: W. UNLESS Pr (Q i) b) ==>
        LEADSTO Pr (!! i :: W. P i) ((!! i :: W. Q i) OR b)",
    RESTRICT_DEF_TAC
    THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC ((RESTRICT_DEF_RULE o BETA_RULE) COMPLETION)) ;;

%----------------------------------------------------------------------------------
                      Few corollaries as in CM88
----------------------------------------------------------------------------------% 

let LEADSTO_cor2 = prove_thm
  (`LEADSTO_cor2`, 
   "!(Pr:^Uprog) p b q. 
    (LEADSTO Pr (p AND b) q) /\ (LEADSTO Pr (p AND (NOT b)) q)
            ==> LEADSTO Pr p q",
    RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
    THEN OLD_IMP_RES_TAC LEADSTO_SIMPLE_DISJ
    THEN SUBGOAL_THEN "((p AND (NOT b)) OR (p AND b)) = (p:^Pred)"
         (\thm. XRULE_ASSUM_TAC (SUBS [thm]))
    THENL [ pred_CYAN_TAC ; ASM_REWRITE_TAC[]]) ;;

let LEADSTO_cor3 = prove_thm
  (`LEADSTO_cor3`, 
   "!(Pr:^Uprog) p q r.
         (LEADSTO Pr p q) /\ (UNLESS Pr r FF) 
          ==> LEADSTO Pr (p AND r) (q AND r)",
    RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC (RESTRICT_SHIFT LEADSTO_PSP)
    THEN XRULE_ASSUM_TAC 
         (REWRITE_RULE [ONCE_REWRITE_RULE [pOR_SYM] pOR_UNIT])
    THEN ASM_REWRITE_TAC[]) ;;

close_theory() ;;
