%----------------------------------------------------------------------------------

  This theory provides a number of basic properties (as in CM88) of the UNLESS,
  STABLE, and Inv (strong invariant) operator.

  Main results are:

    - Reflexivity & Monotonicity of UNLESS
    - UNLESS includes IMP
    - Conjunctivity and Disjuctivity of UNLESS, STABLE, & Inv
    - Compositionality of UNLESS, STABLE, & Inv wrt UNITY parallel composition

  Author: Wishnu Prasetya

----------------------------------------------------------------------------------%    


load_library `WF` ;;
load_theory `def_UNITY` ;;
new_theory `unless`;;
loadf `OLD_RES` ;;
loadf `my_misc` ;;
loadf `faust88` ;;

let Autoload thy =    
    let defs = map fst (definitions thy) in
        map (\name. autoload_theory(`definition`,thy,name)) defs;
    let thms = map fst (theorems thy) in
        map (\name. autoload_theory(`theorem`,thy,name)) thms; 
    delete_cache thy; ();;

Autoload `variables` ;;
Autoload `def_UNITY` ;;

let State = ":*var->*value" ;;
let Pred = ":(*var->*value)->bool" ;;
let Action = ":(*var->*value)->(*var->*value)->bool" ;;
let ACTIONS = ":((*var->*value)->(*var->*value)->bool)->bool" ;;
let VARS = ":*var->bool" ;;
let Uprog = ":^ACTIONS # ^Pred # ^VARS # ^VARS" ;;

let REDUCE_UNLESS_TAC =
    REWRITE_TAC [UNLESS; HOA_DEF] THEN pred_JADE_TAC 0 ;;

letrec UNDISCH_LST_TAC lst =
       if (lst=[]) then ALL_TAC
       else (UNDISCH_TAC (hd lst) THEN UNDISCH_LST_TAC (tl lst)) ;;

let contain_noforall trm =
    (not(type_of (find_term is_forall trm))=":bool") ? true ;;

letrec no_forall_lst lst =
       if (lst = []) then []
       else if contain_noforall (hd lst) 
            then (hd lst).(no_forall_lst (tl lst)) 
       else no_forall_lst (tl lst) ;;

letrec UNDISCH_but_forall_TAC (asml,g) =
    UNDISCH_LST_TAC (no_forall_lst asml) (asml,g) ;;

%----------------------------------------------------------------------------------
    Definition of Inv in terms of Hoare triple
----------------------------------------------------------------------------------%

let Inv_HOA = prove_thm (`Inv_HOA`,
   "Inv (Pr:^Uprog) J = 
    ((|== ((INIT Pr) IMP J))  /\ (!A::PROG Pr. HOA(J,A,J)))",
    REWRITE_TAC [Inv; UNLESS; NOT_FF]
    THEN ONCE_REWRITE_TAC [pAND_SYM; pOR_SYM]
    THEN REWRITE_TAC [pAND_UNIT; pOR_UNIT]) ;;

%----------------------------------------------------------------------------------
                TT and FF are invariant and stable
----------------------------------------------------------------------------------%

let Inv_TT = prove_thm 
  (`Inv_TT`, "!(Pr:^Uprog). Inv Pr TT",
    REWRITE_TAC [Inv] THEN REDUCE_UNLESS_TAC) ;;

let Inv_FF = prove_thm 
  (`Inv_FF`, "!(Pr:^Uprog). Inv Pr TT",
    REWRITE_TAC [Inv] THEN REDUCE_UNLESS_TAC) ;;

let STABLE_TT = prove_thm 
  (`STABLE_TT`, "!(Pr:^Uprog). STABLE Pr TT",
    REWRITE_TAC [STABLE] THEN REDUCE_UNLESS_TAC) ;;

let STABLE_FF = prove_thm 
  (`STABLE_FF`, "!(Pr:^Uprog). STABLE Pr FF",
    REWRITE_TAC [STABLE] THEN REDUCE_UNLESS_TAC) ;;

%----------------------------------------------------------------------------------
    reflexivity of UNLESS
----------------------------------------------------------------------------------%

let UNLESS_REFL = prove_thm 
  (`UNLESS_REFL`,
   "!Pr:^Uprog. !p. UNLESS Pr p p",
   REDUCE_UNLESS_TAC THEN RES_TAC) ;;

let UNLESS_ANTI_REFL = prove_thm 
  (`UNLESS_ANTI_REFL`,
   "!Pr:^Uprog. !p. UNLESS Pr p (NOT p)",
   REDUCE_UNLESS_TAC THEN TAUT_TAC) ;;

%----------------------------------------------------------------------------------
    Monotonicity of UNLESS
----------------------------------------------------------------------------------%

let UNLESS_CONSQ_WEAK = prove_thm 
  (`UNLESS_CONSQ_WEAK`,
   "!Pr:^Uprog. !(p:^Pred) q r.
    (UNLESS Pr p q) /\ (|== (q  IMP r)) ==>  (UNLESS Pr p r)",
    REDUCE_UNLESS_TAC 
    THEN ASM_CASES_TAC "(q:^Pred) s"
    THEN REPEAT (RES_TAC THEN ASM_REWRITE_TAC[])) ;;

%----------------------------------------------------------------------------------
    UNLESS Conjunction and Disjunction laws
----------------------------------------------------------------------------------%

let UNLESS_CONJ = prove_thm
  (`UNLESS_CONJ`,
   "!Pr:^Uprog. !p q p' q'.
     UNLESS Pr p q /\ UNLESS Pr p' q'
     ==>
     UNLESS Pr (p AND p') (((p AND q') OR (p' AND q)) OR (q AND q'))",
    REDUCE_UNLESS_TAC
    THEN ASM_CASES_TAC "(q:^Pred) s"
    THEN ASM_CASES_TAC "(q':^Pred) s"
    THEN REPEAT (OLD_RES_TAC THEN ASM_REWRITE_TAC[])) ;; 

let UNLESS_DISJ = prove_thm
  (`UNLESS_DISJ`,
   "!Pr:^Uprog. !p q p' q'.
      UNLESS Pr p q /\ UNLESS Pr p' q'       ==>
      UNLESS Pr (p OR p') 
                ((((NOT p) AND q') OR ((NOT p') AND q)) OR 
                 (q AND q'))",
    REDUCE_UNLESS_TAC
    THEN RES_TAC THEN UNDISCH_but_forall_TAC 
    THEN TAUT_TAC) ;; 

%----------------------------------------------------------------------------------
    UNLESS simple Conjunction and Disjunction laws
----------------------------------------------------------------------------------%

let UNLESS_SIMPLE_CONJ = 
  prove_thm 
  (`UNLESS_SIMPLE_CONJ`,
   "!Pr:^Uprog. !p q p' q'.
         UNLESS Pr p q /\ UNLESS Pr p' q' ==>
         UNLESS Pr (p AND p') (q OR q')",
    REDUCE_UNLESS_TAC
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [(CONJUNCT2 o SPEC_ALL) DE_MORGAN_THM])
    THEN UNDISCH_ALL_TAC THEN REPEAT STRIP_TAC
    THEN REPEAT (OLD_RES_TAC THEN ASM_REWRITE_TAC[])) ;;

let UNLESS_SIMPLE_DISJ = prove_thm
  (`UNLESS_SIMPLE_DISJ`,
   "!Pr:^Uprog. !p q p' q'.
         UNLESS Pr p q /\ UNLESS Pr p' q' ==>
         UNLESS Pr (p OR p') (q OR q')",
    REDUCE_UNLESS_TAC
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [(CONJUNCT2 o SPEC_ALL) DE_MORGAN_THM])
    THEN UNDISCH_ALL_TAC THEN REPEAT STRIP_TAC
    THEN REPEAT (OLD_RES_TAC THEN ASM_REWRITE_TAC[])) ;;

%----------------------------------------------------------------------------------
    STABLE and Inv Conjunction and Disjunction laws
----------------------------------------------------------------------------------%

let STABLE_GEN_DISJ = prove_thm
  (`STABLE_GEN_DISJ`,
   "!(Pr:^Uprog) W f.
       (!i:*ind::W. STABLE Pr (f i)) ==> STABLE Pr (??i::W. f i)",
    REWRITE_TAC [STABLE] THEN REDUCE_UNLESS_TAC
    THEN EXISTS_TAC "i:*ind" THEN ASM_REWRITE_TAC[]
    THEN RES_TAC) ;;
    
let STABLE_GEN_CONJ = prove_thm
  (`STABLE_GEN_CONJ`,
   "!(Pr:^Uprog) W f.
       (!i:*ind::W. STABLE Pr (f i)) ==> STABLE Pr (!!i::W. f i)",
    REWRITE_TAC [STABLE] THEN REDUCE_UNLESS_TAC
    THEN RES_TAC) ;;

let STABLE_SIMPLE_DISJ = prove_thm
  (`STABLE_SIMPLE_DISJ`,
   "!(Pr:^Uprog) p q.
       STABLE Pr p /\ STABLE Pr q  ==> STABLE Pr (p OR q)",
    REWRITE_TAC [STABLE] THEN REDUCE_UNLESS_TAC
    THEN RES_TAC THEN ASM_REWRITE_TAC[]) ;;

let STABLE_SIMPLE_CONJ = prove_thm
  (`STABLE_SIMPLE_CONJ`,
   "!(Pr:^Uprog) p q.
       STABLE Pr p /\ STABLE Pr q  ==> STABLE Pr (p AND q)",
    REWRITE_TAC [STABLE] THEN REDUCE_UNLESS_TAC
    THEN RES_TAC THEN ASM_REWRITE_TAC[]) ;;

let Inv_CONJ = prove_thm
  (`Inv_CONJ`,
   "!H J (Pr:^Uprog). Inv Pr H /\ Inv Pr J  ==>  Inv Pr (H AND J)",
    REWRITE_TAC [Inv_HOA; HOA_DEF] 
    THEN pred_JADE_TAC 3) ;;

let Inv_DISJ = prove_thm
  (`Inv_DISJ`,
   "!H J (Pr:^Uprog). Inv Pr H /\ Inv Pr J  ==>  Inv Pr (H OR J)",
    REWRITE_TAC [Inv_HOA; HOA_DEF] 
    THEN pred_JADE_TAC 3) ;;

%----------------------------------------------------------------------------------
    more general UNLESS Conjunction and Disjunction laws
----------------------------------------------------------------------------------%

let lemma1 = prove
  ("(!i. W i ==> (P:*ind->(^Pred)) i s) ==> (!i. W i ==> P i s \/ Q i s)",
    REPEAT STRIP_TAC THEN RES_TAC THEN ASM_REWRITE_TAC[]) ;;

let lemma2 = prove
  ("~(?i. W i /\ (Q:*ind->(^Pred)) i s) = (!i. W i ==> ~ Q i s)",
    REWRITE_TAC [NOT_EXISTS_CONV "~(?i:*. W i /\ Q i)"]
    THEN EQ_TAC THEN REPEAT STRIP_TAC
    THEN RES_TAC) ;;

let MY_RULE thm =
    let asm1 = ASSUME "(!i. W i ==> (P:*ind->(^Pred)) i s)" in
    let asm2 = ASSUME "(!i. W i ==> ~((Q:*ind->(^Pred)) i s))" in
    let asm3 = ASSUME "(x:^Action) s t" in
    let asm4 = ASSUME "(PROG (Pr:^Uprog)) x" in
    let th1  = (SPEC "x:^Action" o UNDISCH o SPEC "i:*ind") thm in
    let th2  = (SPEC_ALL o REWRITE_RULE [asm4]) th1 in
    let asm5 = ASSUME "(W:*ind->bool) i" in
    let th3  = REWRITE_RULE [MATCH_MP asm1 asm5; MATCH_MP asm2 asm5; asm3]
               th2 in
    (GEN "i:*ind" o DISCH "(W:*ind->bool) i") th3 ;;

let UNLESS_GEN_CONJ = prove_thm
  (`UNLESS_GEN_CONJ`,
       "!Pr:^Uprog. !W (P:*ind->(^Pred)) Q.
       (!i::W. UNLESS Pr (P i) (Q i)) ==>
    UNLESS Pr (!!i::W. P i) ((!!i::W. (P i) OR (Q i)) AND (??i::W. Q i))",
    REDUCE_UNLESS_TAC
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [MP lemma1 
              (ASSUME "!i. W i ==> (P:*ind->(^Pred)) i s") ])
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [lemma2])
    THEN XRULE_ASSUM_TAC MY_RULE
    THEN ASM_CASES_TAC "(?i. W i /\ (Q:*ind->^Pred) i t)"
    THEN ASM_REWRITE_TAC[]
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [lemma2])
    THEN REPEAT STRIP_TAC THEN REPEAT RES_TAC) ;;

let lemma3 = prove
  ("~(!i. W i ==> ~(P:*ind->^Pred) i s \/ Q i s) = 
     (?i. W i /\ P i s /\ ~(Q i s))",
    REWRITE_TAC [NOT_FORALL_CONV "~(!i. W i ==> ~(P:*ind->^Pred) i s \/ Q i s)"]
    THEN EQ_TAC THEN REPEAT STRIP_TAC 
    THEN EXISTS_TAC "i:*ind" 
    THEN UNDISCH_ALL_TAC THEN TAUT_TAC) ;;

let UNLESS_GEN_DISJ = prove_thm
  (`UNLESS_GEN_DISJ`,
   "!Pr. !W (P:*ind->(^Pred)) Q.
    (!i::W. UNLESS Pr (P i) (Q i)) ==>
    UNLESS Pr (??i::W. P i) 
              ((!!i::W. (NOT (P i)) OR (Q i)) AND (??i::W. Q i))",
    REDUCE_UNLESS_TAC
    THEN ASM_CASES_TAC "?i. W i /\ (P:*ind->^Pred) i t"
    THEN ASM_REWRITE_TAC[]
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [
         (CONJUNCT1 o SPEC_ALL) DE_MORGAN_THM])
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [lemma3; lemma2])
    THEN CONJ_TAC
    THENL [ REPEAT STRIP_TAC 
            THEN OLD_RES_TAC THEN ASM_REWRITE_TAC[] ;
            UNDISCH_ALL_TAC 
            THEN  REPEAT STRIP_TAC
            THENL [ EXISTS_TAC "i':*ind" 
                    THEN REPEAT (OLD_RES_TAC THEN ASM_REWRITE_TAC[]) ;
                    EXISTS_TAC "i:*ind" 
                    THEN REPEAT (OLD_RES_TAC THEN ASM_REWRITE_TAC[])]]);;

%----------------------------------------------------------------------------------
    a cancelation law for UNLESS
----------------------------------------------------------------------------------%

let UNLESS_CANCEL = prove_thm
  (`UNLESS_CANCEL`,
   "!Pr:^Uprog. !(p:^Pred) q r.
     UNLESS Pr p q /\ UNLESS Pr q r ==> UNLESS Pr (p OR q) r",
    REDUCE_UNLESS_TAC 
    THEN ASM_CASES_TAC "(q:^Pred) s"
    THEN REPEAT (OLD_RES_TAC THEN ASM_REWRITE_TAC[])) ;;

%----------------------------------------------------------------------------------
    UNLESS includes IMP, and some corollaries.
----------------------------------------------------------------------------------%

let UNLESS_IMP_LIFT1 = prove_thm
  (`UNLESS_IMP_LIFT1`,
   "!Pr:^Uprog. 
    !p q. (|== (p IMP q)) ==> UNLESS Pr p q",
    REDUCE_UNLESS_TAC
    THEN ASM_CASES_TAC "(q:^Pred) t"
    THEN ASM_REWRITE_TAC[]
    THEN XRULE_ASSUM_TAC (GEN_ALL o CONTRAPOS o SPEC "s:^State")
    THEN RES_TAC) ;;

let UNLESS_IMP_LIFT2 = prove_thm
  (`UNLESS_IMP_LIFT2`,
   "!Pr:^Uprog. !p q. 
   (|== ((NOT p) IMP q)) ==> UNLESS Pr p q",
    REDUCE_UNLESS_TAC
    THEN ASM_CASES_TAC "(p:^Pred) t"
    THEN ASM_REWRITE_TAC[] THEN RES_TAC) ;;

let UNLESS_cor3 = prove_thm
  (`UNLESS_cor3`,
   "!Pr:^Uprog. !p q r.
        (UNLESS Pr p (q OR r)) = (UNLESS Pr (p AND (NOT q)) (q OR r))",
    RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC THEN EQ_TAC
    THEN REDUCE_UNLESS_TAC
    THENL
    [ OLD_RES_TAC THEN OLD_RES_TAC
      THEN ASM_REWRITE_TAC[] THEN TAUT_TAC ;
      XRULE_ASSUM_TAC (REWRITE_RULE [(CONJUNCT2 o SPEC_ALL) DE_MORGAN_THM])
      THEN UNDISCH_ALL_TAC THEN REPEAT STRIP_TAC
      THEN REPEAT (OLD_RES_TAC THEN ASM_REWRITE_TAC[]) ] ) ;;

let UNLESS_cor4 = prove_thm
  (`UNLESS_cor4`,
   "!Pr:^Uprog. !p q r.
        UNLESS Pr (p OR q) r ==> UNLESS Pr p (q OR r)",
    REDUCE_UNLESS_TAC
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [(CONJUNCT2 o SPEC_ALL) DE_MORGAN_THM])
    THEN UNDISCH_ALL_TAC THEN REPEAT STRIP_TAC
    THEN REPEAT (OLD_RES_TAC THEN ASM_REWRITE_TAC[]) ) ;;

let UNLESS_cor5 = prove_thm 
  (`UNLESS_cor5`,
   "!Pr:^Uprog. !W (P:*ind->(^Pred)) Q.
    (!i::W. UNLESS Pr (P i) ((P i) AND (Q i))) ==>
    UNLESS Pr (!!i::W. P i) ((!!i::W. P i) AND (??i::W. Q i))",
    REDUCE_UNLESS_TAC
    THEN DISJ1_TAC 
    THEN REPEAT STRIP_TAC THEN OLD_RES_TAC
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [prove("(p==>F) ==> ~p", TAUT_TAC)])
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [lemma2])
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [(CONJUNCT1 o SPEC_ALL) DE_MORGAN_THM])
    THEN REPEAT (OLD_RES_TAC THEN ASM_REWRITE_TAC[]) ) ;;

%----------------------------------------------------------------------------------
    Compositionality of UNLESS, Inv, and STABLE wrt PAR
----------------------------------------------------------------------------------%

let UNLESS_PAR = prove_thm (`UNLESS_PAR`,
   "! p q Pr (Qr:^Uprog).
    UNLESS Pr p q  /\ UNLESS Qr p q  ==> UNLESS (Pr PAR Qr) p q",
    REWRITE_TAC [UNLESS; PAR; UNLESS; PROG; pOR_DEF]
    THEN RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
    THEN RES_TAC ) ;;

let UNLESS_PAR_i = prove_thm 
  (`UNLESS_PAR_i`,
   "UNLESS (Pr:^Uprog) p q  /\ UNLESS Qr p q  = UNLESS (Pr PAR Qr) p q",
    REWRITE_TAC [UNLESS; PAR; UNLESS; PROG; pOR_DEF; HOA_DEF]
    THEN pred_JADE_TAC 0
    THEN FAUST_TAC) ;;
    % Run time: 8.5s
      Garbage collection time: 2.5s
      Intermediate theorems generated: 1770 %

let Inv_PAR = prove_thm (`Inv_PAR`,
   "! J Pr (Qr:^Uprog).
       (Inv Pr J) /\ (Inv Qr J) ==> (Inv (Pr PAR Qr) J)",
    REWRITE_TAC [PAR; Inv_HOA; HOA_DEF; INIT; PROG]
    THEN pred_JADE_TAC 3) ;;

let STABLE_PAR = prove_thm
  (`STABLE_PAR`,
   "!(Pr:^Uprog) Qr p. STABLE Pr p /\ STABLE Qr p ==> STABLE (Pr PAR Qr) p",
    REWRITE_TAC [STABLE; UNLESS_PAR]) ;;

let STABLE_PAR_i = prove_thm
  (`STABLE_PAR_i`,
   "STABLE (Pr:^Uprog) p /\ STABLE Qr p = STABLE (Pr PAR Qr) p",
    REWRITE_TAC [STABLE; UNLESS_PAR_i]) ;;

close_theory() ;;



